/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *          Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _Frame_H
#define _Frame_H
#include "streamringbuffer.h"
#include <utils/iplImageWrapper.h>
#include <cv.h>
#include <string>
#include <vector>
#include "packets/imu_raw_packets.h"

#include <timeutils/timer.hpp>
#include <vslam/features/feature_points.hpp>
#include <vslam/matcher/match.hpp>
#include <vslam/map/feature_map.hpp>
#include <vslam/core/pose_as_matrix.hpp>

namespace vslam {
namespace map {

class landmark;

}  // namespace map
}  // namespace vslam

using namespace std;

template <typename T>
class DataStream
{
protected:
  std::vector<T> stream;
public:
  int maximum_size;
  int last_write;
  bool wrapped_around;
  
  DataStream(unsigned int _maximum_size=100000) : maximum_size(_maximum_size), last_write(-1), wrapped_around(false)
  {
    stream.resize(maximum_size);
  }
  
  ~DataStream()
  {
  }
  
  void append(T& data)
  {
    if (last_write + 1 >= maximum_size)
      wrapped_around = true;
    
    last_write = (last_write + 1) % maximum_size;        
    stream[last_write] = data;
  }
  
  void clear()
  {
    last_write = -1;
    stream.clear();
  }
  
  T& operator[](int idx)
  {
    return stream[idx];
  }    
};


class Frame
{
public:

  Frame();
  ~Frame();
  
  static uint64_t timespec_to_mu_s(const timespec& ts)
  {
      return ts.tv_sec * 1e6 + ts.tv_nsec / 1000;
  }
  std::string get_filename() const;
  
  long long number;
  int cam_id;
  
  struct timespec cam_ts;  
  struct timespec os_ts;
  struct timespec physical_event_ts;

  IplImageWrapper video;

  bool is_log;  
  double delta_t;
  double total_time;

  std::vector<IMURDataPacket> imu_measurements;
  static DataStream<IMURDataPacket> imu_measurement_log;

  // extracted features
  vslam::feature_points features;

  // matches (stored for camera 0 Frame only)
  std::vector<vslam::matcher::stereo_match> stereo_features;
  std::vector<vslam::matcher::mono_feature> mono_features;
  
  vslam::pose_as_matrix world2cam;

  typedef std::pair<const vslam::matcher::stereo_match*, const vslam::map::landmark*> map_match;
  std::vector<map_match> map_matches;

  typedef std::pair<const vslam::matcher::mono_feature*, const vslam::map::landmark*> mono_match;
  std::vector<mono_match> mono_matches;

  vslam::map::feature_map feature_map;
  std::vector<vslam::map::feature_map::inter_frame_match> inter_frame_matches;

  unsigned stat_known_landmarks;
  unsigned stat_new_landmarks;
  
  timeutils::timer t;
};

typedef StreamRingBuffer<Frame> Buffer;
#endif
