/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *          Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

// C++ STL Header
#include <limits>
#include <sstream>
#include <string>
#include <iostream>
#include <istream>
#include <iterator>
#include <iomanip>

#include <cstdlib>
#include <cstdio>
#include <sys/mman.h>

#include <sched.h>

#include <camera/file_camera.h>

#include "processing_stack.h"
#include "plugin_visualisation.h"
#include "plugin_test.h"
#include "plugin_dump.h"
#include "plugin_compare.h"

#include "plugin_null.h"
#include "plugin_imu.h"
#include "plugin_extract_features.h"
#include "plugin_stereo_match.h"
#include "plugin_optimizer.h"
#include "plugin_dense_stereo.h"
#include "processing_thread.h"
#include "visualisation_window.h"
#include "transformations.h"
#include "image_capture_thread.h"
#include "imu_acquisition_file.h"
#include <sys/resource.h>

#include "program_config.h"
#include "frame.h"

enum data_input_type
{
    INPUT_TYPE_CAM =0,
    INPUT_TYPE_LOG_LEGACY,
    INPUT_TYPE_LOG,
    INPUT_TYPE_LOG_RAW    
};

int wrapped_main(int argc, char** argv)
{

  // Create buffer for ard's arguments
  char** ard_argv = new char*[argc];
  int ard_argc = 0;

  // Create buffer for arguments for program options (without any ard stuff)
  char** po_argv = new char*[argc];
  int po_argc = 0;
  
  // Separate ard arguments from the rest
  separate_arguments(argc, argv, ard_argc, ard_argv, po_argc, po_argv);

  // Propagate non-ard arguments through boost::program_options
  po::variables_map vm;
  program_config pc;
  do_program_options(po_argc, po_argv, vm, pc);

  // Lock memory
  if (mlockall(MCL_CURRENT | MCL_FUTURE) == 0)
    std::cerr << "Memory successfully locked!" << std::endl;
  else
    std::cerr << "Error: locking of memory failed!" << std::endl;

  struct rlimit l;
  getrlimit(RLIMIT_MEMLOCK, &l);
  std::cerr << l.rlim_cur << " " << l.rlim_max << std::endl;
  
  std::string log_dir("");
  enum data_input_type data_input = INPUT_TYPE_CAM;

  if(vm.count("log_normal")) 
  {	
    log_dir = pc.log_dir;
    data_input = INPUT_TYPE_LOG;
  }
  else if(vm.count("log_legacy"))
  {	
    log_dir = pc.log_dir;
    data_input = INPUT_TYPE_LOG_LEGACY;
  }
  else if(vm.count("log_raw"))
  {
    log_dir = pc.log_dir;
    data_input = INPUT_TYPE_LOG_RAW;
  }
  
  if(vm.count("priority"))
  {
    if (pc.priority >= 0 && pc.priority < 100)
    {
      std::cerr << "Trying to set priority " << pc.priority << "."<< std::endl;
      if (pc.priority >= sched_get_priority_min(SCHED_FIFO) && pc.priority <= sched_get_priority_max(SCHED_FIFO))
      {
        struct sched_param sp;
		memset(&sp, 0, sizeof(sp));
        sp.sched_priority = pc.priority;
        if (sched_setscheduler(0, SCHED_FIFO, &sp) == 0)
        {
          // Test if could set the highest rtprio
          std::cerr << "Success! Priority of process will never be higher than " << pc.priority << " using policy SCHED_FIFO."<< std::endl;
          // And set the actual lower one
          sp.sched_priority = pc.priority-5;
          sched_setscheduler(0, SCHED_FIFO, &sp);
        }
        else
        {
          std::cerr << "Failure! Priority could not could not be set! " << std::endl;
        }
      }
      else
        std::cerr << "Error: priority value is not a valid value!" << std::endl;
      
    }
    else
    {
      std::cerr << "Command line error: priority value is missing!" << std::endl;
      exit(1);
    }
  }

  bool do_mainloop(true);
  
  Generic_Camera *c;
  Generic_Camera *c2;

  if (data_input != INPUT_TYPE_CAM)  
  {
      c = new File_Camera(log_dir, "l", pc.image_width, pc.image_height, pc.play_once, do_mainloop, pc.skip_frames);
      c2 = new File_Camera(log_dir, "r", pc.image_width, pc.image_height, pc.play_once, do_mainloop, pc.skip_frames);
  }
  else
  {
    std::cerr << "Require log file\n";
    exit(1);
  }


  std::ifstream in;
  if (data_input != INPUT_TYPE_CAM)
  {      
      in.open(std::string(log_dir + "/calibration.txt").c_str());
      if (!in.is_open())
      {
          std::cerr << "Error opening calibration file from log directory. Are you sure it exists?" << std::endl;
          exit(1);
      }
  }
  else
  {     
     char *home_path  = getenv ("HOME");
     in.open(std::string(std::string(home_path) + "/.slam_vision/calibration.txt").c_str());
      if (home_path == 0 || !in.is_open())
      { 
          std::cerr << "Error opening calibration file from ~/.slam_vision directory.  Are you sure it exists?" << std::endl;
          exit(1);
      }      
  }
  
  Transform::CalibrationFromStream(in);
  
   Buffer *fb_left = new Buffer(pc.framebuffer_size);
   Buffer *fb_right = new Buffer(pc.framebuffer_size);
  
   images_rb_t *img_left = new images_rb_t();
   images_rb_t *img_right = new images_rb_t();

   ImageCaptureThread* capture_left = new ImageCaptureThread(*img_left, *img_right, *c, 0);
   ImageCaptureThread* capture_right = new ImageCaptureThread(*img_right, *img_left, *c2, 1);
  
   // Create ProcessingStack
   ProcessingStack* processingstack_left  = new ProcessingStack("Left camera vision stack");
   ProcessingStack* processingstack_right = new ProcessingStack("Right camera vision stack");

   imu_raw_packets_rb_t* imu_data_rb = new imu_raw_packets_rb_t();

   vslam::map::map global_map(feature_extraction_config::descriptor_size);

   vslam::optimizer optimizer(Transform::left2imu.inverse(),
                              Transform::cam.get_left2right() * Transform::left2imu.inverse(),
                              pc.optimizer_verbosity, pc.velocity_sigma);

   // LEFT and RIGHT stack interleaved
   processingstack_left ->stack.push_back(new PluginIMU("IMU (left)",fb_left, *imu_data_rb, pc.lag_imu));
   processingstack_right->stack.push_back(new PluginIMU("IMU (right)",fb_right, *imu_data_rb, pc.lag_imu));
   processingstack_left ->stack.push_back(new PluginTest("Test (left)", fb_left, fb_right));
   processingstack_right->stack.push_back(new PluginTest("Test (right)", fb_right, fb_left));
   if(pc.extract_features)
   {

     processingstack_left ->stack.push_back(new PluginExtractFeatures("ExtractFeatures (left)", fb_left, fb_right, pc.feature_limit));
     processingstack_right->stack.push_back(new PluginExtractFeatures("ExtractFeatures (right)", fb_right, fb_left, pc.feature_limit));
     processingstack_left ->stack.push_back(new PluginStereoMatch("StereoMatch (left)", fb_left, fb_right, pc));
     processingstack_right->stack.push_back(new PluginNull("Null (StereoMatch right)", fb_right));
     processingstack_left ->stack.push_back(new PluginOptimizer("Optimizer (left)", fb_left, fb_right, optimizer, global_map, pc));
     processingstack_right->stack.push_back(new PluginDenseStereo("Dense Stereo (right)", fb_right, fb_left, pc, optimizer));
     if(pc.dump_landmarks
		|| pc.dump_trajectory)
	 {
		 processingstack_left ->stack.push_back(new PluginDump("Dump (left)", fb_left, fb_right,
															   optimizer.get_poses(), global_map,
															   pc));
		processingstack_right->stack.push_back(new PluginNull("Null (Dump right)", fb_right));
     }
	 if (pc.compare_trajectory)
	 {
		 processingstack_left ->stack.push_back(new PluginCompare("Compare (left)", fb_left, fb_right,
															   optimizer.get_poses(), global_map,
															   pc));
		processingstack_right->stack.push_back(new PluginNull("Null (Compare right)", fb_right));
	 }
   }
   VisualisationWindow* vw = 0;
   
   if (!pc.nox)
   {
   vw = new VisualisationWindow(pc.image_width, pc.image_height, pc.dsf);

   PluginVisualisation* pvl = new PluginVisualisation("Visualisation (left)",fb_left, fb_right, global_map, pc.image_width, pc.image_height, pc.dsf, pc, vw);
   PluginVisualisation* pvr = new PluginVisualisation("Visualisation (right)", fb_right, fb_left, global_map, pc.image_width, pc.image_height, pc.dsf, pc, vw, pvl);
   processingstack_left ->stack.push_back(pvl);
   processingstack_left->stack.push_back(pvr);
   processingstack_right->stack.push_back(new PluginNull("Null", fb_right));
   processingstack_right->stack.push_back(new PluginNull("Null", fb_right));
   }
   
   
   std::ofstream timeleft("timing_left.dat"), timeright("timing_right.dat");

   ProcessingThread* pt_left = new ProcessingThread(0, c, *img_left, pc.lag_cam, pc.rotate_left_cam, timeleft);
   ProcessingThread* pt_right = new ProcessingThread(1, c2, *img_right, pc.lag_cam, pc.rotate_right_cam, timeright);
  
   pt_left->setBuffer(fb_left);
   pt_right->setBuffer(fb_right);

   pt_left->setProcessingStack(processingstack_left );
   pt_right->setProcessingStack(processingstack_right);
   
   pt_left->init();
   pt_right->init();

   processingstack_right->setup_barriers();

   Thread* imuqt;

   {
       imuqt = new IMUAcquisitionFile(*imu_data_rb, log_dir+"/imu_data.dat");       
   }
  
   pt_left->start(pc.priority);
   pt_right->start(pc.priority);   
   
   imuqt->start(pc.priority+10);  
   
   capture_right->start(pc.priority);  
   capture_left->start(pc.priority);

   SDL_Event event;  
   while(do_mainloop)
   {
      if (pc.nox)
	  {
		  usleep(50000);
		  continue;
	  }
	   
      while (SDL_PollEvent(&event))
      {
          if (event.type == SDL_QUIT || (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE)) 
          {
              do_mainloop = false;
              break;
          }             
          else
          {
              processingstack_right->handle_event(event);
              processingstack_left ->handle_event(event);             
          }
      }
     
      vw->update();
      SDL_Delay(50);
   }

   imuqt->stop();
   delete imuqt;

   capture_left->stop();
   capture_right->stop();

   pt_left->stop();
   pt_right->stop();

   delete pt_left;
   delete pt_right;
   delete capture_left;
   delete capture_right;
 
   delete processingstack_left ;
   delete processingstack_right;

   delete fb_left;
   delete fb_right;

   delete c;
   delete c2;

   if (!pc.nox)
	   SDL_Quit();
  
  return 0;
}

int main(int argc, char **argv)
{
	try
	{
		return wrapped_main(argc, argv);
	}
	catch(const std::exception &e)
	{
		std::cerr << "exception: " << e.what() << std::endl;
	}
	catch(const char *e)
	{
		std::cerr << "exception: " << e << std::endl;
	}
	return 1;
}
