/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *          Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "plugin_compare.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>

#include "program_config.h"

const static double delta_norm_threshold = 1e-2;

PluginCompare::PluginCompare(std::string _name, Buffer * _buffer, Buffer * _other_buffer,
					   const vslam::optimizer::pose_list &trajectory, const vslam::map::map &global_map,
					   const program_config &pc)
    : ProcessingPlugin(_name, _buffer), other_buffer(_other_buffer),
	  trajectory_(trajectory),
	  global_map_(global_map),
	  pc_(pc)
{
}


PluginCompare::~PluginCompare()
{
}

static void fail(const char *msg)
{
	std::cerr << "CMP: error: " << msg << std::endl;
	throw msg;
}

Result PluginCompare::process(Frame * data)
{
	typedef vslam::optimizer::Pose pose_t;
	
    // Only camera 0 should do some stuff
    if (data->cam_id == 0)
    {
        // Get other Frame from other camera
        //Frame *other_data = other_buffer->getPointer(other_buffer->curWrite());

		std::stringstream suffix;
		suffix << std::setfill('0') << std::setw(4) << data->number << ".dat";

		std::stringstream fname;
		fname << pc_.dump_dir << "/trajectory-" << suffix.str();
		std::ifstream f(fname.str().c_str());

		if (!f.is_open())
			fail("could not open trajectory file to compare with");

		pose_t pose;

		vslam::optimizer::pose_list::const_iterator i;
		vslam::optimizer::pose_list::size_type k = 0;
		double max_delta_norm = 0;

		for (i = trajectory_.begin();
			 f.is_open() && !f.eof() && f.good()
				 && i != trajectory_.end(); ++i, ++k)
		{
			f >> pose.orient.x();
			if (f.eof())
				fail("premature EOF");
			f >> pose.orient.y() >> pose.orient.z() >> pose.orient.w()
			  >> pose.pos.x() >> pose.pos.y() >> pose.pos.z()
			  >> pose.vel.x() >> pose.vel.y() >> pose.vel.z();

			const pose_t reference = (**i);
			Eigen::Matrix<double, pose_t::DOF, 1> delta;
			
			pose.boxminus(delta.data(), reference);
			max_delta_norm = std::max(delta.norm(), max_delta_norm);

			if (delta.norm() > delta_norm_threshold)
			{
				std::cerr << "CMP: found different pose #" << k << " in trajectory #" << data->number << ": " << delta.norm() << std::endl;
				std::cerr.precision(20);
				std::cerr << reference.orient.x() << " " << reference.orient.y() << " " << reference.orient.z() << " " << reference.orient.w() << " "
						  << reference.pos.x() << " " << reference.pos.y() << " " << reference.pos.z()
						  << reference.vel.x() << " " << reference.vel.y() << " " << reference.vel.z()
						  << std::endl;
				std::cerr << pose.orient.x() << " " << pose.orient.y() << " " << pose.orient.z() << " " << pose.orient.w() << " "
						  << pose.pos.x() << " " << pose.pos.y() << " " << pose.pos.z()
						  << pose.vel.x() << " " << pose.vel.y() << " " << pose.vel.z()
						  << std::endl;
				
				fail("trajectory differs");
			}
		}

		if (k > trajectory_.size())
		{
			std::cerr << "CMP: trajectory #" << data->number << " reference size " << k << std::endl;
			fail("less poses than in reference trajectory");
		}
		else if (k < trajectory_.size())
		{
			std::cerr << "CMP: trajectory #" << data->number << " reference size " << k << std::endl;
			fail("more poses than in reference trajectory");
		}
		std::cout << "PluginCompare: Maximal difference was: " << max_delta_norm << std::endl;
    }    

    return Ok;  
}

bool PluginCompare::event(SDL_Event &event)
{
    bool event_handled = false;
    
    return event_handled;
}
