/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *          Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include <iomanip>
#include <sstream>
#include <utility>

#include <Eigen/Core>

#include <octomap/octomap.h>

#include <vslam/core/optimizer.hpp>
#include "plugin_dense_stereo.h"
#include "transformations.h"

#include <highgui.h>

typedef std::pair<vslam::optimizer::Pose_id, vslam::dense_stereo::pointcloud> scan;
typedef std::deque<scan> scan_list;

static scan_list scans;

PluginDenseStereo::PluginDenseStereo(std::string _name, Buffer * _buffer, Buffer * _other_buffer, const program_config& pc, const vslam::optimizer& opti)
    : ProcessingPlugin(_name, _buffer), other_buffer(_other_buffer),
	  ds_(Transform::cam, pc.min_disparity, pc.disparity_range),
	  pc_(pc),
	  opti_(opti)
{
}


PluginDenseStereo::~PluginDenseStereo()
{
}


Result PluginDenseStereo::process(Frame * data)
{   

    // Only camera 1 should do some stuff
    if (data->cam_id == 1
		&& opti_.initialized())
    {
		Frame *other_data = other_buffer->getPointerAt(other_buffer->getCurrentWriteIdx());
		cv::Mat left(other_data->video.p);
		cv::Mat right(data->video.p);

		scans.push_back(std::make_pair(opti_.last_pose(),
									   vslam::dense_stereo::pointcloud()));
		
		timeutils::timer t;
		ds_(left, right, scans.back().second);

		std::cerr << "dense stereo: " << t.elapsed() << "us" << std::endl;

		if (scans.size() == size_t(pc_.dump_octomap))
			dump_octree();

		if (pc_.dump_disparities)
			dump_disparity(data->number);
    }    

    return Ok;  
}

void PluginDenseStereo::dump_disparity(const long long &number)
{
	const cv::Mat &disparity = ds_.disparity();

	float disparity_min = 1e9;
	float disparity_max = 0;
	for (cv::MatConstIterator_<float> it = disparity.begin<float>();
		 it != disparity.end<float>(); ++it)
	{
		if (*it >= 0)
		{
			disparity_min = std::min(disparity_min, *it);
			disparity_max = std::max(disparity_max, *it);
		}
	}

	cv::Mat disp(disparity.size(), CV_8UC1, cv::Scalar(0));
	cv::MatConstIterator_<float> src = disparity.begin<float>();
	cv::MatIterator_<uchar> dst = disp.begin<uchar>();
	for (; src != disparity.end<float>() && dst != disp.end<uchar>(); ++src, ++dst)
	{
		const float &val = *src;
		if (val > 0)
			*dst =  (val - disparity_min) * 255 / (disparity_max - disparity_min);
	}

	std::ostringstream s;
	s << "disparity-" << std::setw(5) << std::setfill('0') << number << ".png";

	using namespace cv;
	
	unsigned w = ds_.left().size().width;
	
	Mat large(ds_.left().size().height, w * 3, CV_8UC1, cv::Scalar(0));
  
	Mat l = large.colRange(0,w);
	Mat d = large.colRange(w, 2 * w);
	Mat r = large.colRange(2 * w, 3 * w);

	cv::resize(ds_.left_undistrect(), l, l.size());
	cv::resize(ds_.right_undistrect(), r, r.size());
	cv::resize(disp, d, d.size());

	
	//cv::imwrite(s.str(), disp);
	cv::imwrite(s.str(), large);
}

bool PluginDenseStereo::event(SDL_Event &event)
{
    bool event_handled = false;

	switch(event.type)
	{
	case SDL_KEYUP:
		if(event.key.keysym.sym==SDLK_d)
		{
			//std::cout << "dense: event" << std::endl;

			opti_.big_lock(boost::bind(&self::dump_octree, this));
			
			event_handled = true;
		}
		else if(event.key.keysym.sym == SDLK_p)
		{
			std::cout << "pose-dump" << std::endl;
			opti_.big_lock(boost::bind(&self::dump_poses, this));
			event_handled = true;
		}
		break;
	default:
		break;
	}
    
    return event_handled;
}

void PluginDenseStereo::dump_octree()
{
	typedef vslam::camera::Vector vect;
	using namespace octomap;

	boost::shared_ptr<ScanGraph> g(new ScanGraph());
	
	for (scan_list::const_iterator i = scans.begin();
		 i != scans.end(); ++i)
	{
		const vslam::optimizer::Pose& pose = *i->first;

		Pointcloud *scan = new Pointcloud();

		pose6d scan_pose(point3d(pose.pos.x(), pose.pos.y(), pose.pos.z()),
						 octomath::Quaternion(pose.orient.w(), pose.orient.x(), pose.orient.y(), pose.orient.z()));

		for (vslam::dense_stereo::pointcloud::const_iterator p = i->second.begin();
			 p != i->second.end(); ++p)
		{
			double d = p->norm();
			if (pc_.min_stereo_range < d && d < pc_.max_stereo_range)
			{
				vect endp_ = Transform::left2imu.apply(ds_.left2rectified().transpose() * (*p));
				scan->push_back(endp_[0], endp_[1], endp_[2]);
			}
		}
		
		g->addNode(scan, scan_pose);
		
		if (i != scans.begin())
			g->connectPrevious();
	}

	std::stringstream path;
	path << pc_.dump_dir << "/dense_stereo.graph";
	g->writeBinary(path.str().c_str());
}

void PluginDenseStereo::dump_poses()
{
	std::ofstream posefile("poses.dat");
	posefile << "# position 1-3, velocity 1-3, orientation 1-4\n";
	for (scan_list::const_iterator i = scans.begin();
		 i != scans.end(); ++i)
	{
		const vslam::optimizer::Pose& pose = *i->first;
		posefile << pose.pos.transpose() << ", " << pose.vel.transpose() << ", " << pose.orient.coeffs().transpose() << std::endl;
	}
	posefile.close();
}
