/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *          Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>

#include <boost/bind.hpp>

#include "plugin_optimizer.h"
#include "transformations.h"

#include <vslam/core/optimizer.hpp>

#include <vslam/map/map.hpp>
#include <vslam/camera/camera.hpp>

typedef vslam::camera::Vector Vector;


PluginOptimizer::PluginOptimizer(std::string _name, Buffer * _buffer, Buffer * _other_buffer,
                                 vslam::optimizer& opti,
                                 vslam::map::map &global_map,
                                 const program_config& pc)
    : ProcessingPlugin(_name, _buffer), other_buffer(_other_buffer),
	  opti_(opti),
	  global_map_(global_map),
	  imu_initialization_measurements(0),
	  pc(pc)
{
	gyro_bias.setZero();
	init_gravity.setZero();
}


PluginOptimizer::~PluginOptimizer()
{
}

/**
 * transforms a cv::Keypoint to an Eigen::Vector2d and returns a meaningful sigma
 */

void PluginOptimizer::add_measurement(const vslam::matcher::stereo_match &feature,
									  const vslam::map::landmark &landmark)
{
	Eigen::Vector2d left_normalized, right_normalized;
	double sigma_left  = Transform::cam.get_cam(true) .keypoint2Normalized(feature.left,  left_normalized);
	double sigma_right = Transform::cam.get_cam(false).keypoint2Normalized(feature.right, right_normalized);
	
	opti_.add_landmark_measurement(
			left_normalized, sigma_left,
			right_normalized, sigma_right,
			landmark);
}

void PluginOptimizer::add_measurement(const vslam::matcher::mono_feature &feature,
									  const vslam::map::landmark &landmark, const vslam::camera & cam, bool is_left)
{
	Eigen::Vector2d normalized;
	double sigma = cam.keypoint2Normalized(feature.keypoint, normalized);
	opti_.add_mono_landmark_measurement(normalized, sigma,
			landmark, is_left);
}



void PluginOptimizer::process_known_feature(Frame * data,
											const vslam::matcher::stereo_match &feature,
											const vslam::map::landmark &landmark,
											const float &dist)
{
	if(data->map_matches.size() + data->mono_matches.size() >= pc.feature_association_limit)
		return;
	// project landmark to left and right image:
	const Vector lm_world(landmark.position->data()); 
	const Vector lm_in_left_cam = data->world2cam * lm_world;

	double max_im_dist2 = Transform::cam.max_dist2(lm_in_left_cam, feature.left.pt, feature.right.pt);
	
	// check distance between projected landmark and features in pixel, (before optimization):
	if (sqrt(max_im_dist2) < pc.association_threshold) 
	{
		// and pass to slom
		add_measurement(feature, landmark);
		
		// store for later visualization
		data->map_matches.push_back(std::make_pair(&feature, &landmark));
	}
}
void PluginOptimizer::process_mono_feature(Frame * data,
											const vslam::matcher::mono_feature &feature,
											const vslam::map::landmark &landmark,
											const float &dist)
{
	if(data->map_matches.size() + data->mono_matches.size() >= pc.feature_association_limit)
		return;
	assert (data->cam_id == 0);
	
	bool left = (feature.dist_l > feature.dist_r);
	float dist_sig = left ? feature.dist_l : feature.dist_r;
	
	// make sure signature distance is lower than distance to nearest feature in other image:
	if(dist > dist_sig)
		return;
	Frame *cam_data = left ? data : other_buffer->getPointerAt(other_buffer->getCurrentWriteIdx());
	const vslam::camera cam = Transform::cam.get_cam(left);
	
	// Now check distance in image:
	const Vector lm_world(landmark.position->data()); 
	Vector lm_in_cam = cam_data->world2cam * lm_world;

	double dist_img = cam.dist2(lm_in_cam, feature.keypoint.pt);
	
	if(sqrt(dist_img) < pc.association_threshold)
	{
		add_measurement(feature, landmark, cam, left);
		// for visualization of match
		cam_data->mono_matches.push_back(std::make_pair(&feature, &landmark));
	}
}

void PluginOptimizer::process_unknown_feature(Frame * data, const vslam::matcher::stereo_match &feature,
											  const float &dist_new, const float &dist_known)
{
	// previously unseen feature

	vslam::map::map &map = global_map_;
	
	// calculate position based on stereo and current pose
	const Eigen::Vector3d &p = feature.world_point_in_left_camera();
	Vector p_in_left_camera(p.x(), p.y(), p.z());

	Vector p_in_world_ = opti_.current_imu2world() * Transform::left2imu * p_in_left_camera;
	Eigen::Vector3d p_in_world(p_in_world_[0], p_in_world_[1], p_in_world_[2]);
			
	// check distance to nearest neighbor in world coordinates
	float dist;
	const vslam::map::landmark *closest = map.closest_landmark(p_in_world, &dist);

	if (!closest
		|| (p.norm() < pc.max_landmark_distance
			&& dist > pc.min_relative_landmark_distance * p.norm()))
	{
		// pass new landmark to slom if threshold exceeded
		vslam::landmark_id lm_id = opti_.add_landmark(p_in_world);
		vslam::map::landmark &m = map.push_back(vslam::map::landmark(lm_id,
									     feature.signature));
		add_measurement(feature, m);

		// store for later visualization
		//data->map_matches.push_back(std::make_pair(&feature, &m));
	}
}


bool PluginOptimizer::process_imu(const std::vector<IMURDataPacket> imu_measurements)
{
	const int imu_initialization_measurements_required = pc.imu_initialization_frames;
	for (std::vector<IMURDataPacket>::const_iterator i = imu_measurements.begin();
			i != imu_measurements.end(); ++i)
	{
		// add IMU measurement:
		vec3 a = Eigen::Vector3d(i->a[0], i->a[1], i->a[2]);
		vec3 w = Eigen::Vector3d(i->w[0], i->w[1], i->w[2]);
		
		if(imu_initialization_measurements < imu_initialization_measurements_required) 
		{
			gyro_bias += w;
			init_gravity += a;
			imu_initialization_measurements++;
			//				std::cerr << "Gyro bias: " << gyro_bias.transpose() << ", Init Gravity: " << init_gravity.transpose() << std::endl;
		}
		else if (imu_initialization_measurements == imu_initialization_measurements_required) 
		{
			imu_initialization_measurements++;
			gyro_bias /= imu_initialization_measurements_required;
			init_gravity /= imu_initialization_measurements_required;
			std::cerr << "Gyro bias: " << gyro_bias.transpose() << ", Init Gravity: " << init_gravity.transpose() << std::endl;
			opti_.init(
					Eigen::Vector3d(0,0,pc.start_height),
					init_gravity);
		}
		else 
		{
			w -= gyro_bias;
			opti_.add_imu_measurement(a, w,
					1.0/pc.imu_sample_freq);
		}
	}
	
	return (imu_initialization_measurements >= imu_initialization_measurements_required);
}

void PluginOptimizer::timing(const char* what, timeutils::timer &t) const
{
	if(pc.optimizer_verbosity >= 1)
	{
		fprintf(stderr, "OPTIMIZER: %s %lld\n", what, t.elapsed());
		t.restart();
	}
}

Result PluginOptimizer::process(Frame * data)
{
	// Only camera 0 should do some stuff
	if (data->cam_id != 0)
		return Ok;
	fprintf(stderr, "PluginOptimizer::process called\n");
	timeutils::timer t;
	
	bool imu_initialized = process_imu(data->imu_measurements);
	timing("imu", t);
	
	// Get other Frame from other camera
	Frame *other_data = other_buffer->getPointerAt(other_buffer->getCurrentWriteIdx());

	if(! imu_initialized){
		data->world2cam = (vslam::pose_as_matrix(vslam::optimizer::orientation_from_gravity(init_gravity), Eigen::Vector3d(0,0,pc.start_height)) * Transform::left2imu).inverse();
//		data->world2cam = (opti_.current_imu2world() * Transform::left2imu).inverse();
		other_data->world2cam = Transform::cam.get_left2right() * data->world2cam;
		return Ok;
	}
	
	data->map_matches.clear();
	data->mono_matches.clear();
	other_data->mono_matches.clear();
	
	// next frame always starts after imu_measurements:
	opti_.next_frame();
	timing("next_frame", t);
	
	// update world2cam transformations for feature matching
	data->world2cam = (opti_.current_imu2world() * Transform::left2imu).inverse();
	other_data->world2cam = Transform::cam.get_left2right() * data->world2cam;
	
	
	// look for (stereo) features in map
	global_map_.nearest_neighbors(data->stereo_features,
			boost::bind(&self::process_known_feature, this, data, _1, _2, _3),
			boost::bind(&self::process_unknown_feature, this, data, _1, _2, _3));
	timing("stereo_features", t);
	
	if(data->map_matches.size() < pc.feature_association_limit)
	{
		// look for mono features (only if not enough stereo matches)
		global_map_.nearest_neighbors(data->mono_features, 
				boost::bind(&self::process_mono_feature, this, data, _1, _2, _3));
		timing("mono_features", t);
	}
	// do the actual optimization:
	opti_.step(pc.optimization_steps);
	timing("step", t);
	
	
	// update world2cam transformations for visualization
	data->world2cam = (opti_.current_imu2world() * Transform::left2imu).inverse();
	other_data->world2cam = Transform::cam.get_left2right() * data->world2cam;
	
	return Ok;  
}


typedef SLOM::internal::IMeasurement_Holder IMeasurement_Holder;
typedef IMeasurement_Holder::holder<vslam::optimizer::stereo_measurement> s_type;
typedef std::deque<const s_type*> st_measurements; 

typedef vslam::optimizer::stereo_measurement::id stereo_id;

struct extract_measurements
{
	st_measurements result;
	void operator()(const IMeasurement_Holder* m)
	{
		const s_type * s_m = dynamic_cast<const s_type*>(m);
		if(s_m)
		{
			Eigen::Vector4d res;
			s_m->eval(res.data(), false);
			if(res.start<2>().squaredNorm() <= 1)
			{
//				std::cout << res.start<2>().squaredNorm() << " ";
				result.push_back(s_m);
//				std::cout << result.size() << " ";
			}
		}
	}
};

typedef std::multimap<vslam::landmark_id, std::pair<int, Eigen::Matrix<double, 2, 1, Eigen::DontAlign> > > lm_collector;


bool PluginOptimizer::event(SDL_Event &event)
{
    bool event_handled = false;
    
    if(event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_l)
    {
    	// poses are imu2world, but we need left2world = imu2world * left2imu 
    	const vslam::optimizer::pose_list &poses = opti_.get_poses();
    	std::ofstream pose_file((pc.dump_dir + "/cams.txt").c_str());
    	Eigen::Quaterniond l2i_q(Transform::left2imu.get_rotation());
    	
    	lm_collector landmarks;
    	
    	int last_pose = 0; // number of last pose with enough measurements
    	for(size_t i = 0; i<poses.size(); ++i)
    	{
    		extract_measurements collector;
    		SLOM::SparseFunction::traverse_measurements(poses[i], collector);
    		if(collector.result.size() < 6)
    		{
//    			std::cerr << "Only " << collector.result.size() << " measurements\n";
    			continue;
    		}

    		for(st_measurements::const_iterator it = collector.result.begin(); it!=collector.result.end(); ++it)
    		{
    			const vslam::optimizer::stereo_measurement & sm = (*it)->m_;
//    			std::cerr << sm.lm.ptr << " ";
    			landmarks.insert(std::make_pair(sm.lm, std::make_pair(last_pose, sm.coord_left)));
    		}
    		
    		Eigen::Quaterniond orient = poses[i]->orient * l2i_q;
    		Eigen::Vector3d pos = poses[i]->orient * Transform::left2imu.get_translation() + poses[i]->pos;
    		
    		pose_file << orient.w() << " " << orient.vec().transpose() << " " << pos.transpose() << std::endl;
    		++last_pose;
    	}
    	
    	
    	pose_file.close();
    	int lm_count = 0, meas_count = 0;
    	std::ofstream lm_file((pc.dump_dir + "/pts.txt").c_str());
    	for(lm_collector::const_iterator it = landmarks.begin(); it!=landmarks.end(); /* */)
    	{
    		vslam::landmark_id id = it->first;
    		int count = landmarks.count(id);
    		if(count >= 2) // at least 2 measurements per landmark 
    		{
    			++lm_count; meas_count += count;
    			lm_file << it->first->transpose() << " " << count;
    			for(int c = 0; c<count; ++c)
    			{
    				lm_file << " " << it->second.first << " " << it->second.second.transpose();
    				++it;
    			}
    			lm_file << std::endl;
    		}
    		else
    		{
    			advance(it, count);
    		}
    	}
    	
    	lm_file.close();
    	std::cout << "PluginOptimizer: wrote " << last_pose << " poses, " << lm_count << " landmarks with " << meas_count << " Associations\n"; 
    	event_handled = true;
    } 
    else if(event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_j)
    {
    	// dump jacobian
    	ofstream out("jacobian.dat");
    	opti_.dump_jacobian(out);
    }
    
    return event_handled;
}
