/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *          Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "plugin_visualisation.h"
#include <iostream>
#include <iomanip>
#include <limits>
#include <SDL_gfxPrimitives.h>
#include <cmath>
#include <sstream>
#include <cmath>
#include <highgui.h>
#include "transformations.h"

#include <vslam/map/map.hpp>
#include <vslam/camera/camera.hpp>

#include "feature_extraction_config.hpp"

PluginVisualisation::PluginVisualisation(std::string _name, Buffer * _buffer, Buffer * _other_buffer,
										 const vslam::map::map &global_map,
										 int _width, int _height, int _dsf, const program_config &pc, VisualisationWindow *vw, PluginVisualisation *other_visualisation)
    : ProcessingPlugin(_name, _buffer),
	  other_buffer(_other_buffer),
	  global_map_(global_map),
	  do_draw_video_image(true),
	  do_write_rendered_image(false && "DEPRECATED"),
	  dump_dir(pc.dump_dir),
	  dsf(_dsf),
	  vis_dsf(static_cast<double>(_dsf)),
	  left(0), upper(0), drag(false),
	  width(_width), height(_height),
	  pc(pc),
	  vw_(vw),
	  other_visualisation_(other_visualisation)
{
  Uint32 rmask, gmask, bmask, amask;

#if SDL_BYTEORDER == SDL_BIG_ENDIAN
  rmask = 0xff000000;
  gmask = 0x00ff0000;
  bmask = 0x0000ff00;
  amask = 0x000000ff;
#else
  rmask = 0x000000ff;
  gmask = 0x0000ff00;
  bmask = 0x00ff0000;
  amask = 0xff000000;
#endif
  
  visualisation = SDL_CreateRGBSurface(SDL_SRCCOLORKEY|SDL_RLEACCEL, width/dsf, height/dsf, 32, rmask, gmask, bmask, amask);
}

PluginVisualisation::~PluginVisualisation()
{
  SDL_FreeSurface(visualisation);
}

void PluginVisualisation::draw_cross(const cv::Point2f &p_, int r, int g, int b, int a)
{
	cv::Point2f p(p_.x/vis_dsf, p_.y/vis_dsf);
	
	float cL = 1+5.0/vis_dsf;
	
	// rounded coordinates of cross corners:
	int lef = p.x - cL + 0.5;
	int rig = p.x + cL + 0.5;
	int bot = p.y - cL + 0.5;
	int top = p.y + cL + 0.5;
	
	lineRGBA(visualisation, lef, top, rig, bot, r, g, b, a);
	lineRGBA(visualisation, lef, bot, rig, top, r, g, b, a);
	
	if(pc.thick_crosses)
	{
		lineRGBA(visualisation, lef+1, bot, rig, top-1, r, g, b, a);
		lineRGBA(visualisation, lef, bot+1, rig-1, top, r, g, b, a);
		lineRGBA(visualisation, lef+1, top, rig, bot+1, r, g, b, a);
		lineRGBA(visualisation, lef, top-1, rig-1, bot, r, g, b, a);
	}
}

void PluginVisualisation::draw_plus(const cv::Point2f &p_, int r, int g, int b, int a)
{
	cv::Point2f p(p_.x/vis_dsf, p_.y/vis_dsf);
	
	float cL = 1+5/vis_dsf;
	
	int lef = p.x - cL + 0.0;
	int rig = p.x + cL + 1.0;
	int bot = p.y - cL + 0.0;
	int top = p.y + cL + 1.0;
	
	lineRGBA(visualisation, lef, p.y, rig, p.y, r, g, b, a);
	lineRGBA(visualisation, p.x, bot, p.x, top, r, g, b, a);
	if(pc.thick_crosses)
	{
		lineRGBA(visualisation, lef, p.y+1, rig, p.y+1, r, g, b, a);
		lineRGBA(visualisation, p.x+1, bot, p.x+1, top, r, g, b, a);
	}
}

void PluginVisualisation::draw_line(const cv::Point &p1_, const cv::Point &p2_,
									int r, int g, int b, int a)
{
	using namespace cv;

	Point p1(p1_.x/vis_dsf, p1_.y/vis_dsf);
	Point p2(p2_.x/vis_dsf, p2_.y/vis_dsf);
	
	lineRGBA(visualisation,  
			 p1.x, 
			 p1.y, 
			 p2.x, 
			 p2.y, r, g, b, a);
}

void PluginVisualisation::draw_text(int x, int y, const std::string &text, int r, int g, int b, int a)
{
	stringRGBA(visualisation, x, y, text.c_str(), r, g, b, a);
}

void PluginVisualisation::draw_filled_circle(const cv::Point &p_, int radius, int r, int g, int b, int a)
{
	using namespace cv;

	Point p(p_.x/vis_dsf, p_.y/vis_dsf);
	filledCircleRGBA(visualisation, p.x, p.y, radius, r, g, b, a);
}


void draw_video_image_subpixel(unsigned char* src, unsigned char* dst, double dsf, int dst_width, int dst_height, unsigned int width_step)
{
  // Destination ptrs
  int* p1_0 = (int*) dst;
  int* p1 = p1_0;
  int* p1_end = p1_0 + dst_width * dst_height;
  
  // Source ptrs
  unsigned char* p2_0 = (unsigned char*) src;
  unsigned char* p2 = p2_0;
  double acc_dsf = dsf;
  double offset = 0.0;

  while(p1 != p1_end)
  {    
    *(p1++) = (int) *p2 + ((int)*p2 << 8) + ((int)*p2 << 16) + ((int)0xff << 24);
    offset += dsf;
    p2 = p2_0 + (int) offset;
    
    if ((offset - floor(offset)) == 0 && (p2 - p2_0) % (int)((double)dst_width * dsf) == 0)
    {
      if(acc_dsf != 1.0)
      {
        offset += -(static_cast<double>(dst_width) * dsf);        
        acc_dsf += dsf;        
      }
      else
      {
        offset += (static_cast<double>(width_step) - (static_cast<double>(dst_width) * dsf));
        acc_dsf = dsf;
      }

      p2 = p2_0 + (int) offset;    
    }    
  }
}

void draw_video_image(unsigned char* src, unsigned char* dst, int dsf, int dst_width, int dst_height, unsigned int width_step)
{
  // Destination ptrs
  int* p1_0 = (int*) dst;
  int* p1 = p1_0;
  int* p1_end = p1_0 + dst_width * dst_height;
  
  // Source ptrs
  unsigned char* p2_0 = (unsigned char*) src;
  unsigned char* p2 = p2_0;

  while(p1 != p1_end)
  {    
    *(p1++) = (int) *p2 + ((int)*p2 << 8) + ((int)*p2 << 16) + ((int)0xff << 24);
    p2 +=dsf;
    
    if ((p2 - p2_0) % (dst_width * dsf) == 0)
      p2 += (dsf-1) * width_step + width_step - (dst_width * dsf);
  }
}

int PluginVisualisation::calc_grid_line( vslam::camera::image_point coords[], // output array, minimal size [grid_steps+1][2]
		const vslam::camera &cam, const vect &p0, const vect &dir, int grid_steps)
{
	double lambdaMin = 0 - pc.grid_linewidth, lambdaMax = grid_steps * pc.grid_stepsize + pc.grid_linewidth;
	cam.sensorVisibleIntervalOfLine(p0, dir, lambdaMin, lambdaMax, true, true);
	if(lambdaMin > lambdaMax)
		return 0; // no points visible
	// get indexes of first and last grid point:
	int y_start = std::max(static_cast<int>(std::ceil(lambdaMin/pc.grid_stepsize)), 1);
	int y_end   = std::min(static_cast<int>(std::floor(lambdaMax/pc.grid_stepsize)), grid_steps - 1);
	// calculate start and end-point:
	int ci = 0; // current coordinates index
	cam.sensor2ImageClipped(p0 + lambdaMin * dir, coords[ci++]);
	for(int i=y_start; i<=y_end; ++i)
	{
		cam.sensor2ImageClipped(p0 + (i*pc.grid_stepsize)*dir, coords[ci++]);
	}
	cam.sensor2ImageClipped(p0 + lambdaMax * dir, coords[ci++]);
	return ci;
}

void PluginVisualisation::draw_grid_line(const vslam::camera &cam, const vect &p0, const vect &p1, const vect &dir, int grid_steps, bool zero_line)
{
	vslam::camera::image_point coords[grid_steps + 1]; // calculated coordinates, column major
	Sint16 px_coords[2][2*grid_steps + 2]; // input array for filledPolygonRGBA, row major

	int steps0 = calc_grid_line(coords, cam, p0, dir, grid_steps);
	if(steps0 == 0)
		return;
	for(int i=0; i<steps0; ++i)
	{
		px_coords[0][i] = coords[i].x / vis_dsf;
		px_coords[1][i] = coords[i].y / vis_dsf;
	}

	int steps1 = calc_grid_line(coords, cam, p1, dir, grid_steps);
	for(int i=0; i<steps1; ++i)
	{
		int idx = steps1 + steps0 - 1 - i;
		px_coords[0][idx] = coords[i].x / vis_dsf;
		px_coords[1][idx] = coords[i].y / vis_dsf;
	}
	int r=255, g=0, b=0;
	if(zero_line)
		g=128;

	filledPolygonRGBA(visualisation, &px_coords[0][0], &px_coords[1][0], steps0 + steps1, r, g, b, 128);

}

void PluginVisualisation::draw_grid(const vslam::camera::Transform &world2cam, const vslam::camera &cam)
{
	for(int x=0; x<=pc.grid_x_steps; ++x)
	{
		double x1 = pc.grid_x_min + x*pc.grid_stepsize;
		vect p0 = world2cam * vect(x1-pc.grid_linewidth, pc.grid_y_min, 0.0);
		vect p1 = world2cam * vect(x1+pc.grid_linewidth, pc.grid_y_min, 0.0);
		vect dir = world2cam.get_rotation() * vect(0,1,0);
		
		draw_grid_line(cam, p0, p1, dir, pc.grid_y_steps, (fabs(x1)< pc.grid_linewidth));
	}
	for(int y=0; y<=pc.grid_y_steps; ++y)
	{
		double y1 = pc.grid_y_min + y*pc.grid_stepsize;
		vect p0 = world2cam * vect(pc.grid_x_min, y1-pc.grid_linewidth, 0.0);
		vect p1 = world2cam * vect(pc.grid_x_min, y1+pc.grid_linewidth, 0.0);
		vect dir = world2cam.get_rotation() * vect(1,0,0);
		
		draw_grid_line(cam, p0, p1, dir, pc.grid_x_steps, (fabs(y1)< pc.grid_linewidth));
	}
}

template<class Feature>
unsigned 
PluginVisualisation::draw_landmark_matches(const std::vector<std::pair<const Feature*, const vslam::map::landmark*> > &map_matches, 
		cv::KeyPoint Feature::* l_r, const vslam::camera &cam, const vslam::camera::Transform &world2cam,
		int r, int g, int b, int a)
{
	unsigned inliers = 0;
	for (typename std::vector<std::pair<const Feature*, const vslam::map::landmark*> >::const_iterator i = map_matches.begin();
			i != map_matches.end(); ++ i)
	{
		const Feature &feature = *i->first;
		const vslam::map::landmark &landmark = *i->second;
		
		vect lm_cam = world2cam.apply(*landmark.position);
		vslam::camera::image_point lm_im;
		bool visible = cam.sensor2ImageClipped(lm_cam, lm_im);
		if(visible) {
			const cv::Point2f &match_pt = (feature.*l_r).pt;
			draw_line(match_pt, lm_im, r, g, b, a);
			draw_cross(match_pt, r, g, b, a);
			
			Eigen::Vector2f a(&match_pt.x);
			Eigen::Vector2f b(&lm_im.x);
			
			if ((b-a).norm() < 10)
			{
				++inliers;
			}
		}	  
	}
	
	return inliers;
}

Result PluginVisualisation::process(Frame *)
{
	Frame * data = buffer->getPointerAt(buffer->getCurrentWriteIdx());
	
	std::cerr << "processing time " << (data->cam_id == 0 ? "left" : "right") << ": " << data->t.elapsed() << std::endl;

  if(do_draw_video_image)
  {
    if (left < 0)
      left = 0;    

    if (left > data->video.width() - visualisation->w * vis_dsf)
      left = data->video.width() - visualisation->w * vis_dsf;

    if (upper < 0)
      upper = 0;

    if (upper > data->video.height() - visualisation->h * vis_dsf)
      upper = data->video.height() - visualisation->h * vis_dsf;

    if (vis_dsf < 1.0)
      draw_video_image_subpixel(data->video.imageData() + upper * width + left, (unsigned char*) visualisation->pixels, vis_dsf, visualisation->w, visualisation->h, data->video.width());
    else
      draw_video_image(data->video.imageData() + upper * width + left, (unsigned char*) visualisation->pixels, static_cast<int>(vis_dsf), visualisation->w, visualisation->h, data->video.width());

  }

  const vslam::camera &cam = ::Transform::cam.get_cam(data->cam_id == 0);
  draw_grid(data->world2cam, cam);

  for (vslam::feature_points::const_iterator i = data->features.begin();
	   i != data->features.end(); ++i)
  {
	  //std::cout << "kp response: " << i->response << std::endl;
	  draw_cross(i->kp().pt, 150, 150, 150, 128);
  }

  //std::cerr << "vis cam: " << data->cam_id << std::endl;
  Frame *other_data = other_buffer->getPointerAt(other_buffer->getCurrentWriteIdx());
  
  Frame *d = data->cam_id == 0 ? data : other_data;
  cv::KeyPoint vslam::matcher::stereo_match::* l_r = data->cam_id == 0 ? &vslam::matcher::stereo_match::left : & vslam::matcher::stereo_match::right; 
  for (std::vector<vslam::matcher::stereo_match>::const_iterator i = d->stereo_features.begin();
	   i != d->stereo_features.end(); ++i)
  {
	  draw_cross((*i.*l_r).pt, 0, 255, 0, 255);
  }

  for (std::vector<vslam::map::feature_map::inter_frame_match>::const_iterator i = d->inter_frame_matches.begin();
	   i != d->inter_frame_matches.end(); ++i)
  {
	  draw_line((i->left->*l_r).pt, (i->right->*l_r).pt, 255, 255, 0, 255);
	  draw_plus((i->left->*l_r).pt, 255, 255, 0, 255);
  }
  data->inter_frame_matches.clear();

  // draw all landmarks currently in the map
  Transform &world2cam = data->world2cam;
  const vslam::map::map::landmarks_list &landmarks = global_map_.get_landmarks();
  for(vslam::map::map::landmarks_list::const_iterator it = landmarks.begin(); it != landmarks.end(); ++it) {
	  vect lm_cam = world2cam.apply(*it->position);
	  vslam::camera::image_point lm_im;
	  bool visible = cam.sensor2ImageClipped(lm_cam, lm_im);
	  if(visible) {
			draw_plus(lm_im, 0, 255, 255, 255);
	  }
  }

  
  // draw matches of features to landmarks
  const std::vector<Frame::map_match> &map_matches = d->map_matches;
  unsigned inliers = draw_landmark_matches(map_matches, l_r, cam, world2cam, 255, 255, 0, 255);
  inliers += draw_landmark_matches(data->mono_matches, &vslam::matcher::mono_feature::keypoint, cam, world2cam, 255, 128, 0, 255);
  
  Eigen::Matrix<int,2,1> pos(5,5);
  
  int height = visualisation->h;
  int width = visualisation->w;
  
  if(pc.visualize_verbosity > 0)
  {
    pos.y() = height - 15;
    int baselineskip = 12;
    {
      std::stringstream s;
      s << "frame:" << std::setw(4) << data->number;
      draw_text(width - 8*s.str().length() - 5, pos.y() , s.str(), 255, 255, 255, 255);
    }
    if(pc.visualize_verbosity > 1) {
    {
      std::stringstream s;
      s << "+ reprojected map features: " << std::setw(4) <<  landmarks.size();
      draw_text(pos.x(), pos.y() , s.str(), 0, 255, 255, 255);
      pos.y() -= baselineskip;
    }
    {
      std::stringstream s;
      s << "x fused with map (monocular):" << std::setw(3) <<  data->mono_matches.size();
      draw_text(pos.x(), pos.y() , s.str(), 255, 128, 0, 255);
      pos.y() -= baselineskip;
    }
    {
      std::stringstream s;
      s << "x fused with map (stereo):  " << std::setw(4) <<  d->map_matches.size();
      draw_text(pos.x(), pos.y() , s.str(), 255, 255, 0, 255);
      pos.y() -= baselineskip;
    }
    {
      std::stringstream s;
      s << "x stereo matched:           " << std::setw(4) <<  d->stereo_features.size();
      draw_text(pos.x(), pos.y() , s.str(), 0, 255, 0, 255);
      pos.y() -= baselineskip;
    }
    }
    if(pc.visualize_verbosity > 2){
      pos.x() = width - 5;
      pos.y() += baselineskip;
      std::ostringstream s;
      s.precision(4);
      s << "current position:";
      
      vect position = data->world2cam.inverse().get_translation();
      for(int i=0; i<3; ++i){
        s << std::setw(8) << std::fixed << position(i);
        draw_text(pos.x() - 8*s.str().length(), pos.y() + i * baselineskip, s.str(), 255, 255, 255, 255);
        s.str("");
      }
    }
  }
  
  if (pc.dump_images & (1 << data->cam_id) )
  {
      std::string prefix;
      if (data->cam_id == 0)
      {
          prefix = "l";
      }
      else
      {
          prefix = "r";
      }
      
//      uint64_t os_ts_in_mu_s = data->os_ts.tv_sec * 1e6 + data->os_ts.tv_nsec / 1000;
//      uint64_t cam_ts_in_mu_s = data->cam_ts.tv_sec * 1e6 + data->cam_ts.tv_nsec / 1000;
      
      std::stringstream fname;
      //fname << prefix << "-"<< os_ts_in_mu_s << "-"<< cam_ts_in_mu_s << ".bmp";
      fname << dump_dir << "/" << prefix << std::setfill('0') << std::setw(4) << data->number << ".bmp";
      //cvSaveImage(fname.str().c_str(), visualisation->pixels);
      SDL_SaveBMP(visualisation, fname.str().c_str());
      counter++;
  }

  
  if (data->cam_id != 0)
  {
	SDL_Surface *left = 0;
    if (other_visualisation_)
	  left = other_visualisation_->visualisation;	
    vw_->blit(left, visualisation);
  }

  return Ok;  
}

bool PluginVisualisation::event(SDL_Event &event)
{
  bool event_handled = false;
  switch (event.type)
  {    

  case SDL_MOUSEBUTTONDOWN: 
  {
      
    if (event.button.button == SDL_BUTTON_WHEELUP)
    {       
      // Check how many subpixel scales fit into our window
      double max_subpixel_scale = 1.0;
      while(floor(static_cast<double>(width/dsf) * (max_subpixel_scale / 2.0)) / (max_subpixel_scale / 2.0) == static_cast<double>(width/dsf))
        max_subpixel_scale /= 2.0;        

      // Zoom in 
      if (vis_dsf > max_subpixel_scale)
      {          
        vis_dsf /= 2.0;        

        // Add half frame in current sample ratio
        left += width / 2 / (dsf/vis_dsf); 
        upper += height / 2 / (dsf/vis_dsf);
      }
        
      event_handled = true;
    }
    else if (event.button.button == SDL_BUTTON_WHEELDOWN)
    {    
      // Zoom out
      if (vis_dsf < dsf)
      {            
        vis_dsf *= 2.0;

        // Add half frame in previous sample ratio
        left -= width / 2 / (dsf/(vis_dsf / 2.0));
        upper -= height / 2 / (dsf/(vis_dsf / 2.0));
      }          
             
      event_handled = true;
    }
    else if (event.button.button == SDL_BUTTON_MIDDLE)
    {
      int global_mouse_x = (event.button.x % visualisation->w * dsf - width/2) / (dsf/vis_dsf);
      int global_mouse_y = (event.button.y * dsf - height/2) / (dsf/vis_dsf);      
          
      left += global_mouse_x;
      upper += global_mouse_y;

      event_handled = true;
    }
    break;
  }  
  default: ;
    break;
  }
    
  return event_handled;
}
