/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *          Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "processing_thread.h"
#include <iostream>
#include <time.h>
#include "image_capture_thread.h"
#include <sys/time.h>
#include <utils/sseHelpers.h>

boost::barrier ProcessingThread::bar(2);
int ProcessingThread::frame_count[2];

ProcessingThread::ProcessingThread(int cam_id, Generic_Camera* _cam, images_rb_t &_img_rb,
                                   double _camera_lag, bool _image_rotated, std::ostream &timing_file)
	: stack(0),
	  rb(0),
	  _kill(false),
	  camId(cam_id),
	  cam(_cam),
	  img_rb(_img_rb),
	  camera_lag(_camera_lag),
	  image_rotated(_image_rotated),
	  timing_file(timing_file)
{
}

void ProcessingThread::setProcessingStack(ProcessingStack * _stack) {
  stack_mutex.lock();
  stack=_stack;
  stack->set_timingfile(&timing_file);
  stack_mutex.unlock();
}

ProcessingThread::~ProcessingThread()
{
}

void ProcessingThread::setBuffer(Buffer * _rb) {
  rb=_rb;
}

Buffer * ProcessingThread::getBuffer() const {
  return rb;
}

ProcessingStack * ProcessingThread::getStack() const {
  return stack;
}

void ProcessingThread::kill() {
 _kill=true; 
 while(is_running()) {
   usleep(100);
 }
}

bool ProcessingThread::init() {
  capture_mutex.lock();
  capture_mutex.unlock();
  timing_file << "#";
  for(std::vector<ProcessingPlugin*>::const_iterator it = stack->stack.begin(); it != stack->stack.end(); ++it)
  {
	  timing_file << " " << (*it)->getName();
  }
  timing_file << std::endl;
  return true;
}

bool ProcessingThread::stop() {
//  capture_mutex.lock();
//  Thread::stop();
//  capture_mutex.unlock();
	boost::mutex::scoped_lock lock(stack_executing_mutex);
	std::cerr << "stopping processing thread " << this << std::endl;
	pthread_cancel(handle);
	running = false;
  join();
  return true;
}

void ProcessingThread::join() {
	pthread_join(handle, 0);
}

static
void rotate_and_cpy (const unsigned char* src, int w, unsigned char* ds)
{
  assert(w % 16 == 0);
    
  SSEDQWord mask;
  for (int i=0; i < 16; i++)
  {
      mask.ub[i] = 15 - i;
  }
  
   asm (
     "movl %0, %%esi     \n\t"
     "movl %1, %%edi     \n\t"

     "movl %%esi, %%edx  \n\t"
     "subl $16, %%edx    \n\t"

     "addl %2, %%esi     \n\t"
     "subl $16, %%esi    \n\t"

     "movdqa %3, %%xmm1  \n\t"

 "1:  movdqu (%%esi), %%xmm0  \n\t"
     "pshufb %%xmm1, %%xmm0 \n\t"
     "movdqa %%xmm0, (%%edi)  \n\t"
     "subl $16, %%esi         \n\t"
     "addl $16, %%edi          \n\t"
     "cmp %%esi, %%edx        \n\t"
     "jne 1b                  \n\t"    
    : : "m" (src), // %0
        "m" (ds), // %1
        "m" (w), // %2
        "m" (mask) // %3
     : "%xmm0", "%xmm1", "%esi", "%edi", "%edx", "memory"
     );  
}

bool ProcessingThread::keep_running() const
{
	return is_running();
}

bool ProcessingThread::synchronize()
{
	bar.wait();
	return true;
}

void ProcessingThread::run() {

	Image myImage;
	t_0_found = false;

	while(keep_running())
	{
		if (rb!=0)
		{
			int idx=rb->getCurrentWriteIdx();
			Frame * d=rb->getPointerAt(idx);
			
			//capture_mutex.lock();
			if (keep_running())
			{
				img_rb.pop(myImage);
				frame_count[camId] = myImage.frame_count;
				if (!synchronize())
					return;
        
				while(frame_count[camId] < frame_count[1-camId]
					  && is_running())
				{
					img_rb.pop(myImage);
            
					frame_count[camId] = myImage.frame_count;
				}

				if (d->video.empty()
					|| (unsigned int)(d->video.width()) != myImage.width
					|| (unsigned int)(d->video.height()) != myImage.height)
				{
					d->video = IplImageWrapper(myImage.width, myImage.height, IPL_DEPTH_8U, 1);
				}
			
        
				d->cam_id=camId;
				d->cam_ts = myImage.timestamp;
				d->is_log = myImage.is_log;          
				d->os_ts = myImage.os_timestamp;

				double os_ts_without_lag = (d->os_ts.tv_sec + d->os_ts.tv_nsec / 1e9) - camera_lag;
				d->physical_event_ts.tv_sec = static_cast<time_t>(os_ts_without_lag);
				d->physical_event_ts.tv_nsec = (os_ts_without_lag - static_cast<double>(d->physical_event_ts.tv_sec)) * 1e9; 

				d->number = myImage.frame_count;

				if (!t_0_found)
				{
					t_0 = d->os_ts;
					t_0_found = true;
				}
				d->delta_t = (d->os_ts.tv_sec + d->os_ts.tv_nsec / 1e9) - (t_0.tv_sec + t_0.tv_nsec / 1e9);	


				if (!image_rotated)
					memcpy(d->video.imageData(), myImage.data, myImage.width * myImage.height);	
				else
					rotate_and_cpy(myImage.data, myImage.width * myImage.height, d->video.imageData());

				if (stack!=0)
				{
					boost::mutex::scoped_lock lock(stack_executing_mutex);
			
					stack->process(d);
				}
			}
			//capture_mutex.unlock();
		}
	}
}
