/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *          Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "program_config.h"
#include <cstring>
#include <iostream>
#include <assert.h>
#include <cstdlib>
#include <fstream>

void separate_arguments(int ac, char** av, int& ard_argc, char** ard_argv, int& po_argc, char** po_argv)
{
  // Find ard's argument in argument list
  ard_argc = 1;
  ard_argv[0] = strdup(av[0]);

  for (int i=0; i < ac; i++)
  {
    std::string av_str(av[i]);
    int num_params(0);
    if (av_str.substr(1,3).compare("ard") == 0)
    {
      std::string av_substr(av_str.substr(1,7));

      if (av_substr.compare("ardishm") == 0 || av_substr.compare("ardoshm") == 0 || av_substr.compare("ardoudp") == 0)
      {
        num_params = 2;
      }
      else if(av_substr.compare("ardiudp") == 0) 
      {
        num_params = 1;
      }
      else
      {
        std::cerr << "Found unknown ard program option. Exiting." << std::endl;
        exit(1);
      }      
      
      int j(0);
      for (; j <= num_params && i+j < ac; j++)
      {  
        ard_argv[ard_argc] = strdup(av[i+j]); 
        strcpy(av[i+j], "");
        ard_argc++;
      }      
      
      i += j-1;
    }    
  }

  // Fill the program options buffer with non ard arguments
  for (int i=0; i < ac; i++)
  {
    if (strcmp(av[i], "") != 0)
    {
      po_argv[po_argc] = strdup(av[i]);
      po_argc++;
    }    
  }

  // Test
  assert(po_argc + ard_argc == ac + 1);
}

void do_program_options(int po_argc, char** po_argv, po::variables_map &vm, program_config &pc)
{

  try {
    
    // Declare a group of options that will be 
    // allowed only on command line
    po::options_description generic("Generic options");
    generic.add_options()
      ("version,v", "print version string")
      ("help", "produce help message")    
      ;

    po::options_description gui("Gui options");
    gui.add_options()
      ("nox,x", po::value<bool>(&pc.nox)->default_value(false)->implicit_value(true), "disable graphical output")
      ;
	
    po::options_description camera("Camera options");
    camera.add_options()
      ("left_cam_serial", po::value<std::string>(&pc.left_cam_serial)->default_value("02-2165A-06330"), "Serial no of left cam")
      ("right_cam_serial", po::value<std::string>(&pc.right_cam_serial)->default_value("02-2165A-06327"), "Serial no of right cam")
      ("image_width", po::value<int>(&pc.image_width)->default_value(1024), "Image width in px")
      ("image_height", po::value<int>(&pc.image_height)->default_value(768), "Image height in px")
      ("region_x", po::value<int>(&pc.region_x)->default_value(0), "X offset of ROI")
      ("region_y", po::value<int>(&pc.region_y)->default_value(0), "Y offset of ROI")
      ("exposure", po::value<int>(&pc.exposure)->default_value(1500), "Exposure time in us")
      ("gain", po::value<int>(&pc.gain)->default_value(11), "Gain")
      ("frame_rate", po::value<double>(&pc.frame_rate)->default_value(25.0), "Frame-rate")
      ("rotate_left_cam", po::value<bool>(&pc.rotate_left_cam)->default_value(false)->implicit_value(true), "Rotate left cam 180 deg")
      ("rotate_right_cam", po::value<bool>(&pc.rotate_right_cam)->default_value(false)->implicit_value(true), "Rotate right cam 180 deg")
      ("dsf", po::value<int>(&pc.dsf)->default_value(2), "Visualisation down scale factor")
      ;
  

    po::options_description imu("IMU options");
    imu.add_options()
        ("imu_dev", po::value<std::string>(&pc.imu_dev)->default_value("/dev/ttyUSB0"), "Serial device (e.g. /dev/ttyUSB0)")
        ("imu_sample_freq", po::value<int>(&pc.imu_sample_freq)->default_value(512), "IMU sample frequency")
        ;  
    po::options_description timing_options("Timing options");
    timing_options.add_options()
      ("lag_imu", po::value<double>(&pc.lag_imu)->default_value(0.0), "Lag of incoming IMU wrt to physical event in seconds")
      ("lag_cam", po::value<double>(&pc.lag_cam)->default_value(0.015), "Lag of incoming camera images wrt to physical event in seconds")
      ;    

    po::options_description optimizer_options("Optimizer options");
    optimizer_options.add_options()
      ("imu_initialization_frames", po::value(&pc.imu_initialization_frames)->default_value(512), "Number of IMU measurements required for initialization")
      ("start_height", po::value(&pc.start_height)->default_value(0.1), "Height of first pose in meters")
      ("optimization_steps", po::value(&pc.optimization_steps)->default_value(1), "Number of Gauss-Newton iterations per frame")
      ("association_threshold", po::value(&pc.association_threshold)->default_value(50), "Threshold in pixels until which features are seen as associated (before optimization)")
      ("optimizer_verbosity", po::value(&pc.optimizer_verbosity)->default_value(1), "Optimizer verbosity. 0=none, 1=little, 2=everything")
      ("max_landmark_distance", po::value(&pc.max_landmark_distance)->default_value(3), "maximal distance of landmark to left camera in order to registered")
      ("min_relative_landmark_distance", po::value(&pc.min_relative_landmark_distance)->default_value(.05), "minimal relative distance between nearest landmark and distance to camera in order to be registered")
      ("velocity_sigma", po::value(&pc.velocity_sigma)->default_value(-1), "sigma of usual velocity, if negative value is ignored")
      ;

    po::options_description dense_stereo_options("Dense stereo options");
    dense_stereo_options.add_options()
      ("min_stereo_range", po::value(&pc.min_stereo_range)->default_value(0.05), "Minimum range [m] for dense stereo data to be added to the voxel map")
      ("max_stereo_range", po::value(&pc.max_stereo_range)->default_value(0.6), "Maximum range [m] for dense stereo data to be added to the voxel map")
      ("min_disparity", po::value(&pc.min_disparity)->default_value(0), "Minimum disparity for block matcher (in pixel)")
      ("disparity_range", po::value(&pc.disparity_range)->default_value(80), "Disparity range (i.e. max_disparity - min_disparity) for block matcher (in pixel)")
      ("dump_disparities", po::value(&pc.dump_disparities)->default_value(false)->implicit_value(true), "Dump stereo disparity images")
      ;

    po::options_description calibration_options("Calibration options");
    dense_stereo_options.add_options()
      ("calib_plate_features_vertical", po::value(&pc.calib_plate_features_vertical)->default_value(7), "Number of vertical features on calibration plate.")
      ("calib_plate_features_horizontal", po::value(&pc.calib_plate_features_horizontal)->default_value(7), "Number of horizontal features on calibration plate.")
      ("dist_features", po::value(&pc.dist_features)->default_value(0.11125), "Distance between features (both dimensions) in [m]")
        ;

    po::options_description more_options("More options");
    more_options.add_options()
      ("save_dir", po::value<std::string>(&pc.save_dir)->default_value("/home/olli/saved/"), "Directory where to save log data")
      ("framebuffer_size", po::value<int>(&pc.framebuffer_size)->default_value(25), "Size of the framebuffer which stores past images")
      ("priority,p", po::value<int>(&pc.priority)->default_value(50), "Sets RT-prio")  
      ("log_normal,l", po::value<std::string>(&pc.log_dir), "Load logfile") 
      ("play_once", po::value<bool>(&pc.play_once)->default_value(false)->implicit_value(true), "Play logfile just once") 
      ("skip_frames", po::value<unsigned int>(&pc.skip_frames)->default_value(1), "Skip n frames") 
      ("dump_dir", po::value(&pc.dump_dir)->default_value("dump"), "directory where to dump result images and landmarks") 
      ("dump_images", po::value(&pc.dump_images)->default_value(0)->implicit_value(3), "dump images 0=off, 1=left, 2=right, 3=both") 
      ("dump_octomap", po::value(&pc.dump_octomap)->default_value(-1), "dump octomap .graph file at given frame (-1=off)") 
      ("dump_landmarks", po::value(&pc.dump_landmarks)->default_value(false)->implicit_value(true), "dump landmarks from map")
      ("dump_trajectory", po::value(&pc.dump_trajectory)->default_value(false)->implicit_value(true), "dump trajectory")
      ("compare_trajectory", po::value(&pc.compare_trajectory)->default_value(false)->implicit_value(true), "compare trajectory")
      ("dump_stereo_matches", po::value(&pc.dump_stereo_matches)->default_value(-1), "dump stereo matches at interval given in frames, or disable with -1")
      ("extract_features", po::value(&pc.extract_features)->default_value(true), "enable extraction of features (mandatory for SLAM!)")
      ("feature_limit", po::value(&pc.feature_limit)->default_value(200), "maximal number of extracted features")
      ("feature_association_limit", po::value(&pc.feature_association_limit)->default_value(50), "maximal number of associated features per frame")
// Visualization options:
      ("thick_crosses", po::value(&pc.thick_crosses)->default_value(false)->implicit_value(true), "use thick crosses in visualization")
      ("grid_stepsize", po::value(&pc.grid_stepsize)->default_value(0.1), "distance between grid lines")
      ("grid_linewidth", po::value(&pc.grid_linewidth)->default_value(0.005), "width of grid lines")
      ("grid_x_min", po::value(&pc.grid_x_min)->default_value(-0.1), "smallest grid x position (initial viewing direction)")
      ("grid_y_min", po::value(&pc.grid_y_min)->default_value(-2.0), "smallest grid y position (initially looking left")
      ("grid_x_steps", po::value(&pc.grid_x_steps)->default_value(12), "Number of grid lines in x direction")
      ("grid_y_steps", po::value(&pc.grid_y_steps)->default_value(30), "Number of grid lines in y direction")
      ("visualize_verbosity", po::value(&pc.visualize_verbosity)->default_value(2), "verbosity level for visualization 0: off, 1: just frame, 2: also landmarks and matches 3: all")
      ;

    po::options_description cmdline_options;
    cmdline_options.add(generic).add(gui).add(camera).add(imu).add(timing_options).add(optimizer_options).add(dense_stereo_options).add(more_options);
  
    po::options_description config_file_options;
    config_file_options.add(camera).add(imu).add(timing_options).add(more_options);

    po::options_description visible("Allowed options");
    visible.add(generic).add(gui).add(camera).add(imu).add(timing_options).add(optimizer_options).add(dense_stereo_options).add(more_options);
        
    store(po::command_line_parser(po_argc, po_argv).
          options(cmdline_options).run(), vm);
  
    notify(vm);

    // Look for a configuration file in the users home_directory    
    char *home_path  = getenv ("HOME");
    std::ifstream in;
    in.open(std::string(std::string(home_path) + "/.slam_vision/slam_vision.cfg").c_str());
    std::cerr << std::string(std::string(home_path) + "/.slam_vision/slam_vision.cfg") << std::endl;
    if (home_path != 0 && in.is_open())
    {
      store(parse_config_file(in, config_file_options), vm);
      std::cerr << "Loaded slamvision config from file" << std::endl;
    }

    notify(vm);

    if (vm.count("help")) {
      visible.print(std::cout);
      exit(1);
      //return;
    }

    if (vm.count("version")) {
      return;
    }
  }
  catch(std::exception& e)
  {
    std::cout << e.what() << "\nTry --help";
    exit(0);
  }
}



