/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *          Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "visualisation_window.h"
#include <sstream>
#include <iostream>

VisualisationWindow::VisualisationWindow(int _image_width, int _image_height, int _dsf)
	: display(0), image_width(_image_width), image_height(_image_height), dsf(_dsf), use2(0)
 {
  if ( display )
  {
    return;
  }

  if ( SDL_Init(SDL_INIT_VIDEO) < 0 )
  {
    exit(1);
  }
  
  SDL_VideoModeOK(2 * image_width / dsf, image_height / dsf, 24 /* 32 */, 0); // SDL_ANYFORMAT | SDL_DOUBLEBUF | SDL_HWSURFACE
  display = SDL_SetVideoMode(2 * image_width / dsf, image_height / dsf, 24 /* 32 */, 0); //SDL_ANYFORMAT | SDL_DOUBLEBUF | SDL_HWSURFACE

  if ( display == 0 )
  {
    exit(1);
  }

  Uint32 rmask, gmask, bmask, amask;
  
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
  rmask = 0xff000000;
  gmask = 0x00ff0000;
  bmask = 0x0000ff00;
#else
  rmask = 0x000000ff;
  gmask = 0x0000ff00;
  bmask = 0x00ff0000;
#endif

  SDL_WM_SetCaption( "VSLAM", NULL );    
  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

  left = SDL_CreateRGBSurface(SDL_SRCCOLORKEY|SDL_RLEACCEL, image_width/dsf, image_height/dsf, 32, rmask, gmask, bmask, 0);
  right = SDL_CreateRGBSurface(SDL_SRCCOLORKEY|SDL_RLEACCEL, image_width/dsf, image_height/dsf, 32, rmask, gmask, bmask, 0);
  left2 = SDL_CreateRGBSurface(SDL_SRCCOLORKEY|SDL_RLEACCEL, image_width/dsf, image_height/dsf, 32, rmask, gmask, bmask, 0);
  right2 = SDL_CreateRGBSurface(SDL_SRCCOLORKEY|SDL_RLEACCEL, image_width/dsf, image_height/dsf, 32, rmask, gmask, bmask, 0);
}

VisualisationWindow::~VisualisationWindow()
{
}

void VisualisationWindow::update()
{
  SDL_Rect r1;
  r1.x = 0;
  r1.y = 0;
  r1.w = image_width/dsf;
  r1.h = image_height/dsf;

  SDL_Rect r2;
  r2.x = image_width/dsf;
  r2.y = 0;
  r2.w = image_height/dsf;
  r2.h = image_width/dsf;

  if (!use2)
  {
	  boost::try_mutex::scoped_lock lock(blit_mutex);
	  // Blit to screen
	  //std::cout << "blitting to screen" << std::endl;
	  if ( SDL_BlitSurface(left, 0, display, &r1) )
	  {
		  std::cerr << "Blit failed" << std::endl;
	  }
	  if ( SDL_BlitSurface(right, 0, display, &r2) )
	  {
		  std::cerr << "Blit failed" << std::endl;
	  }
  }
  else
  {
	  boost::try_mutex::scoped_lock lock(blit_mutex2);
	  // Blit to screen
	  //std::cout << "blitting to screen" << std::endl;
	  if ( SDL_BlitSurface(left2, 0, display, &r1) )
	  {
		  std::cerr << "Blit failed" << std::endl;
	  }
	  if ( SDL_BlitSurface(right2, 0, display, &r2) )
	  {
		  std::cerr << "Blit failed" << std::endl;
	  }
  }
  SDL_Flip(display);
}

void VisualisationWindow::blit(SDL_Surface *l, SDL_Surface *r)
{
  SDL_Rect r1;
  r1.x = 0;
  r1.y = 0;
  r1.w = image_width/dsf;
  r1.h = image_height/dsf;

  boost::try_mutex::scoped_try_lock lock(blit_mutex);
  if (lock)
  {
	  //std::cout << "blitting to tmp" << std::endl;
	  if ( SDL_BlitSurface(l, 0, left, &r1) )
	  {
		  std::cerr << "Blit failed" << std::endl;
	  }
	  if ( SDL_BlitSurface(r, 0, right, &r1) )
	  {
		  std::cerr << "Blit failed" << std::endl;
	  }
	  use2 = 0;
  }
  else
  {
	  boost::try_mutex::scoped_try_lock lock(blit_mutex2);
	  if (lock)
	  {
		  //std::cout << "blitting to tmp" << std::endl;
		  if ( SDL_BlitSurface(l, 0, left2, &r1) )
		  {
			  std::cerr << "Blit failed" << std::endl;
		  }
		  if ( SDL_BlitSurface(r, 0, right2, &r1) )
		  {
			  std::cerr << "Blit failed" << std::endl;
		  }
		  use2 = 1;
	  }
  }
}
