/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "file_camera.h"

#include <sstream>
#include <signal.h>
#include <inttypes.h>
#include <iostream>
#include <cstring>
#include <iomanip>

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>

#include <cv.h>
#include <highgui.h>

int File_Camera::instances = 0;
Posix_Semaphore File_Camera::sync_sem[2];

// Use first Camera that matches the serial - remove Code duplication
File_Camera::File_Camera(std::string load_directory, std::string prefix,
						 unsigned int width, unsigned int height,
						 bool _play_once,
                                                 bool &_do_mainloop,
                                                 unsigned _skip_frames)
	: directory(load_directory),
	  frame_counter(0),
	  play_once(_play_once),
	  do_mainloop(_do_mainloop),
          skip_frames(_skip_frames)
{
  this_instance = instances++;
  if (this_instance == 0) {
    other_instance = 1;         
  }    
  else
  {        
    other_instance = 0;    
  }           
  sync_sem[this_instance] = Posix_Semaphore(1);

  // Acquire a list of filenames
  DIR *dp;
  struct dirent *dirp;
  if((dp  = opendir(directory.c_str())) == NULL) {
    std::cerr << "Error opening " << directory << std::endl;
  }
    
  while ((dirp = readdir(dp)) != NULL) {
    if (strcmp(dirp->d_name,".")!=0 && strcmp(dirp->d_name,"..")!=0 && dirp->d_name[0] == prefix.c_str()[0]) {
      file_names.push_back(std::string(dirp->d_name));
    }
  }
  closedir(dp);
  // File names may be unsorted -> sort
  file_names.sort();
  file_name_iterator = file_names.begin();
  currentImage = new Image(width, height, 1);
}

void File_Camera::repoll()
{
}

void File_Camera::wait()
{   
  // Open file get image    
  IplImage* img = cvLoadImage(std::string(directory + "/" + (*file_name_iterator)).c_str(), 0);   
  assert((int) currentImage->width == img->width && (int) currentImage->height && img->height);

  // Copy data
  if (img)
  {
    memcpy(currentImage->data, img->imageData, img->width * img->height);
    cvReleaseImage(&img);
  }
  
  // Increment counter
  currentImage->frame_count = frame_counter++;
  currentImage->is_log = true;

  // Set the timestamp
  std::string ts_string((*file_name_iterator).substr(2, (*file_name_iterator).find_last_of('.')-2));

  std::string os_timestamp(ts_string.substr(0, ts_string.find_first_of('-')));
  std::string timestamp(ts_string.substr(ts_string.find_first_of('-')+1, ts_string.size()-1));

  std::stringstream s(timestamp);
  uint64_t time_in_msec;
  s >> time_in_msec;
  currentImage->timestamp.tv_sec = time_in_msec / 1e6;
  currentImage->timestamp.tv_nsec = (uint64_t)(time_in_msec % 1000000) * 1000;

  std::stringstream s2(os_timestamp);
  s2 >> time_in_msec;
  currentImage->os_timestamp.tv_sec = time_in_msec / 1e6;
  currentImage->os_timestamp.tv_nsec = (uint64_t)(time_in_msec % 1000000) * 1000;

  // Enable frame skipping  
  unsigned int already_skipped(0);  
  while(++already_skipped <= skip_frames)
  {
      if (file_name_iterator++ == file_names.end())
      {          
          file_name_iterator = file_names.begin();
          if (play_once)
          {
              // instruct mainloop to exit
              do_mainloop = false;
          }  
      }
  }
    
  usleep(80000);

  // Wait
  sync_sem[this_instance].down();
  sync_sem[other_instance].up();
}

File_Camera::~File_Camera()
{
  delete currentImage;
}
