/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef IPLIMAGEWRAPPER_INCLUDED
#define IPLIMAGEWRAPPER_INCLUDED

#include <stdlib.h>
#include <algorithm>
#include <cxcore.h>

//! Class encapsulating an \c IplImage* pointer
/*! The class provides mainly a destructors, a copy constructor and a
    = operator. The sole purpose of this class is to allow having an
    image as a class member without the need to provide these
    operators.

    The main IplImage* members are wrapped by accessor functions.

    The object owns the image stored.
*/
class IplImageWrapper 
{
 public:
  //! Pointer to the actual image
  IplImage* p;

  IplImageWrapper ()
    :p(NULL)
    {}

    IplImageWrapper (int w, int h, int depth=IPL_DEPTH_8U, int channels=1)
      :p(NULL)      
      {
	create (w, h, depth, channels);	
      }

    IplImageWrapper (const IplImageWrapper& img2)
      {
	if (img2.p!=NULL) p = cvCloneImage (img2.p);
	else p = NULL;
      }

    IplImageWrapper& operator= (const IplImageWrapper& img2)
      {
	clear();
	if (img2.p!=NULL) p = cvCloneImage (img2.p);
	else p = NULL;
	return *this;	
      }	
      
    
    ~IplImageWrapper ()
      {
	clear();
      }
    

    void clear ()
    {
      if (p!=NULL) {
	cvReleaseImage (&p);
	p=NULL;
      }      
    } 

    void create (int w, int h, int depth=IPL_DEPTH_8U, int channels=1)
    {
      clear();		   
      CvSize size;
      size.width = w;
      size.height = h;	
      p = cvCreateImage (size, depth, channels);
    }

    CvSize size () const {      
      CvSize size;
      size.width = p->width;
      size.height = p->height;      
      return size;
    }

    bool empty () const
    {
      return p==NULL || p->imageData==NULL;
    }    

    IplImage* operator() (){return p;}
    const IplImage* operator() ()const {return p;}

    operator IplImage* (){return p;}
    operator const IplImage* ()const {return p;}


    unsigned char* imageData () {return (unsigned char*) p->imageData;}
    const unsigned char* imageData () const {return (unsigned char*) p->imageData;}
    
    int width () const {return p->width;}
    int height() const {return p->height;}    
    int widthStep() const {return p->widthStep;}
    
    //! Pixel access (only on byte images)
    unsigned char& operator() (int x, int y) {return (unsigned char&) p->imageData[x+y*p->widthStep];}
    const unsigned char& operator() (int x, int y) const {return (unsigned char&) p->imageData[x+y*p->widthStep];}

    void fill (unsigned char b) {memset(p->imageData, b, p->imageSize);}                     

    void random () {
      for (int y=0; y<height(); y++) for (int x=0; x<width(); x++) (*this)(x,y)=rand() & 0xff;
    }					             				      
};

inline void swap (IplImageWrapper& a, IplImageWrapper& b)
{
  std::swap<IplImage*> (a.p, b.p);
}
#endif
