/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MUTEX_H
#define MUTEX_H

#include <pthread.h>

/* A Mutex. Don't use stupid pthread directly.
 * A Mutex is very easy to use. Just construct
 * and make lock() and unlock(). That's it.
 * Looks much nicer and is more readable.
 *
 */

class Mutex
{
public:
    Mutex(){pthread_mutex_init(&mutex_, 0);};
    virtual ~Mutex(){pthread_mutex_destroy(&mutex_);};
    bool try_lock()
    {
	if ( pthread_mutex_trylock(&mutex_) != 0) return false;
	else return true;
    };

    void lock(){ pthread_mutex_lock(&mutex_); };
    void unlock(){pthread_mutex_unlock(&mutex_); };


private:
    Mutex( const Mutex&); // Makes no sense
    Mutex& operator=(const Mutex&); // Don't do that!

    pthread_mutex_t mutex_;
};


/* The Guard Class is the same as BSmart::SYNC Makro but doesn't
 * make such a big magic from its function. And you can use more than
 * one Mutex per class. See Ringbuffer for example.
 */

template<typename Mutex>
class Guard
{
public:
    explicit Guard(Mutex& mutex): mutex_(mutex){mutex_.lock();};
    ~Guard(){ mutex_.unlock(); };
private:
    Guard(const Mutex&); // No Copy-Constructor
    void operator=(const Guard&);
    Mutex& mutex_;
};

#endif
