/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef RINGBUFFER_H
#define RINGBUFFER_H

/* This provides a Small ringbuffer
 * you write (push) and see if it worked, if not the Buffer is
 * full. You can pop a Element, then new ELements can be inserted
 * Should be straight forward
 */

#include "posix_semaphore.h"
#include "mutex.h"
#include <iostream>
/* A Ringbuffer. What else should i say?
 * Use push() to put elements into the buffer
 * Use Pop to get ist back
 * push blocks when buffer is full, maybe it would be useful to have
 * a non-blocking push, but therefore the Semaphore Class has to be changed
 * The fist Template-Parameter is the type stored in the Buffer
 * the Second is the size of the buffer (number of elements)
 */

template <typename T,  int max_size = 5>
class Ringbuffer
{
public:
    Ringbuffer(): max_size_( max_size ), write_idx( 0 ), read_idx( 0 ) // , sem_write(max_size_-1)
    {
        data_ = new T[max_size_];
        for ( int i = 0; i< max_size_; ++i ) sem_write.up();
    };

    virtual ~Ringbuffer() { delete[] data_;};

    bool push_non_blocking(const T& input)
    {
	if (! sem_write.try_down() ) return false;
	Guard<Mutex> g(mutex);
	data_[write_idx] = input;
	++write_idx %= max_size_;
	sem_res.up();
	return true;
    };

    bool push(const T& input)
    {
	sem_write.down();
	Guard<Mutex> g(mutex);
	data_[write_idx] = input;
	++write_idx %= max_size_;
	sem_res.up();
	return true;
    };

    // Are there any readable Elements
    // its easier than counting the sem_val
    int size()
    {
	return sem_res.get_val();
    }

    // Can we write without blocking?
    // only useful if there is only one writer
    bool writeable()
    {
	return (sem_write.get_val() != 0);
    }

    void pop(T& t)
    {
	sem_res.down();
	Guard<Mutex> g(mutex);
	t = data_[read_idx];
	++read_idx %= max_size_;
	sem_write.up();
    };

    // Just inspect the next readable element
    void inspect(T& t)
    {
        Guard<Mutex> g(mutex);
	t = data_[read_idx];
    }

    // These operators simulate (very badly)
    // the cout,cerr,clog streams
    // you can write :
    //  rb << some_type;
    // but you can't write:
    // rb << some_type << some_type;
    virtual Ringbuffer<T,max_size>& operator<<(const T& in)
    {
	push_non_blocking(in);
	return *this;
    }

    virtual Ringbuffer<T, max_size>& operator>>(T& out)
    {
	pop(out);
	return *this;
    }

protected:

private:
    Ringbuffer(const Ringbuffer&);     // Don't copy Ringbuffers
    void operator=(const Ringbuffer&); // Don't assign Ringbuffers
    Mutex mutex;
    Posix_Semaphore sem_res;
    Posix_Semaphore sem_write;
    int max_size_;
    T* data_;
    unsigned int write_idx;
    unsigned int read_idx;
    volatile int elements_in_buffer;
};

#endif
