/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Oliver Birbach <oliver.birbach@dfki.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef THREAD_H
#define THREAD_H

#include <pthread.h>  // pthread
#include <unistd.h>   // usleep
#include <stdexcept>  // runtime_error
#include <cstdlib>
#include <cstring>
#include <signal.h>
#include <iostream>

#ifndef THREAD_EXCEPTION
#define THREAD_EXCEPTION

class thread_exception : public std::runtime_error
{
public:
    thread_exception() : runtime_error("generic thread execption error")
    {
    }

    thread_exception(const char* what_in) : runtime_error(what_in)
    {
    }
};

#endif

// A class encapsulating a pthread
class Thread
{
public:
    pthread_t handle;	   // The pthread-handle
    volatile bool running; // A flag which indicates the state of the thread

public:
    // Constructor
    Thread()
    {
	handle = 0;
	running = false;
    }

    // Destructor
    // Stops the thread, if it is still running
    virtual ~Thread()
    {
	if ( running )
	    stop();
    }

    // run implements the thread's main function
    void virtual run() = 0;

    // Creates a new pthread and starts it
    // throws thread_exception if it fails
    void start()
    {
	if(running) throw thread_exception("thread already running");
	running = true;        
	int ret_val = pthread_create( &handle, 0, &Thread::init_thread, this);
	if ( ret_val != 0 ) throw thread_exception(strerror(ret_val));
    }
    
    void start(int priority)
    {            
      start();
      
      int ret;
      int policy = SCHED_FIFO;
      sched_param param;
      param.sched_priority = priority;
      ret = pthread_setschedparam (handle, policy, &param);
    }

    // Stops the thread
    // The thread is given about a third of a second to stop running.
    // After that, pthread_cancel and pthread_join are called
    void stop()
    {
	running = false;        
        pthread_cancel(handle);
        usleep(150000); //Wait for 150 ms
//	pthread_join(handle, 0);
    }

    // Joins the thread
    void join()
    {
//	pthread_join(handle, 0);
    }

    // Announces stop so can stop
    // call join afterwards
    void announce_stop()
    {
	running = false;
    }

    // Returns the state of the thread
    bool is_running() const
    {
	return running;
    }

protected:
    // Initializes the pthread parameters and starts
    // the run() function
    static void* init_thread(void* param)
    {
	Thread* cl = (Thread*) param;
	pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, 0);
	pthread_setcanceltype  (PTHREAD_CANCEL_ASYNCHRONOUS, 0);
	cl->run();
	return 0;
    }
};

#endif //THREAD_H
