/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef BOOST_CHECK_CLOSE_VECTOR_HH_
#define BOOST_CHECK_CLOSE_VECTOR_HH_


#include <boost/test/floating_point_comparison.hpp>

#include <Eigen/Core>

namespace boost {

namespace test_tools {

using unit_test::readonly_property;


template<class ftype, int dim1, int dim2>
class close_at_tolerance<Eigen::Matrix<ftype, dim1, dim2> >{
public:
	// Public typedefs
	typedef bool result_type;
	typedef ftype FPT;
	
	typedef Eigen::Matrix<ftype, dim1, dim2> comp_type;
	
	// Constructor
	template<typename ToleranceBaseType>
	explicit    close_at_tolerance( percent_tolerance_t<ToleranceBaseType>  tolerance, 
			floating_point_comparison_type          fpc_type = FPC_STRONG ) 
	: p_fraction_tolerance( tt_detail::fpt_abs( static_cast<FPT>(0.01)*tolerance.m_value ) )
	, p_strong_or_weak( fpc_type ==  FPC_STRONG )
	{}
	template<typename ToleranceBaseType>
	explicit    close_at_tolerance( fraction_tolerance_t<ToleranceBaseType> tolerance, 
			floating_point_comparison_type          fpc_type = FPC_STRONG ) 
	: p_fraction_tolerance( tt_detail::fpt_abs( tolerance.m_value ) )
	, p_strong_or_weak( fpc_type ==  FPC_STRONG )
	{}
	
	bool        operator()( comp_type left, comp_type right ) const
	{
		FPT diff = ( left - right ).norm();
		FPT d1   = tt_detail::safe_fpt_division( diff, right.norm() );
		FPT d2   = tt_detail::safe_fpt_division( diff, left.norm() );
		
		return p_strong_or_weak 
		? (d1 <= p_fraction_tolerance && d2 <= p_fraction_tolerance) 
				: (d1 <= p_fraction_tolerance || d2 <= p_fraction_tolerance);
	}
	
	// Public properties
	readonly_property<FPT>  p_fraction_tolerance;
	readonly_property<bool> p_strong_or_weak;
};

}
}

#endif /* BOOST_CHECK_CLOSE_VECTOR_HH_ */
