/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#define BOOST_TEST_MODULE imu_test
#include <boost/test/included/unit_test.hpp>
#include "boost_check_close_vector.hh"


#include <vslam/core/optimizer.hpp>

#include "random_vector.hh"


BOOST_AUTO_TEST_SUITE( imu_test_suite )

using namespace vslam;


random_vector<> r;

BOOST_AUTO_TEST_CASE( random_generator_test )
{
	int step_num = 1000000;
	Eigen::Matrix3d cov;
	cov.setZero();
	Eigen::Vector3d mean;
	mean.setZero();
	
	for(int i=0; i<step_num*0; ++i)
	{
		Eigen::Vector3d val = r.get_normal<3>();
		mean += val;
		cov += val * val.transpose();
	}
	mean /= step_num;
	cov /= step_num;
	
	std:: cout << mean.transpose() << "\n" << cov << std::endl;
	
	mean.setZero(); cov.setZero();
	SO3 q0;
	double delt[] = {0.002147, 0.0003762, -0.001005};
	q0.boxplus(delt, -1);
	for(int i=0; i<step_num; ++i)
	{
		SO3 qd( Eigen::Quaterniond(r.get_uniform_sphere<4>()) );
		Eigen::Vector3d delta;
		qd.boxminus(delta.data(), q0);
		mean += delta;
		cov += delta * delta.transpose();
	}
	mean /= step_num;
	cov /= step_num;
	
	std:: cout << mean.transpose() << "\n" << cov << std::endl;
	
	
	
	
}


BOOST_AUTO_TEST_CASE( initialization )
{
	for(int i=0; i<16; ++i)
	{
		vec3 a = 9.81 * r.get_uniform_sphere<3>();
		vec3 p = r.get_uniform<3>();
		
		SO3 imu2world = optimizer::orientation_from_gravity(a);
		
		
		// measured gravitation should point upwards: 
		BOOST_CHECK_CLOSE_FRACTION(imu2world * a, Eigen::Vector3d(0,0,9.81), 1e-6);
		
		// rotated x-vector is supposed to have y-component of 0:
		BOOST_CHECK_SMALL( (imu2world * Eigen::Vector3d(1,0,0) )[1], 1e-6 );
	}
}

BOOST_AUTO_TEST_CASE( accumulation )
{
	double dt = 1.0/512;
	
	SO3 start_orient = SO3::exp(r.get_uniform<3>() );
	vec3 start_pos = r.get_uniform<3>(), start_vel = r.get_uniform<3>();
	optimizer::Pose start_pose(start_orient, start_pos, start_vel);
	
	for(double gravity = 0; gravity < 10; gravity += 0.981)
	{
		optimizer::imu_accumulator imu_acc;
		optimizer::Pose current_pose = start_pose;
		Eigen::Vector3d g_vec(0, 0, gravity);
		for(int i=0; i<512; ++i)
		{
			// acceleration with 1m/s^2 and rotation with 1deg/s in arbitrary direction:
			vec3 a = r.get_uniform_sphere<3>(); // world coordinates
			vec3 w = 3.14/180 * r.get_uniform_sphere<3>(); // imu coordinates
			
			// FIXME order of transformations does make a difference!
			// better would be to use 'exact' integration, assuming w is constant over time
			current_pose.orient = current_pose.orient * SO3::exp(w, dt);
			vec3 a_meas = current_pose.orient.conjugate() * (a + g_vec);
			current_pose.pos += (current_pose.vel + a * dt/2 ) * dt;
			current_pose.vel += a * dt;
			
			imu_acc.accumulate(a_meas, w, dt);
		}
		
		optimizer::Pose endpose = imu_acc.integrate_pose(start_pose, gravity);
		BOOST_CHECK_CLOSE_FRACTION(Eigen::Vector3d(endpose.pos), Eigen::Vector3d(current_pose.pos), 1e-6);
		BOOST_CHECK_CLOSE_FRACTION(Eigen::Vector3d(endpose.vel), Eigen::Vector3d(current_pose.vel), 1e-6);
		vec3 delta_orient;
		endpose.orient.boxminus(delta_orient.data(), current_pose.orient);
		BOOST_CHECK_SMALL(delta_orient.norm(), 1e-6);
	}
}

BOOST_AUTO_TEST_CASE( check_noises )
{
	// generate measurements, accumulate with noises added and check covariance of result.
	double dt = 1.0/512, gravity = 9.81;
	
	double acc_noise = 1e-1, gyr_noise = 512e-0;
	
	int sample_size = 4096;
	
	SO3 start_orient = SO3::exp(r.get_uniform<3>() );
	vec3 start_pos = r.get_uniform<3>(), start_vel = r.get_uniform<3>();
	optimizer::Pose start_pose(start_orient, start_pos, start_vel);
	
	Eigen::Matrix<double, optimizer::Pose::DOF, optimizer::Pose::DOF> cov;
	cov.setZero();
	
	for(int n=0; n < sample_size; ++n)
	{
		random_vector<> tr; // always generate the same trajectory
		optimizer::imu_accumulator imu_acc;
		optimizer::Pose current_pose = start_pose;
		Eigen::Vector3d g_vec(0, 0, gravity);
		for(int i=0; i<512; ++i)
		{
			// acceleration with 1m/s^2 and rotation with 1deg/s in arbitrary direction:
			vec3 a = tr.get_uniform_sphere<3>(); // world coordinates
			vec3 w = 3.14/180 * tr.get_uniform_sphere<3>(); // imu coordinates
			
			// FIXME order of transformations does make a difference!
			// better would be to use 'exact' integration, assuming w is constant over time
			current_pose.orient = current_pose.orient * SO3::exp(w, dt);
			vec3 a_meas = current_pose.orient.conjugate() * (a + g_vec) + r.get_normal<3>() * acc_noise;
			vec3 w_meas = w + r.get_normal<3>() * gyr_noise;
			current_pose.pos += (current_pose.vel + a * dt/2 ) * dt;
			current_pose.vel += a * dt;
			
			imu_acc.accumulate(a_meas, w_meas, dt);
		}
		
		optimizer::Pose endpose = imu_acc.integrate_pose(start_pose, gravity);
		
		Eigen::Matrix<double, optimizer::Pose::DOF, 1> delta;
		endpose.boxminus(delta.data(), current_pose);
		
		cov += delta * delta.transpose();
	}
	std::cout << cov / sample_size << std::endl;
	std::cerr << std::flush;
}

BOOST_AUTO_TEST_SUITE_END()
