/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include <string>
#include <fstream>

#include <Eigen/Core>
#include <vslam/camera/calibrator.hpp>
#include <framework/transformations.h>
#include <mtk/mean_and_covar.hpp>

#include <mtk/types/pose.hpp>


typedef vslam::camera::image_point image_point;


class line_extractor : public std::ifstream
{
	typedef std::ifstream base;
	std::istringstream output;
public:
	line_extractor(const std::string &input_name) : base(input_name.c_str()) {}
	
	bool good()
	{
		return !base::eof() &&  !base::bad();
	}
	
	std::istringstream& get_line(bool verbose = false)
	{
		std::string line;
		std::getline(*this, line);
		if(verbose)
			std::cerr << line << std::endl;
		output.str(line);
		return output;
	}
};

void parse_points(std::istream &inp, std::vector<image_point>& out, int num_points, bool interleaved)
{
	out.resize(num_points);
	for(int i=0; i<num_points; ++i)
	{
		inp >> out[i].x;
		if(interleaved)
			inp >> out[i].y;
	}
	if(interleaved)
		return;
	for(int i=0; i<num_points; ++i)
	{
		inp >> out[i].y;
	}
}


//MTK_MAKE_POSE3D(pose, double, pos, orient, )
typedef MTK::trafo<MTK::SO3<double> > pose;

double parse_qualisys(std::istream &inp, vslam::pose_as_matrix &star2qualiworld)
{
	// FIXME everything hardcoded ...
	for(int i=0; i<11; ++i)
	{
		std::string dummy;
		std::getline(inp, dummy);
	}
	int num_values = 100;
	std::vector<pose, Eigen::aligned_allocator<pose> > poses;
	poses.reserve(num_values);
	
	for(int i=0; i<num_values; ++i)
	{
		Eigen::Vector3d pos;
		Eigen::Matrix3d orient;
		double roll, pitch, yaw;
		double res;
		
		for(int i=0; i<3; ++i)
			inp >> pos(i);
		inp >> roll >> pitch >> yaw >> res; // values are ignored
		
		for(int i=0; i<3; ++i)
			for(int j=0; j<3; ++j)
				inp >> orient(j,i);
		
		if(res < 0)
			continue;
		
		poses.push_back(pose(pos/1000, Eigen::Quaterniond(orient)));
	}
		
	pose mean;
	Eigen::Matrix<double, 6, 6> cov;

	double res = MTK::mean_and_covariance(mean, cov, poses, 16);
	double sigma = cov.diagonal().maxCoeff();
	std::cout << "res: " << res << ", sigma: " << sigma << std::endl;
	
	star2qualiworld = vslam::pose_as_matrix(mean.orient, mean.pos);

	
	return sigma;
}

int main(int argc, char **argv)
{
	if(argc < 2)
	{
		std::cerr << "Please specify input directory\n";
	}

	int num_x = 7, num_y = 7;
	double gridlength = 0.11125;
	double acc_sigma = 0.05;
	double im_sigma = .3;

	std::string dir(argv[1]);
	
	Transform::CalibrationFromFile(dir + "/calibration.txt");
	
	std::cout.precision(16);
	Transform::CalibrationToStream(std::cout, true);

	vslam::calibrator calib;
	
	calib.initialize(gridlength, num_x, num_y, Transform::cam, Transform::left2imu, false);

	line_extractor imu(dir + "/imu.dat");
	line_extractor left(dir + "/left.dat");
	line_extractor right(dir + "/right.dat");

	bool interleaved = false;
	int count = 1;
	std::vector<image_point> left_p, right_p;
	while(imu.good())
	{
		int c;
		Eigen::Vector3d acc;
		imu >> c >> acc(0) >> acc(1) >> acc(2);
		if(!imu.good())
			break;
		assert(c==count);

		parse_points(left, left_p, num_x*num_y, interleaved);
		parse_points(right, right_p, num_x*num_y, interleaved);

		std::ostringstream gt_path;
		gt_path << dir << "/gt_" << count << ".tsv";
		
		std::ifstream groundtruth(gt_path.str().c_str());
		
		
		vslam::pose_as_matrix gt_pose;
		
		double gt_sigma = parse_qualisys(groundtruth, gt_pose);
		if(((count !=1 && count != 6) || !interleaved) && count != 90 && count != 60)
		{
			calib.add_frame(acc, acc_sigma, left_p, right_p, im_sigma);
			
			//FIXME: obtaining sigma as variance in measurements gives unrealistic value
			calib.add_groundtruth(gt_pose, 1e-2+0*gt_sigma*1e2);
		}
		++count;
	}
	calib.analyze();
	
	calib.optimize();
	calib.analyze(true);
	std::cout << "\n\n";
	
	
	calib.get_result(Transform::cam, Transform::left2imu);
	
	Transform::CalibrationToStream(std::cout, true);
	
	vslam::pose_as_matrix star2left, checker2world;
	calib.get_result(star2left, checker2world);
	
	std::cout << "\n" << star2left << std::endl << checker2world << std::endl;

	return 0;
}
