/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */


#include <fstream>

#include <vslam/core/pose_as_matrix.hpp>
#include "calibration/qualisys_from_file.hpp"

#include <slom/Estimator.hpp>
#include <slom/BuildMeasurement.hpp>
#include <slom/CallBack.hpp>

typedef vslam::pose_as_matrix pose;
SLOM_RANDOMVAR_ID(pose);

SLOM_BUILD_MEASUREMENT(pose_to_pose, pose::DOF, 
                       ((pose, quali2world)) ((pose, imu2star)), 
                       ((pose, star2quali)) ((pose, imu2world))
)

void pose_to_pose::eval(MTK::vectview<double, pose::DOF> ret) const
{
	pose imu2world_ = *quali2world * star2quali * *imu2star;
	imu2world.boxminus(ret, imu2world_);
}

SLOM_BUILD_MEASUREMENT(fix_1st_pose, 3, 
                       ((pose, quali2world)) ((pose, imu2star)), 
                       ((pose, star2quali)) ((pose, imu2world))
)

void fix_1st_pose::eval(MTK::vectview<double, 3> ret) const
{
	pose imu2world_ = *quali2world * star2quali * *imu2star;
	ret = 1e6 * (imu2world_.get_translation() - imu2world.get_translation());
}


std::istream& get_vslam_pose(std::istream& in, Eigen::Quaterniond &q , Eigen::Vector3d &p, Eigen::Vector3d &v)
{
	in >> q.x() >> q.y() >> q.z() >> q.w()
	   >> p.x() >> p.y() >> p.z()
	   >> v.x() >> v.y() >> v.z();
	
	std::string dummy;
	std::getline(in, dummy);
	return in;
}


int main(int argc, char **argv)
{
	if(argc < 5){
		std::cerr << "Usage:\n " << argv[0] << " <calibfile> <estimation> <groundtruth> <output> [offset [frequency]]\n";
		return 1;
	}
	
	SLOM::Estimator e;
	SLOM::CallBack cb(e, 3);
	e.setCallBack(cb);
	
	// if offset parameter is passed do a fit between trajectories
	bool fit = (argc > 5);
	double offset = fit ? atof(argv[5]) : 0;
	
	std::cout << argv[5] << " " << offset << std::endl;
	// frequency of groundtruth vs vslam (default 10)
	double frequency = argc > 6 ? atof(argv[6]) : 10;
	
	pose star2left, left2imu;
	std::ifstream calib(argv[1]);
	calib >> left2imu;
	calib >> star2left;
	pose imu2star = star2left.inverse() * left2imu.inverse();
	
	pose_id imu2star_var = e.insertRV(imu2star, fit);
	pose_id quali2world_var;

	std::ifstream estimation(argv[2]), groundtruth(argv[3]);
	std::ofstream output(argv[4]);
	
	Eigen::Quaterniond orient;
	Eigen::Vector3d pos, vel;
	get_vslam_pose(estimation, orient, pos, vel);
	assert(estimation.good());

	pose imu2world(orient, pos);
	
	std::cout << "imu2star: " << imu2star << "\nimu2world[0]: " << imu2world << std::endl;
	
	qualisys::skip_header(groundtruth);
	
	double vslam_line = offset;

	pose qualy2world;
	
	std::deque<pose_to_pose::id> measurements;
	for(int line=0; groundtruth.good(); ++line)
	{
		pose star2qualy;
		
		double res = qualisys::readline(groundtruth, star2qualy);
		
		if(res < 0 || !groundtruth.good())
		{
			std::cerr << "Bad line\n";
			continue;
		}
		
		if(!quali2world_var)
		{
			qualy2world = imu2world * imu2star.inverse() * star2qualy.inverse();
			quali2world_var = e.insertRV(qualy2world);
			std::cerr << "qualy2world: " << qualy2world << std::endl;
			e.insertMeasurement(fix_1st_pose(quali2world_var, imu2star_var, star2qualy, imu2world));
		}
		
		if(vslam_line <= line)
		{
			measurements.push_back(
					e.insertMeasurement(pose_to_pose(quali2world_var, imu2star_var, star2qualy, imu2world))
			);
			vslam_line += frequency;
			get_vslam_pose(estimation, orient, pos, vel);
			if(!estimation.good()) break;
			imu2world = pose(orient, pos);
			// now calculate imu2world and dump result:

			pose imu2world_quali = qualy2world * star2qualy * imu2star;

			Eigen::Quaterniond q(imu2world_quali.get_rotation());
			//output << q.coeffs().transpose() << " " << imu2world_quali.get_translation().transpose() << std::endl;
		}
	}

	for(int i=0; i<50; ++i)
	{
		double gain = e.optimizeStep(false);
		if(0<= gain && gain < 1e-8) break;
	}
	
	std::cerr << qualy2world << std::endl << *quali2world_var << std::endl;
	
	std::ofstream quali_out("quali_aligned.dat");
	for(size_t i=0; i<measurements.size(); ++i)
	{
		const pose_to_pose &m = *measurements[i];
		pose imu2world_ = *quali2world_var * m.star2quali * *imu2star_var;
		Eigen::Quaterniond q(imu2world_.get_rotation());
		output << q.coeffs().transpose() << " " << imu2world_.get_translation().transpose();
		Eigen::Matrix<double, pose_to_pose::DIM, 1> res;
		m.eval(res);
		output << " " << res.start<3>().squaredNorm() << " " << res.end<3>().squaredNorm() << std::endl;
	}
	
	return 0;
}
