/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Tobias Hammer <hammer@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "imageviewer.h"

#include <boost/filesystem.hpp>
namespace fs = boost::filesystem;
#include <opencv/highgui.h>

#define NO_MTK_SUPPORT 1
#include <vslam/camera/cv_camera.hpp>

#include "stereoimagematches.h"


ImageViewer::ImageViewer( const StereoImageMatches& sim, const std::string& workdir,
                          int window_Width, const std::string& windowName, const vslam::stereo_camera& camera )
    : m_sim(sim), m_workdir( workdir ), m_camera( camera ), m_windowWidth( window_Width ), m_windowName( windowName )
{
    m_leftImg = NULL;
    m_rightImg = NULL;
    m_multiImg = NULL;

    m_activeMatchIdx = 0;
    m_crossSize = 5;
    m_showCircles = true;
    m_showEpilines = true;
    m_minimalDisplay = false;
    m_largeX = false;

    m_leftImg = loadImage( m_sim.lImageName() );
    m_rightImg = loadImage( m_sim.rImageName() );

    checkAndInit();
}



ImageViewer::~ImageViewer()
{
    if ( m_leftImg )
        cvReleaseImage( &m_leftImg );
    if ( m_rightImg )
        cvReleaseImage( &m_rightImg );
    if ( m_multiImg )
        cvReleaseImage( &m_multiImg );
}



float ImageViewer::getScale() const
{
    return m_scale;
}


void ImageViewer::setActiveMatch( size_t point_idx )
{
    m_activeMatchIdx = point_idx;
}


void ImageViewer::show()
{
    newMultiImage();
    drawMatches();

    cvShowImage( m_windowName.c_str(), m_multiImg );
}


// ------------------ private -------------------

void ImageViewer::checkAndInit()
{
    // error check on images
    if ( m_leftImg->width != m_rightImg->width || m_leftImg->height != m_rightImg->height )
        throw "Images have different sizes";

    m_multiWidth = m_leftImg->width * 2;
    m_multiHeight = m_leftImg->height;
    m_scale = 1.0;

    if ( m_multiWidth > m_windowWidth )
    {
        m_scale = (float)m_windowWidth / (2.0*m_leftImg->width);
        m_multiWidth = m_windowWidth;
        m_multiHeight = m_scale * m_leftImg->height;
    }
}


IplImage* ImageViewer::loadImage( std::string filename ) const
{
    IplImage *res = NULL;
    res = cvLoadImage( fs::complete( filename, m_workdir ).string().c_str() );
    if ( res == NULL )
        throw "Failed opening image " + filename;
    return res;
}


void ImageViewer::newMultiImage()
{
    cvReleaseImage( &m_multiImg );

    m_multiImg = cvCreateImage( cvSize( m_multiWidth, m_multiHeight ), m_leftImg->depth, m_leftImg->nChannels );

    cvSetImageROI( m_multiImg, cvRect( 0, 0, m_multiImg->width/2 - 1, m_multiImg->height - 1 ) );
    cvResize( m_leftImg, m_multiImg );
    cvResetImageROI( m_multiImg );
    cvSetImageROI( m_multiImg, cvRect( m_multiImg->width/2, 0, m_multiImg->width/2 - 1, m_multiImg->height - 1 ) );
    cvResize( m_rightImg, m_multiImg );
    cvResetImageROI( m_multiImg );
}



void ImageViewer::drawMatches()
{
    CvScalar red = CV_RGB( 255, 0, 0 );

    if ( !m_minimalDisplay )
    {
        drawEpilines();

        // draw matches
        for ( size_t idx = 0; idx < m_sim.matchCount(); ++idx )
        {
            StereoPoint sp = m_sim.getSPoint( idx );

            CvScalar col = CV_RGB( 0, 255, 255 );
            if ( m_sim.getRating( idx ) != StereoImageMatches::R_UNRATED )
                col = CV_RGB( 255, 255, 0 );

            if ( m_activeMatchIdx == idx )
            {
                // circles in right image, point as mid
                CvPoint mid_r = cvPoint( (int)(sp.r.x * m_scale + 0.5) + m_multiImg->width/2,
                                         (int)(sp.r.y * m_scale + 0.5) ),
                        mid_l = cvPoint( sp.l.x * m_scale + 0.5, sp.l.y * m_scale + 0.5 );

                if ( !m_largeX )
                {
                    cvCircle( m_multiImg, mid_l, 3, red, -1 );
                    drawCircles( mid_r );
                    cvCircle( m_multiImg, mid_r, 1, red, -1 );
                }
                else
                {
                    drawCross( m_multiImg, mid_l.x, mid_l.y, red, 3 );
                    drawCross( m_multiImg, mid_r.x, mid_r.y, red, 3 );
                }
            }
            else
            {
                // draw crosss on both sides for not active
                int thickness = 1;
                drawCross( m_multiImg, sp.l.x * m_scale + 0.5, sp.l.y * m_scale + 0.5, col, thickness );
                drawCross( m_multiImg, m_multiImg->width/2 + sp.r.x * m_scale + 0.5, sp.r.y * m_scale + 0.5, col, thickness );
            }

        }
    }

    else    // m_minimalDisplay == true
    {
        drawEpilines();

        StereoPoint sp = m_sim.getSPoint( m_activeMatchIdx );

        // small red dot on bot sides
        CvPoint mid_r = cvPoint( (int)(sp.r.x * m_scale + 0.5) + m_multiImg->width/2,
                                 (int)(sp.r.y * m_scale + 0.5) ),
                mid_l = cvPoint( sp.l.x * m_scale + 0.5, sp.l.y * m_scale + 0.5 );

        if ( !m_largeX )
        {
            cvCircle( m_multiImg, mid_l, 1, CV_RGB( 255, 0, 0 ), -1 );
            cvCircle( m_multiImg, mid_r, 1, CV_RGB( 255, 0, 0 ), -1 );

            drawCircles( mid_r );
        }
        else
        {
            drawCross( m_multiImg, mid_l.x, mid_l.y, red, 3 );
            drawCross( m_multiImg, mid_r.x, mid_r.y, red, 3 );
        }
    }
}


void ImageViewer::drawEpilines()
{
    // draw epiline for current point
    // set roi of multiimage to shift line to right side AND to clip it (prevents drawing in left image)
    if ( m_showEpilines )
    {
        cvSetImageROI( m_multiImg, cvRect( m_multiImg->width/2, 0, m_multiImg->width/2 - 1, m_multiImg->height - 1 ) );

        StereoPoint active_sp = m_sim.getSPoint( m_activeMatchIdx );
        std::vector<vslam::camera::image_point> epiline_points;
        m_camera.get_epiline( true, vslam::camera::image_point( active_sp.l ), epiline_points );

        if ( epiline_points.size() > 1 )
        {
            for ( size_t i = 0; i < epiline_points.size() - 1; ++i )
            {
                CvPoint p1( epiline_points[i] ), p2( epiline_points[i+1] );
                p1.x = p1.x * m_scale;// + m_multiImg->width/2;
                p2.x = p2.x * m_scale;//  + m_multiImg->width/2;
                p1.y *= m_scale;
                p2.y *= m_scale;
                cvLine( m_multiImg, p1, p2, CV_RGB( 0, 255, 255 ), 1 );
            }
        }

        cvResetImageROI( m_multiImg );
    }
}


void ImageViewer::drawCircles( const CvPoint& mid )
{
    if ( m_showCircles )
    {
/*        cvCircle( m_multiImg, mid, sp.size_right*m_scale, CV_RGB( 255, 200, 200 ), 1 );   // 100%
        cvCircle( m_multiImg, mid, sp.size_right*m_scale*3/4, CV_RGB( 255, 150, 150 ), 1 );   // 75%
        cvCircle( m_multiImg, mid, sp.size_right*m_scale/2, CV_RGB( 255, 100, 100 ), 1 );   // 50%
        cvCircle( m_multiImg, mid, sp.size_right*m_scale/4, CV_RGB( 255, 50, 50 ), 1 );   // 25%
        */
        cvCircle( m_multiImg, mid, StereoImageMatches::getMatchRadius( StereoImageMatches::R_CLOSE )*m_scale, CV_RGB( 255, 200, 200 ), 1 );   // close (<=5*2 atm)
        cvCircle( m_multiImg, mid, StereoImageMatches::getMatchRadius( StereoImageMatches::R_FAR )*m_scale, CV_RGB( 255, 50, 50 ), 1 );   // far (<=20*2 atm)
    }
}


void ImageViewer::drawCross( IplImage *img, int x, int y, CvScalar color, int thickness ) const
{
    int l = x - m_crossSize,
        r = x + m_crossSize,
        t = y - m_crossSize,
        b = y + m_crossSize;

    cvLine( img, cvPoint( l, t ), cvPoint( r, b ), color, thickness );
    cvLine( img, cvPoint( r, t ), cvPoint( l, b ), color, thickness );
}
