/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Tobias Hammer <hammer@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "stereoimagematches.h"

#include "helpers.h"
#include <iostream>
#include <fstream>


StereoImageMatches::StereoImageMatches( const std::string& matchfile )
{
    m_fileTextMap.resize( _R_MAX_ );
    m_fileTextMap[R_CLOSE] = "close";
    m_fileTextMap[R_FAR] = "far";
    m_fileTextMap[R_UNRELATED] = "unrelated";
    m_fileTextMap[R_UNCLEAR] = "unclear";
    m_fileTextMap[R_SPURIOUS] = "spurious";
    m_fileTextMap[R_UNRATED] = "_";

    loadFile( matchfile );
}


StereoImageMatches::~StereoImageMatches()
{
    // nothing
}


void StereoImageMatches::loadFile( const std::string& matchfile )
{
    std::vector<std::string> lines;
    std::vector<std::string> toks;

    m_matchfile = matchfile;
    m_spoints.clear();
    m_rating.clear();

    if ( !Helper::loadFileToVector( matchfile, lines ) || lines.size() < 1 )
        throw "Invalid matchfile";

    // store image filenames from first line
    toks = Helper::splitStringToVector( lines[0], 3 );
    m_limage = toks[1];
    m_rimage = toks[2];

    // parse rest of file
    for ( size_t i = 1; i < lines.size(); ++i )
    {
        toks = Helper::splitStringToVector( lines[i], 6 );
        std::vector<double> dvec = Helper::vector2double( toks );
        m_spoints.push_back( StereoPoint( dvec[0], dvec[1], dvec[2], dvec[3], dvec[4]*2.0, dvec[5]*2.0 ) );     // XXX: factor 2 for scaling!
    }

    tryLoadRating();
}


std::string StereoImageMatches::lImageName() const
{
    return m_limage;
}


std::string StereoImageMatches::rImageName() const
{
    return m_rimage;
}


size_t StereoImageMatches::matchCount() const
{
    return m_spoints.size();
}


StereoPoint StereoImageMatches::getSPoint( size_t pos ) const
{
    return m_spoints[pos];
}


void StereoImageMatches::setRating( size_t pos, int rating )
{
    if ( pos > m_rating.size() )
        throw "Invalid pos for rating";
    m_rating[pos] = rating;
}


int StereoImageMatches::getRating( size_t pos ) const
{
    return m_rating[pos];
}


int StereoImageMatches::getNumUnrated() const
{
    int res = 0;
    for ( size_t i = 0; i < m_rating.size(); ++i )
        if ( m_rating[i] == R_UNRATED )
            res += 1;
    return res;
}


std::string StereoImageMatches::getRatingAsString( size_t pos ) const
{
    switch ( getRating(pos) )
    {
    case R_CLOSE:     return "close (<=5*2 Pixl)";
    case R_FAR:     return "far (<=20*2 Pixel)";
    case R_UNRELATED:     return "unrelated";
    case R_UNCLEAR: return "unclear";
    case R_SPURIOUS:return "spurious";
    case R_UNRATED: return "unrated";
    default: return "ERROR";
    }
}


std::string StereoImageMatches::constructRatingfileName() const
{
    size_t dotpos = m_matchfile.find_last_of( '.' );
    if ( dotpos == std::string::npos )
        throw "matchfile name no '.'' found";
    return m_matchfile.substr( 0, dotpos ) + "_rating.dat";
}


void StereoImageMatches::tryLoadRating()
{
    m_rating = std::vector<int>( m_spoints.size(), R_UNRATED );
    std::vector<std::string> lines;

    // try load contents only if file exists
    if ( Helper::loadFileToVector( constructRatingfileName(), lines ) )
    {
        size_t c = 0;

        if ( lines.size() < 1 )
            throw "Invalid rating file";
        if ( Helper::splitStringToVector( lines[0], 2 )[1] != m_matchfile )
            throw "Rating file association mismatch";

        for ( size_t i = 1; i < lines.size(); ++i )
        {
            std::string l = lines[i];

            int idx = -1;
            for ( size_t j = 0; j < m_fileTextMap.size(); ++j )
                if ( m_fileTextMap[j] == l )
                {
                    idx = (int)j;
                    break;
                }

            if ( idx != -1 )
            {
                if ( c >= m_spoints.size() )
                {
                    if ( !l.empty() )
                        throw "ratingfile has too many entries";
                }
                else
                {
                    m_rating[c] = idx;
                    ++c;
                }
            }
            else
                throw std::string("Invalid line in rating file '") + l + "'";
        }
    }
}


void StereoImageMatches::writeRating() const
{
    std::string fname_tmp = constructRatingfileName();
    std::ofstream fileobj( fname_tmp.c_str(), std::ios_base::out | std::ios_base::trunc );
    if ( !fileobj.is_open() )
        throw "Could not open rating file " + fname_tmp;

    fileobj << "# " << m_matchfile << std::endl;
    for ( size_t i = 0; i < m_rating.size(); ++i )
    {
        if ( m_rating[i] < 0 || m_rating[i] >= _R_MAX_ )
            throw "Internal error (invalid rating index)";
        else
            fileobj << m_fileTextMap[m_rating[i]] << std::endl;
    }
}


float StereoImageMatches::getMatchRadius( int type )
{
    switch (type)
    {
    case R_CLOSE: return 5*2;
    case R_FAR: return 20*2;
    default: throw "getMatchRadius(): wrong type";
    }
}



