/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Tobias Hammer <hammer@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "stereomatchgui.h"

#include <iostream>
#include <fstream>
#include <istream>
#include <fnmatch.h>
#include <cstdio>

#include <boost/filesystem.hpp>
namespace fs = boost::filesystem;

#include <opencv/highgui.h>

#define NO_MTK_SUPPORT 1
#include <vslam/camera/cv_camera.hpp>
#include "transformations.h"

#include "stereoimagematches.h"
#include "imageviewer.h"



// stl helper functors
struct StereoMatchGui::pathname_of {
    std::string operator()( const fs::directory_entry& p ) const {
        return p.path().string(); // or your creativity here
    }
};

struct StereoMatchGui::pathname_filter_neg {
    pathname_filter_neg( const std::string exp_ ) : exp(exp_) {}
    bool operator()( const std::string& path ) const {
        return fnmatch( exp.c_str(), path.c_str(), 0 ) != 0;
    }
    std::string exp;
};


struct StereoMatchGui::pathname_filter {
    pathname_filter( const std::string exp_ ) : exp(exp_) {}
    bool operator()( const std::string& path ) const {
        return !(fnmatch( exp.c_str(), path.c_str(), 0 ) != 0);
    }
    std::string exp;
};



StereoMatchGui::StereoMatchGui( const std::string& path, int maxWidth )
{
    m_maxWidth = maxWidth;
    m_path = fs::complete( path ).string();
    m_crossSize = 5;
    m_maxRating = 3;
    m_skipMode = false;
    m_stereoMode = true;
    m_camfilename = "calibration.txt";

    if ( !boost::filesystem::is_directory( path ) )
        throw "Path does not exists '" + path + "'";

    // read in directory content
    std::vector<std::string> tmpfiles;
    std::transform( fs::directory_iterator( path ), fs::directory_iterator(),
                    std::back_inserter( tmpfiles ),
                    pathname_of() );

    // filter for matchfiles only, decide if stereo or interframe mode
    if ( std::count_if( tmpfiles.begin(), tmpfiles.end(), pathname_filter( "*/stereo_match_????.dat" )) > 0 )
    {
        m_stereoMode = true;
        std::cout << "Stereo match mode" << std::endl;
        std::remove_copy_if( tmpfiles.begin(), tmpfiles.end(), std::back_inserter( m_matchfiles ), pathname_filter_neg( "*/stereo_match_????.dat" ) );
    }
    else if ( std::count_if( tmpfiles.begin(), tmpfiles.end(), pathname_filter( "*/inter_frame_match_????_????.dat" )) > 0 )
    {
        m_stereoMode = false;
        std::cout << "Interframe mode" << std::endl;
        std::remove_copy_if( tmpfiles.begin(), tmpfiles.end(), std::back_inserter( m_matchfiles ), pathname_filter_neg( "*/inter_frame_match_????_????.dat" ) );
    }
    else
    {
        std::cout << "No match files found" << std::endl;
        return;
    }



    // sort remaining files
    std::sort( m_matchfiles.begin(), m_matchfiles.end() );

    // initialize camera from file
/*    std::filebuf camfilebuf;
    if ( !camfilebuf.open( fs::complete( m_camfilename, m_path ).string().c_str(), std::ios::in ) )
        throw "Could not open camerafile: " + fs::complete( m_camfilename, m_path ).string();
    std::istream camfile( &camfilebuf );
    m_camera.set_from_stream( camfile, vslam::camera::legacy );*/
    Transform::CalibrationFromFile( fs::complete( m_camfilename, m_path ).string() );
}


StereoMatchGui::~StereoMatchGui()
{
}


void StereoMatchGui::run()
{
    if ( m_matchfiles.size() == 0 )
    {
        std::cout << "No match files found" << std::endl;
        return;
    }

    // setup gui
    cvNamedWindow( "win", CV_WINDOW_AUTOSIZE );


    // main loop, handle image pairs
    size_t idx = 0;
    while (true)
    {
        int res = handleSingleImagePair( idx );
        switch ( res )
        {
        case FIRST:
            idx = 0;
            break;
        case LAST:
            idx = m_matchfiles.size() - 1;
            break;
        case NEXT:
            if ( idx == m_matchfiles.size() - 1 )
                return;
            ++idx;
            break;
        case PREV:
            if ( idx > 0 )
                --idx;
            break;
        case QUIT:
            return;
        }
    }
}


int StereoMatchGui::handleSingleImagePair( size_t idx )
{
    // resturn value
    int result = NEXT;

    try {
        std::cout << "Rate matches from file " << m_matchfiles[idx] << "(" << idx+1 << "/" << m_matchfiles.size() << ")" << std::endl;


        // setup StereoImageMatches
        StereoImageMatches sim( m_matchfiles[idx] );

        // skip to next image if nothing left to do here
        if ( m_skipMode && sim.getNumUnrated() == 0 )
        {
            // do nothing, fall through
        }
        else
        {
            ImageViewer view( sim, m_path, m_maxWidth, "win", Transform::cam );
            if ( !m_stereoMode )
                view.setShowEpilines( false );

            if ( sim.matchCount() == 0 )
            {
                std::cout << "No matches in this image. Skip to next one!" << std::endl;
                return NEXT;
            }

            size_t point_idx = 0;
            bool stop = false;

            // loop to cycle through points
            do
            {
                if ( m_skipMode )
                {
                    while ( sim.getRating(point_idx) != StereoImageMatches::R_UNRATED )
                    {
                        ++point_idx;
                        if ( point_idx == sim.matchCount() )
                        {
                            std::cout << "Skipped to next image" << std::endl;
                            return NEXT;
                        }
                    }
                }

                std::cout  << "Rate match " << point_idx+1 << "/" << sim.matchCount() << " (" << sim.getNumUnrated() << " left unrated)" <<
                        " - Current rating: " << sim.getRatingAsString( point_idx ) << std::endl;

                view.setActiveMatch( point_idx );
                view.show();

                // keyboard readin loop
                while ( true )
                {
                    int keycode = cvWaitKey(0);
                    if ( (unsigned int)keycode == 0xFFFFFFFF )    // window close code
                    {
                        result = QUIT;
                        stop = true;
                        break;
                    }
                    keycode &= 0xFFFF;  // prevent strange keycodes

                    /*
                      Ratings:
                        '1'     close
                        '2'     far
                        '3'     unrelated
                        'u'     unclear
                        'p'     spurious
                        ' '     skip rating

                      Keys
                        left/right arrow        match nav
                        page up/down            image nav
                        esc/q                   quit

                        's'         skip mode
                        'c'         toggle circle
                        'e'         toggle epilines
                        'm'         toggle minimal display
                        'h'         print help on console
                    */

                    if ( (keycode >= '1' && keycode <= '3') || keycode == 'u' || keycode == 'p' || keycode == ' ' )  // rating || skip -> next pair
                    {
                        int rating_sym = 0;
                        switch (keycode)
                        {
                        case '1': rating_sym = StereoImageMatches::R_CLOSE; break;
                        case '2': rating_sym = StereoImageMatches::R_FAR; break;
                        case '3': rating_sym = StereoImageMatches::R_UNRELATED; break;
/*                        case '4': rating_sym = StereoImageMatches::R_100P; break;
                        case '5': rating_sym = StereoImageMatches::R_OVER100; break;*/
                        case 'u': rating_sym = StereoImageMatches::R_UNCLEAR; break;
                        case 'p': rating_sym = StereoImageMatches::R_SPURIOUS; break;
                        }

                        if ( keycode != ' ' )   // dont change
                            sim.setRating( point_idx, rating_sym );
                        ++point_idx;

                        if ( point_idx >= sim.matchCount() )
                        {
                            result = NEXT;
                            stop = true;
                        }

                        if ( view.getShowMinimal() )
                        {
                            //view.setShowCircles( true );
                            if ( m_stereoMode )
                                view.setShowEpilines( true );
                        }

                        break;
                    }

                    else if ( keycode == 65361 )    // left arrow -> nav through pairs
                    {
                        point_idx = (point_idx + sim.matchCount() - 1) % sim.matchCount();
                        break;
                    }

                    else if ( keycode == 65363 )        // right arrow -> nav through pairs
                    {
                        point_idx = (point_idx + 1) % sim.matchCount();
                        break;
                    }

                    else if ( keycode == 65365 || keycode == 65366 )    // page up || page down  -> switch images
                    {
                        result = keycode == 65365 ? PREV : NEXT;
                        stop = true;
                        break;
                    }

                    else if ( keycode == 27 || keycode == 'q' )   // ESC or quit
                    {
                        result = QUIT;
                        stop = true;
                        break;
                    }

                    else if ( keycode == 's' )      // toggle skipmode
                    {
                        m_skipMode = !m_skipMode;
                        std::cout << "Skipmode set to " << (m_skipMode ? "ON" : "OFF") << std::endl;
                        if ( m_skipMode && sim.getNumUnrated() == 0 )
                        {
                            result = NEXT;
                            stop = true;
                            break;
                        }
                        break;
                    }

                    else if ( keycode == 'c' )       // toggle circle
                    {
                        view.setShowCircles( !view.getShowCircles() );
                        break;
                    }

                    else if ( keycode == 'e' )       // toggle epilines
                    {
                        if ( m_stereoMode )
                        {
                            view.setShowEpilines( !view.getShowEpilines() );
                            break;
                        }
                    }

                    else if ( keycode == 'm' )       // toggle minimal
                    {
                        view.setShowMinimal( !view.getShowMinimal() );
                        // disable/enable epilines and circles if minimal on/off
                        view.setShowCircles( !view.getShowMinimal() );
                        view.setShowEpilines( !view.getShowMinimal() );
                        break;
                    }

                    else if ( keycode == 'x' )
                    {
                        view.setShowLargeX( !view.getShowLargeX() );
                        break;
                    }

                    else if ( keycode == 'h' )       // show help on console
                    {
                        printKeyHints();
                    }

                }

            } while ( !stop );

            sim.writeRating();
        }

    }
    catch ( const char* err )
    {
        std::cerr << "Exception on '" << m_matchfiles[idx] << "': " << err << std::endl;
    }
    catch ( std::string err )
    {
        std::cerr << "Exception on '" << m_matchfiles[idx] << "': " << err << std::endl;
    }

    return result;
}


void StereoMatchGui::printKeyHints()
{
    std::cout << "Keyboard shortcuts:" << std::endl;
    std::cout << "   1, 2, 3             Rating relative to fixed-size circle (inside inner circle, outer circle, outside both )" << std::endl;
    std::cout << "   u                   Mark match as 'unclear'" << std::endl;
    std::cout << "   p                   Mark match as 'spurious'" << std::endl;
    std::cout << "   space               Skip match, won't change rating" << std::endl;
    std::cout << std::endl;
    std::cout << "   s                   Toggle skipmode (autoskip fully rated images)" << std::endl;
    std::cout << "   c                   Toggle size of feature circles" << std::endl;
    std::cout << "   e                   Toggle epilines" << std::endl;
    std::cout << "   m                   Toggle minimal display" << std::endl;
    std::cout << "   x                   Toggle large X for active" << std::endl;
    std::cout << std::endl;
    std::cout << "   left/right arrow    Navigate through matches" << std::endl;
    std::cout << "   page up/down        Navigate through images" << std::endl;
    std::cout << "   esc, q              exit program" << std::endl;
}

