/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CALIBRATOR_HPP_
#define CALIBRATOR_HPP_

#include "camera.hpp"

#include <slom/Estimator.hpp>
#include <slom/CallBack.hpp>
#include <slom/BuildMeasurement.hpp>

#include <mtk/types/SOn.hpp>


#include <Eigen/LU>

namespace vslam {

class calibrator
{
	// FIXME public just for debugging
public:
	SLOM::Estimator e_;
	SLOM::CallBack cb_;
	
	// FIXME replace pose_as_matrix by a Quaternion/vector representation
	typedef pose_as_matrix pose;
	typedef MTK::vect<3, double> vec3;
	typedef MTK::SO3<double> SO3;
	
	typedef MTK::Scalar<double> scalar;
	SLOM_RANDOMVAR_ID(stereo_camera);
	SLOM_RANDOMVAR_ID(pose);
	SLOM_RANDOMVAR_ID(vec3);
	SLOM_RANDOMVAR_ID(SO3);
	
	SLOM_RANDOMVAR_ID(scalar);
	
	typedef camera::image_point image_point;
	typedef std::vector<image_point> image_points;
	
	stereo_camera_id cam_;
	SO3_id left2imu_;
	scalar_id gravity_;
	vec3_id acc_offset_;
	
	pose_id star2left_;
	pose_id checker2world_;
	
	std::vector<vec3> world_points_;
	CvMat world_p_;
	std::vector<pose_id> poses_;
	
	
	
	SLOM_BUILD_MEASUREMENT(acc_measurement, 3, ((pose, pose_)) ((SO3, left2imu)) ((scalar, g)) ((vec3, acc_offset)), ((vec3, acc)) ((double, acc_sigma)))

	SLOM_BUILD_MEASUREMENT(point_measurement, 4, ((pose, pose_)) ((stereo_camera, cam_)), 
	                       ((image_point, left_)) ((image_point, right_)) ((vec3, world_)) ((double, sigma_)))

	SLOM_BUILD_MEASUREMENT_JACOBIAN(groundtruth_measurement, pose::DOF, ((pose, pose_)) ((pose, star2left)) ((pose, checker2world)), 
	                       ((pose, star2world)) ((double, sigma)))

	std::vector<SLOM::MeasID<groundtruth_measurement> > gt_measurements_;

public:
	
	calibrator() : cb_(e_, 2)
	{
		e_.setCallBack(cb_);
	}
	
	// Minimal initialization (no initial guesses)
	void initialize(double gridwidth, int rows, int cols)
	{
		world_points_.clear();
		world_points_.reserve(rows*cols);
		for(int x=0; x<rows; ++x)
			for(int y=0; y<cols; ++y)
			{
				world_points_.push_back(Eigen::Vector3d(x*gridwidth, -y*gridwidth, 0));
			}
		world_p_ = cvMat(world_points_.size(), 3, CV_64F, world_points_.data());
	}
	
	
	void initialize(double gridwidth, int rows, int cols, const stereo_camera& inital_guess, const pose &left2imu_guess, bool fix_transformations)
	{
		initialize(gridwidth, rows, cols);
		cam_ = e_.insertRV(inital_guess, !fix_transformations);
		left2imu_  = e_.insertRV(SO3(Eigen::Quaterniond(left2imu_guess.get_rotation())), !fix_transformations);
		gravity_ = e_.insertRV(scalar(9.81));
		acc_offset_ = e_.insertRV(vec3());
	}

	
	pose find_extrinsics(const camera& cam, const image_points& points)
	{
		cv_mat<double, 3, 1> rotation, translation;

		CvMat points_cv  = cvMat(points.size(), 2, CV_32F, (void*)points.data());

		const CvMat &cam_mat = cam.get_camera_matrix();
		const CvMat &dist_coeff = cam.get_dist_coeffs();
		
		cvFindExtrinsicCameraParams2(&world_p_, &points_cv, &cam_mat, &dist_coeff, &rotation, &translation);
		
		return pose(SO3::exp(rotation), translation);
	}
	
	void next_frame(const image_points &left,
	                const image_points &right,
	                double im_sigma)
	{
		assert(world_points_.size() == left.size() && left.size() == right.size());
		if(!cam_)
		{
			// TODO initialize intrinsics
		}
		const camera &left_cam = cam_->get_cam(true);

		pose world2left_guess = find_extrinsics(left_cam, left);
		
		if(true)
		{
			pose world2right_guess = find_extrinsics(cam_->get_cam(false), right);
			std::cout << "w2l: " << world2left_guess << ",\nw2r: " << world2right_guess << ",\nl2r: " << (world2right_guess * world2left_guess.inverse()) << std::endl;
		}

		pose_id cur_pose = e_.insertRV(world2left_guess.inverse());
		poses_.push_back(cur_pose);
		for(size_t i=0; i<world_points_.size(); ++i)
		{
			SLOM::MeasID<point_measurement> id = 
					e_.insertMeasurement(point_measurement(cur_pose, cam_, left[i], right[i], world_points_[i], im_sigma));
		}
	}
	
	void add_frame(const vec3& acc, double acc_sigma, 
	               const image_points &left,
	               const image_points &right,
	               double im_sigma)
	{
		next_frame(left, right, im_sigma);
		add_gravity(acc, acc_sigma);
	}
	
	void add_gravity(const vec3& acc, double sigma)
	{
		e_.insertMeasurement(acc_measurement(poses_.back(), left2imu_, gravity_, acc_offset_, acc, sigma));
	}
	
	void add_groundtruth(const pose& gt_pose, double sigma)
	{
		if(!star2left_) // first groundtruth
		{
			star2left_ = e_.insertRV(pose(), true); // temporarily store gt_measurement
			checker2world_ = e_.insertRV(pose(), true);
		}
		
		gt_measurements_.push_back(
				e_.insertMeasurement(
						groundtruth_measurement(poses_.back(), star2left_, checker2world_, 
						                        gt_pose, sigma)
				)
		);
	}
	
	void initialize_groundtruth()
	{
		assert(gt_measurements_.size() >= 3 && "Need at least 3 groundtruth measurements!");
		
		// Implementation based on 
		//   Park & Martin: "Robot Sensor Calibration: Solving AX = XB on the Euclidean Group"
		// TODO this is not robust at all!
		// Actually this just initializes the orientation, translational parts are linear and thus easily determined by overall estimation
		
		
		SO3 p_q[3], g_q[3];
		Eigen::Vector3d p_t[3], g_t[3];
		Eigen::Matrix3d P_, G_;
		
		for(int i=0; i<3; ++i)
		{
			const groundtruth_measurement &m = *gt_measurements_[i];
			p_q[i] = SO3(m.pose_->get_rotation());
			p_t[i] = (m.pose_->get_translation());
			g_q[i] = SO3(m.star2world.get_rotation());
			g_t[i] = (m.star2world.get_translation());
		}
		for(int i=1; i<3; ++i)
		{
			P_.col(i-1) = SO3::log(p_q[0].conjugate() * p_q[i]);
			G_.col(i-1) = SO3::log(g_q[0].conjugate() * g_q[i]);
		}
		P_.col(2) = P_.col(0).cross(P_.col(1));
		G_.col(2) = G_.col(0).cross(G_.col(1));
		
		Eigen::Matrix3d X = P_ * G_.inverse();
		Eigen::Quaterniond q_s2c(X);
		
		std::cout << X << std::endl << q_s2c.toRotationMatrix() << std::endl;
		
		std::cout << (p_q[0].conjugate() * p_q[1] * q_s2c).coeffs().transpose() << "\n" 
		          << (q_s2c * g_q[0].conjugate() * g_q[1]).coeffs().transpose() << std::endl;
		
		SO3 q_c2w = g_q[0] * q_s2c.conjugate() * p_q[0].conjugate();
		
		for(int i=0; i<3; ++i)
			std::cout << (q_c2w * p_q[i] * q_s2c).coeffs().transpose() << " "
			          << (g_q[i] ).coeffs().transpose() << std::endl;
		
		Eigen::Vector3d w_Rv[2];
		for(int i=0; i<2; ++i)
			w_Rv[i] = g_t[i] - q_c2w * p_t[i];
		
		analyze(star2left_);
		e_.reinitRV(star2left_, pose(q_s2c, Eigen::Vector3d::Zero()));
		e_.reinitRV(checker2world_, pose(q_c2w, Eigen::Vector3d::Zero()));
		analyze(star2left_);
	}
	
	void optimize() 
	{
		if(gt_measurements_.size() > 0)
		{
			initialize_groundtruth();
			analyze();
		}
		
		e_.changeAlgorithm(SLOM::Estimator::Levenberg, 1e-9);
		for(int i=0; i<50; ++i)
		{
			double gain = e_.optimizeStep();
			if(-1e-10 < gain && gain < 1e-5) break;
		}
	}
	
	void get_result(vslam::stereo_camera& cam, pose &left2imu)
	{
		cam = *cam_;
		left2imu.get_rotation() = left2imu_->toRotationMatrix();
	}
	
	void get_result(pose &star2left, pose &checker2world)
	{
		star2left = *star2left_;
		checker2world = *checker2world_;
	}
	
	template<class VarID>
	void analyze(VarID id) const
	{
		double rss;
		int n = e_.getRSS(rss, id);
		std::cout << "rss: " << rss << ", n: " << n << ", rms: " << std::sqrt(rss/n) << std::endl;
		
	}
	
	void analyze(bool poses = false) const
	{
		std::cout << "Pixel: ";
		analyze(cam_);
		std::cout << "Gravitation: (" << *gravity_ << ", " << acc_offset_->transpose() << ") ";
		analyze(left2imu_);
		if(star2left_)
		{
			std::cout << "Groundtruth: ";
			analyze(star2left_);
		}
		if(!poses)
			return;
		
		std::cout << "Poses:\n";
		for(std::vector<pose_id>::const_iterator it = poses_.begin(); it != poses_.end(); ++it)
		{
			analyze(*it);
		}
	}
	
};

} // namespace vslam

#endif /* CALIBRATOR_HPP_ */
