/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CV_CAMERA_HPP_
#define CV_CAMERA_HPP_

#include <Eigen/Core>
#include <opencv/cv.h>

#ifndef NO_MTK_SUPPORT
#include <mtk/types/SOn.hpp>
#endif //NO_MTK_SUPPORT

#include "../util/wrapped_cv_mat.hpp"

#include "../core/pose_as_matrix.hpp"

namespace vslam {

/**
 * camera wraps an OpenCV camera
 */
class camera
{
	enum {CV_FTYPE = CV_64F};
	typedef double f_type;
	
	typedef cv_mat<double, 3, 3> mat33_t;
	mat33_t camera_matrix_;
	typedef cv_mat<double, 5, 1> coeffs_type;
	coeffs_type dist_coeffs_;
	cv::Size size_;
	
	// normal vectors of the planes bounding the visible area:
	Eigen::Vector3d bound_normals[4];
	
	
public:
	enum file_format
	{
		legacy = 0,
		cv_cam,
	};
	
	typedef pose_as_matrix Transform;
	typedef pose_as_matrix::vec Vector;
	typedef cv::Point2f image_point;

	camera() 
		: camera_matrix_(mat33_t::Identity()),
		  dist_coeffs_(coeffs_type::Zero()),
		  size_(0,0)
	{
	}

	void set_parameter(int width, int height, double f_x, double f_y, double c_x, double c_y)
	{
		size_.width  = width;
		size_.height = height;
		camera_matrix_(0,0) = f_x;
		camera_matrix_(1,1) = f_y;
		camera_matrix_(0,2) = c_x;
		camera_matrix_(1,2) = c_y;
		compute_bound_normals();
	}
	
	void set_distortion(double k1, double k2, double p1, double p2, double k3=0)
	{
		dist_coeffs_ << k1, k2, p1, p2, k3;
		compute_bound_normals();
	}
	void set_from_stream(std::istream &strm, file_format format)
	{
		double k1, k2, p1, p2, k3;
		int width, height;
		double f_x, f_y;
		double c_x, c_y;

		strm >> width >> height >>  f_x >> f_y >> c_x >> c_y >> k1 >> k2 >> p1 >> p2 >> k3;
		set_parameter (width, height, f_x, f_y, c_x, c_y);
		switch(format)
		{
		case legacy:
			// this is just an approximation, and actually only works for some inputs ...
			set_distortion(k2, k2*k2, 0, 0, k2*k2*k2);
			break;
		case cv_cam:
			set_distortion(k1, k2, p1, p2, k3);
			break;
		default:
			throw "Unknown calibration file format";
		}
	}
	void to_stream(std::ostream &ss) const
	{
		ss << size_.width << " " << size_.height << " " 
		   << camera_matrix_(0,0) << " " << camera_matrix_(1,1) << " " 
		   << camera_matrix_(0,2) << " " << camera_matrix_(1,2);
		// Eigen's default operator<< overrides precision settings ...
		for(int i=0; i<dist_coeffs_.size(); ++i)
			ss << " " << dist_coeffs_[i];
		ss << "\n";
	}

	void point2normalized(const image_point &p_im, Eigen::Vector2d &normalized, const CvMat* rot = 0, const CvMat* proj = 0) const
	{
		CvMat pt = cvMat(1,1, CV_32FC2, (void*)&p_im);
		CvMat res = cvMat(1,1, CV_64FC2, normalized.data());
		cvUndistortPoints(&pt, &res, &camera_matrix_, &dist_coeffs_, rot, proj);
	}
	
	double keypoint2Normalized(const cv::KeyPoint &kp, Eigen::Vector2d &normalized) const
	{
		point2normalized(kp.pt, normalized);
		double scale = camera_matrix_.diagonal().start<2>().sum() / 2;
		return kp.size / scale;
	}

	double dist2(const Vector &p_sensor, const image_point &p_im) const
	{
		image_point p_project;
		bool visible = sensor2ImageClipped(p_sensor, p_project);
		if(!visible)
			return 1e10;
		p_project -= p_im;
		return p_project.dot(p_project);
	}

	void getParameter(int &width, int &height, double &f_x, double &f_y, double &c_x, double &c_y) const
	{
		width = size_.width;
		height = size_.height;
		f_x = camera_matrix_(0,0);
		f_y = camera_matrix_(1,1);
		c_x = camera_matrix_(0,2);
		c_y = camera_matrix_(1,2);
	}

	void get_distortion(double& k1, double& k2, double& p1, double& p2, double& k3) const
	{
		k1 = dist_coeffs_(0);
		k2 = dist_coeffs_(1);
		p1 = dist_coeffs_(2);
		p2 = dist_coeffs_(3);
		k3 = dist_coeffs_(4);
	}

	bool sensor2ImageClipped(const Vector& v, image_point &pt) const
	{
		CvMat object_points = cvMat( 1, 1, CV_64FC3, (void*)&v );
		double _zero[] = {0,0,0};
		CvMat zero = cvMat( 3, 1, CV_64FC1, _zero );
		CvMat result = cvMat(1,1, CV_32FC2, &pt);
		cvProjectPoints2(&object_points, &zero, &zero, &camera_matrix_, &dist_coeffs_, &result);
		return v[2] > 0 && visible(pt);
	}
	
	bool visible(const image_point &pt) const
	{
		return (0<=pt.x && pt.x <= size_.width && 0<=pt.y && pt.y<=size_.height);
	}
private:
	void compute_bound_normals()
	{
		// normalize the corner points, and calculate normal vectors through each adjacent pair
		Eigen::Vector2d n[4];
		image_point test(0,0);
		point2normalized(test, n[0]);
		test.x = size_.width;
		point2normalized(test, n[1]);
		test.y = size_.height;
		point2normalized(test, n[2]);
		test.x = 0;
		point2normalized(test, n[3]);
		
		for(int i=0; i<4; ++i)
		{
			double x1=n[i].x(), y1=n[i].y();
			double x2=n[(i+1)%4].x(), y2 = n[(i+1)%4].y();
			bound_normals[i] = Vector(y1 - y2, x2 - x1, x1*y2 - x2*y1);
		}
	}
	
	void intersect_line_with_plane(const Vector &p, const Vector &v, const Vector &n, double &l_min, double &l_max) const
	{
		double vn = v.dot(n);
		double pn = p.dot(n);
		if (vn>1e-3)
		{
			double ll = -pn/vn;
			if (ll>l_min) l_min = ll;
		}
		else if (vn<-1e-3)
		{
			double ll = -pn/vn;
			if (ll<l_max) l_max = ll;
		}
		else if (pn<0 || true) {
			l_min =  1/.0;
			l_max = -1/.0;
		}
	}
public:
	void sensorVisibleIntervalOfLine (const Vector& p, const Vector& v, double& l_min, double& l_max, bool useDistortion, bool intersectWithPassed) const
	{
		if(!intersectWithPassed)
		{
			l_min = -1/.0;
			l_max = +1/.0;
		}
		if(!useDistortion)
		{
			double f_x = camera_matrix_(0,0), f_y = camera_matrix_(1,1);
			double c_x = camera_matrix_(0,2), c_y = camera_matrix_(1,2);
			
			intersect_line_with_plane (p, v, Vector( f_x, 0.0, c_x), l_min, l_max);
			intersect_line_with_plane (p, v, Vector(-f_x, 0.0, (size_.width-c_x)), l_min, l_max);
			intersect_line_with_plane (p, v, Vector(0.0,  f_y, c_y), l_min, l_max);
			intersect_line_with_plane (p, v, Vector(0.0, -f_y, (size_.height-c_y)), l_min, l_max);
		}
		else
		{
			for(int i=0; i<4; ++i)
			{
				intersect_line_with_plane(p, v, bound_normals[i], l_min, l_max);
			}
		}
	}


	const CvMat& get_camera_matrix() const
	{
		return *&camera_matrix_;
	}
	
	const CvMat& get_dist_coeffs() const
	{
		return *&dist_coeffs_;
	}
	
	const cv::Size& get_size() const
	{
		return size_;
	}
	
#ifndef NO_MTK_SUPPORT
	/* ******************************
	 * Calibration/SLoM interface
	 * ******************************/
	enum {DOF = 2 + 2 + 5};
	
	void boxplus(const MTK::vectview<const double, DOF> &vec, double scale=1)
	{
		f_type scale_(scale);
		camera_matrix_.diagonal().start<2>() += scale_ * vec.start<2>().cast<f_type>();
		camera_matrix_.block<2,1>(0,2) += scale_ * vec.segment<2>(2).cast<f_type>();
		dist_coeffs_ += scale_ * vec.end<5>().cast<f_type>();
	}
#endif //NO_MTK_SUPPORT

private:
	// These methods are only accessed by stereo_camera
	friend class stereo_camera;
	friend class rectified_stereo_camera;
};



class stereo_camera
{
protected:
	camera left_;
	camera right_;
	camera::Transform left2right_;

public:
	
	stereo_camera() { }
	
	stereo_camera(const camera &left, const camera &right, const camera::Transform &left2right)
		: left_(left), right_(right), left2right_(left2right)
	{ }
	
	void set_from_stream(std::istream &strm, camera::file_format format)
	{
		left_. set_from_stream(strm, format);
		right_.set_from_stream(strm, format);
		strm >> left2right_;

		left2right_.normalize();
	}

	const vslam::camera& get_cam(bool left) const
	{
		return left ? left_ : right_;
	}
	const vslam::camera::Transform& get_left2right() const
	{
		return left2right_;
	}

	double max_dist2(const camera::Vector &lm_in_left_cam, const camera::image_point &left, const camera::image_point &right) const
	{
		double dist_l = left_ .dist2(lm_in_left_cam, left);
		double dist_r = right_.dist2(left2right_ * lm_in_left_cam, right);

		return std::max(dist_l, dist_r);
	}

	void left2images(const camera::Vector &lm_in_left_cam, camera::image_point &left, camera::image_point &right) const
	{
		left_.sensor2ImageClipped(lm_in_left_cam, left);
		camera::Vector lm_in_right_cam = left2right_ * lm_in_left_cam;
		right_.sensor2ImageClipped(lm_in_right_cam, right);
	}
	
	/**
	 * If left_cam == true, this method computes for a point pt in the left
	 * image a sequence of points in the right image, marking its epiline.
	 * For left_cam == false, left and right are reversed.
	 * max_dist is the maximal distance of the endpoint of the line (+inf is allowed),
	 * num_points is the number of points to be calculated.
	 * 
	 * If the point can't be seen in the other image, false is returned and no data is modified.
	 */
	bool get_epiline(bool left_cam, const camera::image_point& pt,
	                 std::vector<camera::image_point>& result, double max_dist = 1/.0, int num_points = 16) const
	{
		assert(num_points > 1 && "You need at least 2 points for a line!");
		const camera &here = get_cam(left_cam), &there=get_cam(!left_cam);
		camera::Transform here2there = left_cam ? left2right_ : left2right_.inverse();
		
		Eigen::Vector2d normalized;
		here.point2normalized(pt, normalized);
		
		Eigen::Vector3d start = here2there.get_translation();
		Eigen::Vector3d dir(normalized.x(), normalized.y(), 1);
		dir = here2there.get_rotation() * dir.normalized();
		
		double l_min = 0, l_max = max_dist;
		
		there.sensorVisibleIntervalOfLine(start, dir, l_min, l_max, true, true);
		
		if(l_min > l_max)
			return false;
		
		double step = (std::atan(l_max) - std::atan(l_min)) / (num_points - 1);
		
		result.resize(num_points);
		for(int i=0; i<num_points; ++i)
			there.sensor2ImageClipped(start + std::tan(l_min + i*step)*dir, result[i]);
		
		return true;
	}

	
#ifndef NO_MTK_SUPPORT
	// Interface for MTK:
	enum {DOF = 2*camera::DOF + pose_as_matrix::DOF};
	
	void boxplus(const MTK::vectview<const double, DOF> &vec, double scale=1)
	{
		left_.boxplus(vec.start<camera::DOF>().data(), scale);
		right_.boxplus(vec.segment<camera::DOF>(camera::DOF).data(), scale);
		left2right_.boxplus(vec.segment<pose_as_matrix::DOF>(2*camera::DOF).data(), scale);
	}
#endif //NO_MTK_SUPPORT
	
};


} // namespace vslam

#endif /* SM_CAMERA_HPP_ */
