/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef RECTIFIED_STEREO_CAMERA_HPP_
#define RECTIFIED_STEREO_CAMERA_HPP_


#include "cv_camera.hpp"

namespace vslam {

// FIXME this could be renamed to stereo_camera, and stereo_camera renamed to sth else ...

class rectified_stereo_camera : public stereo_camera
{
	typedef camera::mat33_t mat33_t;
	
	mat33_t fundamental_;
	
	CvSize size_; // Size of the rectified image
	
	// Results of cvStereoRectify:
	mat33_t R1_, R2_;
	
	cv_mat<double, 3, 4> P1_, P2_;
	
	cv_mat<double, 4,4> Q_;
	
	CvMat map_lx, map_ly, map_rx, map_ry;
	
	void update_fundamental()
	{
		// Computation due to Hartley/Zisserman formula (9.2), p.244
		for(int i=0; i<3; ++i)
		{
			// F = [t]_x * R
			fundamental_.col(i) = left2right_.get_translation().cross(left2right_.get_rotation().col(i));
		}
		left_.camera_matrix_.transpose().part<Eigen::UpperTriangular>().solveTriangularInPlace(fundamental_.transpose());
		right_.camera_matrix_.transpose().part<Eigen::UpperTriangular>().solveTriangularInPlace(fundamental_);
	}
	
	void update_stereo_map()
	{
		CvMat mat_rot, mat_trans;
		cvInitMatHeader(&mat_rot, 3, 3, camera::CV_FTYPE, (void*)left2right_.get_rotation().data(), 4*CV_ELEM_SIZE(camera::CV_FTYPE));
		cvInitMatHeader(&mat_trans, 3, 1, camera::CV_FTYPE, (void*)left2right_.get_translation().data(), 4*CV_ELEM_SIZE(camera::CV_FTYPE));
		cvStereoRectify(
				&left_.get_camera_matrix(), &right_.get_camera_matrix(), 
				&left_.get_dist_coeffs(), &right_.get_dist_coeffs(), 
				left_.get_size(), &mat_rot, &mat_trans, 
				&R1_, &R2_, &P1_, &P2_, &Q_);
//		, cv::CALIB_ZERO_DISPARITY, -1, size_);
	}
	
	void update_rectification()
	{
		update_fundamental();
		update_stereo_map();
	}
	
public:
	rectified_stereo_camera(const stereo_camera &cam, const cv::Size &new_size = cv::Size()) 
			: stereo_camera(cam), size_(new_size)
	{
		update_rectification();
	}
	
	rectified_stereo_camera& operator=(const stereo_camera & cam)
	{
		stereo_camera::operator=(cam);
		update_rectification();
		return *this;
	}

	const CvMat& get_fundamental_matrix() const
	{
		return *&fundamental_;
	}
	
	camera::Vector stereo_image2world_point(const camera::image_point &left, const camera::image_point &right, double &error) const
	{
		Eigen::Vector2d left_n, right_n;
		left_.point2normalized(left, left_n, &R1_, &P1_);
		right_.point2normalized(right, right_n, &R2_, &P2_);

		camera::Vector result(left_n[0], .5*(left_n[1] + right_n[1]), right_n[0]-left_n[0]);
		if(result[2] > 0 && false) 
		{
			std::cout << " !!! negative disparity !!!" << std::endl;
			error = 1e6;
			return result;
		}
		CvMat point = cvMat( 1, 1, CV_64FC3, result.data());

		cvPerspectiveTransform(&point, &point, &Q_);

		camera::image_point left_back, right_back;
		left2images(R1_.transpose() * result, left_back, right_back);

		error = cv::norm(left_back - left) + cv::norm(right_back - right);

		return R1_.transpose() * result;
	}

};

} // namespace vslam

#endif /* RECTIFIED_STEREO_CAMERA_HPP_ */
