/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include <boost/thread.hpp>

#include <slom/m_estimate.hpp>
#include <mtk/startIdx.hpp>

#include "optimizer.hpp"
#include "../framework/program_config.h"

namespace vslam {



optimizer::optimizer(const pose_as_matrix& imu2cam_left,
                     const pose_as_matrix& imu2cam_right,
                     int verbosity_level, double velocity_sigma)
	: current_pose_valid_(false),
	  landmarks(0), stereo_measurements(0), mono_measurements(0),
	  e(SLOM::Estimator::Levenberg, 10), // TODO adapt initialization parameters
	  cb(e, verbosity_level),
	  verbosity_level(verbosity_level),
	  poses_back_valid(true),
	  needs_reinitialization(true)
{
	e.setCallBack(cb);
	// TODO: initialize by constructor:
	camera_conf_.pose2left = imu2cam_left;
	camera_conf_.pose2right = imu2cam_right;
	camera_conf_.m_bound = 1;
	vel_sigma = velocity_sigma;

	imu_conf_.acc_sigma = 0.02;
	imu_conf_.gyro_sigma = 2*M_PI / 360.0;
}


void optimizer::init(const Eigen::Vector3d &init_pos, 
          const Eigen::Vector3d &init_gravity)
{
	bool optimize_first_pose = false; // decide if orientation of first pose shall be optimized

	boost::mutex::scoped_lock lock(poses_mutex);
	
	// calculate and insert start pose
	poses.push_back(e.insertRV(Pose(orientation_from_gravity(init_gravity), init_pos), optimize_first_pose));
	
	imu_conf_.gravity = init_gravity.norm();
	
	if(optimize_first_pose)
	{
		e.insertMeasurement(start_pose(poses.back(), init_pos));
	}
	current_pose_valid_ = false;
}


bool optimizer::next_frame()
{
	// if last pose is valid, 
	if(poses_back_valid){
		return false;
	}

	boost::mutex::scoped_lock lock(poses_mutex);

	// otherwise add new pose:
	Pose_id lastPose = poses.back();
	poses.push_back(e.insertRV(Pose(last_imu.integrate_pose(*lastPose, imu_conf_.gravity))));
	e.insertMeasurement(IMU_measurement(lastPose, poses.back(), last_imu, imu_conf_));
	if(vel_sigma>0)
	{
		e.insertMeasurement(small_speed_constraint(poses.back(), vel_sigma));
	}


	needs_reinitialization = true;
	poses_back_valid=true;
	last_imu.reset();
	return true;
}


void optimizer::add_landmark_measurement(const Eigen::Vector2d &p_left_normalized, const double &sigma_left,
                                         const Eigen::Vector2d &p_right_normalized, const double &sigma_right, 
                                         const vslam::map::landmark &m)
{
	next_frame();
	++stereo_measurements;
	e.insertMeasurement(
			stereo_measurement(poses.back(), m.position, 
			                   p_left_normalized, sigma_left, 
			                   p_right_normalized, sigma_right, 
			                   camera_conf_));
	needs_reinitialization = true;
}

void optimizer::add_mono_landmark_measurement(const Eigen::Vector2d &p_cam_normalized, const double &sigma,
                                              const vslam::map::landmark &m, bool is_left)
{
	next_frame();
	const pose_as_matrix& pose2cam = is_left ? camera_conf_.pose2left : camera_conf_.pose2right;
	++mono_measurements;
	e.insertMeasurement(
			mono_measurement(poses.back(), m.position, 
			                 p_cam_normalized, sigma, 
			                 pose2cam, camera_conf_.m_bound));
	needs_reinitialization = true;
}


void optimizer::step(int max_steps)
{
	boost::mutex::scoped_lock lock(big_lock_mutex);
	if(verbosity_level >= 1)
		fprintf(stderr, "optimizer: %i Poses, %i Landmarks, %i Stereo, %i Mono\n", 
				poses.size(), landmarks,
				stereo_measurements, mono_measurements);
	if(needs_reinitialization){
		if(poses.size() < 5 || stereo_measurements < 50)
		{
			e.changeAlgorithm(SLOM::Estimator::Levenberg, 10);
			max_steps = 8;
		}
		else
		{
			e.changeAlgorithm(SLOM::Estimator::GaussNewton);
		}
		needs_reinitialization = false;
	}
	for(int i=0; i<max_steps; ++i){
		try {
			double gain = e.optimizeStep();
			if(gain > -1e-10 && gain < 1e-6) break;
		} catch(const char* s) {
			std::cerr << s << "\n *** Trying to optimize using Levenberg ***" << std::endl;
			e.changeAlgorithm(SLOM::Estimator::Levenberg, 100);
			max_steps = 4;
		}
	}
	if(verbosity_level > 1)
	{
		double rss;
		int n = e.getRSS<stereo_measurement>(rss);
		fprintf(stderr, "stereo_measurement RSS: %g, RMS: %g\n",
		        rss, std::sqrt(rss/n));
	}
	current_pose_valid_ = false;
}


////// Measurements //////



/**
 * Applies the perspective of a normalized camera,
 * assuming camera looks in direction (0,0,1).
 */
static inline vec2 apply_perspective(const vec3 &point){
	vec2 res;
	for(int i=0; i<2; ++i){
		res[i] = point[i]/point[2];
	}
	return res;
}

void optimizer::stereo_measurement::eval(MTK::vectview<double, 4> ret, bool numerical_jacobian) const {
	vec3 lm_local = pose->orient.conjugate() * (*lm - pose->pos);

	vec3 lm_left  = conf.pose2left.apply(lm_local);
	vec3 lm_right = conf.pose2right.apply(lm_local);

	ret.start<2>() = (apply_perspective(lm_left)  - coord_left ) / sigma_left;
	ret.end<2>()   = (apply_perspective(lm_right) - coord_right) / sigma_right;
	if(!numerical_jacobian)
	{
		SLOM::m_estimate(ret.data(),  ret.start<2>(), conf.m_bound);
		SLOM::m_estimate(ret.data()+2,ret.end<2>() ,  conf.m_bound);
	}
}

void optimizer::mono_measurement::eval(MTK::vectview<double, 2> ret, bool numerical_jacobian) const {
	vec3 lm_local = pose->orient.conjugate() * (*lm - pose->pos);

	vec3 lm_cam = pose2cam.apply(lm_local);
	

	ret = (apply_perspective(lm_cam) - coord ) / sigma;
	if(!numerical_jacobian)
	{
		SLOM::m_estimate(ret.data(), ret, m_bound);
	}
}

void optimizer::IMU_measurement::eval(MTK::vectview<double,Pose::DOF> ret) const {
	Pose endpose = delta.integrate_pose(*x0, imu.gravity);
	endpose.boxminus(ret, *x1);
	// TODO accumulate covariance within imu_accumulator
	double dt = delta.dt;
	double gyro_scale = imu.gyro_sigma * sqrt(dt);
	double vel_scale = imu.acc_sigma*sqrt(dt);
	MTK::subvector(ret, &Pose::orient) /= gyro_scale;
	MTK::subvector(ret, &Pose::vel)    /= vel_scale;
	MTK::subvector(ret, &Pose::pos)    /= (vel_scale * dt);
}

void optimizer::small_speed_constraint::eval(MTK::vectview<double, 3> ret) const
{
	ret = pose->vel / vel_sigma;
}

void optimizer::start_pose::eval(MTK::vectview<double, DIM> ret) const 
{
	ret.start<3>() = 1000 * (pose->pos - start);
	
	// Start velocity is approximately zero:
	ret.segment<3>(3) = 100 * pose->vel;
	
	// Fix start direction:
	ret(6) = 1000 * (pose->orient * Eigen::Vector3d(1,0,0))[1];
}

} // namespace vslam

