/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_CORE_OPTIMIZER_HPP__
#define __VSLAM_CORE_OPTIMIZER_HPP__

#include <deque>
#include <iostream>


#include "../slom_variables.hpp"
#include "pose_as_matrix.hpp"
#include "../map/landmark.hpp"

#include <slom/BuildMeasurement.hpp>
#include <slom/Estimator.hpp>
#include <slom/CallBack.hpp>

#include <boost/shared_ptr.hpp>
#include <boost/thread.hpp>

class program_config;

namespace vslam {


class optimizer
{
	struct camera_config
	{
		//! transformations from IMU-pose to left and right camera
		pose_as_matrix pose2left;
		pose_as_matrix pose2right;
		
		//! bound for m-estimates
		double m_bound;
	};
	
	struct imu_config
	{
		double acc_sigma;
		double gyro_sigma;
		double gravity;
	};
	
public:
	SLOM_BUILD_RANDOMVAR(
			Pose,
			((SO3, orient))
			((vec3, pos))
			((vec3, vel))
	)
	typedef std::deque<Pose_id> pose_list;
	
	/**
	 * struct which accumulates IMU-measurements.
	 * So far no bias is considered, and noise is assumed to be isotropic. 
	 */
	struct imu_accumulator {
		SO3 dq;
		vec3 dv;
		vec3 dx;
		double dt;
		
		imu_accumulator(){
			reset();
		}
		
		void reset() {
			dq = SO3();
			dv.setZero();
			dx.setZero();
			dt = 0;
		}
		void accumulate(
				const Eigen::Vector3d &acc, 
				const Eigen::Vector3d &gyr,
				const double &delta_t) {
			dq = dq * SO3::exp(gyr, delta_t);
			Eigen::Vector3d acc_world = dq * acc;
			dx += delta_t * (dv + acc_world * (delta_t/2));
			dv += delta_t * acc_world;
			dt += delta_t;
		}
		
		/**
		 * Integrate start pose to end pose, using accumulated IMU measurements and given gravity.
		 * gravity is always assumed to point in direction (0,0,-1).
		 */
		Pose integrate_pose(const Pose& start, double gravity) const {
			const SO3 &q0 = start.orient;
			const vec3 &x0 = start.pos, v0 = start.vel;
			
			SO3 q_end  = q0 * dq;
			vec3 v_end = v0 + q0 * dv;
			vec3 x_end = x0 + dt * v0 + q0 * dx;
			
			v_end[2] -= gravity * dt;
			x_end[2] -= gravity * dt * dt/2;
			
			return Pose(q_end, x_end, v_end);
		}
		
		
	};

// private: // FIXME privateness makes data extraction complicated
	SLOM_BUILD_MEASUREMENT(IMU_measurement, 
			Pose::DOF /* position, velocity and rotation */
			,
			((Pose, x0)) ((Pose, x1)) 
			,
			((imu_accumulator, delta))
			((const imu_config&, imu))
	)
	
	// TODO maybe add everything as mono measurement?
	SLOM_BUILD_MEASUREMENT_JACOBIAN(stereo_measurement, 4, 
			((Pose, pose))
			((landmark_var, lm)),
			((vec2, coord_left)) ((double, sigma_left))
			((vec2, coord_right)) ((double, sigma_right))
			((const camera_config&, conf))
	)
	
	SLOM_BUILD_MEASUREMENT_JACOBIAN(mono_measurement, 2, 
			((Pose, pose))
			((landmark_var, lm)),
			((vec2, coord)) ((double, sigma))
			((const pose_as_matrix&, pose2cam))
			((double, m_bound))
	)

	SLOM_BUILD_MEASUREMENT(small_speed_constraint, 3, 
			((Pose, pose)), ((double, vel_sigma)))
	
	SLOM_BUILD_MEASUREMENT(start_pose, 6+1,
			((Pose, pose)),
			((vec3, start))
	)

	
public:
	optimizer(const pose_as_matrix& imu2cam_left,
	          const pose_as_matrix& imu2cam_right,
	          int verbosity_level, double velocity_sigma);
public:
	
	static Eigen::Quaterniond orientation_from_gravity(const Eigen::Vector3d &init_gravity)
	{
		SO3::Matrix3 so;
		so.row(2) = init_gravity.normalized();
		so.row(1) = so.row(2).cross(Eigen::Vector3d(1,0,0)).normalized();
		so.row(0) = so.row(1).cross(so.row(2));
		
		//std::cerr << so << std::endl;
		return Eigen::Quaterniond(so);
	}
	
	void init(const Eigen::Vector3d &init_pos, 
			const Eigen::Vector3d &init_gravity);

	bool initialized() const
	{
		boost::mutex::scoped_lock lock(poses_mutex);
		
		return !poses.empty();
	}
	
	bool next_frame();
	
	landmark_id add_landmark(const landmark_var& pos)
	{
		++landmarks;
		return e.insertRV(pos);
	}
	
	void add_landmark_measurement(const Eigen::Vector2d &p_left_normalized, const double &sigma_left,
								  const Eigen::Vector2d &p_right_normalized, const double &sigma_right, 
								  const vslam::map::landmark &m);
	
	void add_mono_landmark_measurement(const Eigen::Vector2d &p_cam_normalized, const double &sigma,
								  const vslam::map::landmark &m, bool is_left);

	void add_imu_measurement(const Eigen::Vector3d &acc,
							 const Eigen::Vector3d &gyr,
							 const double &delta_t)
	{
		// integrate IMU-measurements, make last pose invalid
		poses_back_valid = false;
		
		last_imu.accumulate(acc, gyr, delta_t);
		current_pose_valid_ = false;
	}

	
	/**
	 * Get the current pose based on last estimation and integration of 
	 * following IMU-measurements.
	 * FIXME shall be usable concurrently with \c step
	 */
	const pose_as_matrix& current_imu2world()
	{
		if (!current_pose_valid_)
		{
			boost::mutex::scoped_lock lock(poses_mutex);
			const Pose& pose = last_imu.integrate_pose(*poses.back(), imu_conf_.gravity);
			
			current_pose_ << pose.orient.toRotationMatrix(), pose.pos;
			
			current_pose_valid_ = true;
		}
		return current_pose_;
	}
	
	/**
	 * returns reference pose from last frame.
	 * can be accessed by members orient and pos.
	 */
	const Pose_id last_pose() const
	{
		boost::mutex::scoped_lock lock(poses_mutex);
		
		return poses.back();
	}
	

	/**
	 *  One optimization step.
	 */
	void step(int max_steps = 1);

	template<typename Callback>
	void big_lock(Callback c) const
	{
		boost::mutex::scoped_lock lock(big_lock_mutex);

		c();
	}

	const pose_list &get_poses() const
	{
		return poses;
	}
	
	void dump_jacobian(std::ostream &out)
	{
		e.dumpJacobian(out);
	}

private:
	
	camera_config camera_conf_;
	
	double vel_sigma;
	
	imu_config imu_conf_;

	mutable pose_as_matrix current_pose_;
	mutable bool current_pose_valid_;
	
	
	
	// Data storage for SLoM:
	pose_list poses;
	mutable boost::mutex poses_mutex;
	mutable boost::mutex big_lock_mutex;
	int landmarks;
	int stereo_measurements;
	int mono_measurements;
	
	imu_accumulator last_imu;
	
	SLOM::Estimator e;
	SLOM::CallBack cb;
	
	int verbosity_level;
	
	// new IMU-measurements make the last pose invalid
	bool poses_back_valid;
	// when new variables are inserted, SLoM must be reinitialized
	bool needs_reinitialization;
	
	
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

} // namespace vslam
#endif // __VSLAM_CORE_OPTIMIZER_HPP__
