/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_CORE_POSE_AS_MATRIX_HPP__
#define __VSLAM_CORE_POSE_AS_MATRIX_HPP__

#include <Eigen/Core>
#include <Eigen/Geometry>
#ifdef DEPRECATED
#include <trafo3d/matTransf.h>
#endif //DEPRECATED

#ifndef NO_MTK_SUPPORT
#include <mtk/types/SOn.hpp>
#endif //NO_MTK_SUPPORT

namespace vslam {

struct pose_as_matrix : public Eigen::Matrix<double, 3, 4, Eigen::DontAlign | Eigen::RowMajor> {
	typedef Eigen::Matrix<double, 3, 4, Eigen::DontAlign | Eigen::RowMajor> base;
	typedef Eigen::Matrix<double, 3, 1> vec;
	typedef Eigen::BlockReturnType<base, 3, 3>::Type rotation_type;
	typedef Eigen::BlockReturnType<base, 3, 1>::Type translation_type;

	const rotation_type get_rotation() const
	{
		return corner<3,3>(Eigen::TopLeft);
	}
	rotation_type get_rotation()
	{
		return corner<3,3>(Eigen::TopLeft);
	}
	const translation_type get_translation() const
	{
		return corner<3,1>(Eigen::TopRight);
	}
	translation_type get_translation()
	{
		return corner<3,1>(Eigen::TopRight);
	}
	
	vec apply(const vec& x) const {
		vec res;
		res = get_rotation() * x + get_translation();
		return res;
	}
	
	vec operator*(const vec& x) const
	{
		return apply(x);
	}
	
	pose_as_matrix operator*(const pose_as_matrix& m) const
	{
		pose_as_matrix ret;
		ret << get_rotation() * m.get_rotation(), get_rotation()*m.get_translation() + get_translation();
		return ret;
	}
	
	pose_as_matrix() 
	{
		setIdentity();
	}
	
	pose_as_matrix(const Eigen::Quaterniond& q, const Eigen::Vector3d &v)
	{
		*this << q.toRotationMatrix(), v;
	}
	
#ifdef DEPRECATED
	template<class T>
	pose_as_matrix(const MatTransform<T>& trafo) {
		*this = trafo;
	}
	
	
	template<class T>
	pose_as_matrix& operator=(const MatTransform<T>& trafo) {
		base::operator=(Eigen::Matrix<T, 3, 4, Eigen::RowMajor>::Map(trafo.data));
		return *this;
	}
#endif //DEPRECATED
	pose_as_matrix inverse() const
	{
		rotation_type rot = get_rotation();
		pose_as_matrix ret;
		ret << rot.transpose(), -rot.transpose() * get_translation();
		return ret;
	}
	
	void normalize()
	{
		Eigen::Quaterniond q(get_rotation());
		get_rotation() = q.normalized().toRotationMatrix();
	}
	
	
#ifndef NO_MTK_SUPPORT
	// MTK requirements:

	typedef MTK::SO3<double> SO3;

	enum {DOF = SO3::DOF + 3};
	typedef double scalar;
	
	void boxplus(const MTK::vectview<const double, DOF> &vec, double scale=1)
	{
		SO3 q = Eigen::Quaterniond(get_rotation());
		q.boxplus(vec.data(), scale);
		get_rotation() = q.toRotationMatrix();
		get_translation() += scale * vec.end<3>();
	}
	
	void boxminus(MTK::vectview<double, DOF> res, const pose_as_matrix& other) const
	{
		SO3 q1 = Eigen::Quaterniond(get_rotation());
		SO3 q2 = Eigen::Quaterniond(other.get_rotation());
		q1.boxminus(res.data(), q2);
		res.end<3>() = get_translation() - other.get_translation();
	}
#endif //NO_MTK_SUPPORT
	
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};



inline std::istream& operator>>( std::istream& in, pose_as_matrix& m )
{
	/* possible formats:
	 * a11 a12 a13 t1 a21 a22 a23 t2 a31 a32 a33 t3
	 * [ a11,a12,a13,t1,a21,a22,a23,t2,a31,a32,a33,t3 ]
	 * [ a11,a12,a13,t1 ] [ a21,a22,a23,t2 ] [ a31,a32,a33,t3 ]
	 */
	char c = 0;
	in >> c;
	if ( c == '[' ) {
		for (int i=0; i<3; ++i) {
			for (int j=0; j<4; ++j) {
				in >> m(i,j) >> c;
				if ( j<3 && c!=',' ) { in.clear(std::ios_base::badbit); return in; }
			}
			if ( i<2 ) {
				if (c==']') {
					in >> c;
					if ( c!='[' ) { in.clear(std::ios_base::badbit); return in; }
				}
			} else {
				//in >> c;
				if ( c!=']' ) { in.clear(std::ios_base::badbit); return in; }
			}
		}
	} else {
		in.putback(c);
		for (int i=0; i<3; ++i) {
			for (int j=0; j<4; ++j) {
				in >> m(i,j);
			}
		}
	}
	return in;
}

inline std::ostream& operator<<(std::ostream& out, const pose_as_matrix &pose)
{
	// Eigen's operator<< ignores/changes current precision settings, so we output manually
	for(int r=0; r<3; ++r)
		for(int c=0; c<4; ++c)
			out << pose(r,c) << " ";
	return out;
}

} // namespace vslam





#endif // __VSLAM_CORE_POSE_AS_MATRIX_HPP__
