/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_DENSE_STEREO_HPP__
#define __VSLAM_DENSE_STEREO_HPP__

#include <vector>

#include <Eigen/Core>
#include <cv.h>
#include <vslam/camera/camera.hpp>

namespace vslam {

class dense_stereo
{
public:
	typedef std::vector<Eigen::Vector3d> pointcloud;
	typedef Eigen::Matrix<double, 3, 3, Eigen::RowMajor> rotation_mat;
	
	dense_stereo(const vslam::stereo_camera &cam, int min_disparity, int disparity_range)
		: bm_(0, disparity_range),
		  left_(cv::Size(256, 192), CV_8U), // FIXME hardcoded downscale
		  right_(left_.size(), CV_8U)
	{
		// additional BM parameters
		bm_.state->minDisparity = min_disparity;
		bm_.state->speckleWindowSize = 8;
		bm_.state->speckleRange = 4;

		int width, height;
		double f_x, f_y, c_x, c_y;
		double k1, k2, k3, p1, p2;

		// left camera
		cam.get_cam(true).getParameter(width, height, f_x, f_y, c_x, c_y);
		double scale_factor = left_.size().width / double(width);
		f_x *= scale_factor;
		f_y *= scale_factor;
		c_x *= scale_factor;
		c_y *= scale_factor;

		double cameraMatrix1_d[] = {f_x,   0, c_x,
									0,   f_y, c_y,
									0,     0,   1};
		cv::Mat cameraMatrix1(3,3,CV_64F, cameraMatrix1_d);

		cam.get_cam(true).get_distortion(k1, k2, p1, p2, k3);
  
		double distCoeffs1_d[] = {k1, k2, p1, p2, k3};
		cv::Mat distCoeffs1(5,1,CV_64F, distCoeffs1_d);
		
		// right camera
		cam.get_cam(false).getParameter(width, height, f_x, f_y, c_x, c_y);
		scale_factor = right_.size().width / double(width);
		f_x *= scale_factor;
		f_y *= scale_factor;
		c_x *= scale_factor;
		c_y *= scale_factor;

		double cameraMatrix2_d[] = {f_x,   0, c_x,
									0,   f_y, c_y,
									0,     0,   1};
		cv::Mat cameraMatrix2(3,3,CV_64F, cameraMatrix2_d);

		cam.get_cam(false).get_distortion(k1, k2, p1, p2, k3);
  
		double distCoeffs2_d[] = {k1, k2, p1, p2, k3};
		cv::Mat distCoeffs2(5,1,CV_64F, distCoeffs2_d);

		Eigen::Matrix<double, 3, 3, Eigen::RowMajor> R_d = cam.get_left2right().get_rotation();
		Eigen::Vector3d T_d = cam.get_left2right().get_translation();
		cv::Mat R(3,3,CV_64F,R_d.data());

		cv::Mat T(3,1,CV_64F,T_d.data());

		const cv::Size &size = left_.size();

		cv::Mat R1(3,3,CV_64F,left2rectified_.data());
		cv::Mat R2, P1, P2;

		cv::stereoRectify(cameraMatrix1, distCoeffs1, cameraMatrix2, distCoeffs2, size, R, T,
						  R1, R2, P1, P2, Q, .8, cv::Size(), 0, 0, cv::CALIB_ZERO_DISPARITY);

		cv::initUndistortRectifyMap(cameraMatrix1, distCoeffs1, R1, P1, size, CV_32FC1, map1_1, map2_1);

		cv::initUndistortRectifyMap(cameraMatrix2, distCoeffs2, R2, P2, size, CV_32FC1, map1_2, map2_2);
	}

	void operator()(const cv::Mat &left, const cv::Mat &right, pointcloud &c) const
	{
		cv::resize(left, left_, left_.size());
		cv::resize(right, right_, right_.size());
		
		cv::remap(left_, left_undistrect_, map1_1, map2_1, cv::INTER_LINEAR);
		cv::remap(right_, right_undistrect_, map1_2, map2_2, cv::INTER_LINEAR);

		bm_(left_undistrect_, right_undistrect_, disparity_, CV_32F);

		cv::reprojectImageTo3D(disparity_, _3dImage_, Q, true);

		typedef cv::Vec<float,3> VT;
		for (cv::MatConstIterator_<VT> i = _3dImage_.begin<VT>();
			 i != _3dImage_.end<VT>(); ++i)
		{
			// process only inliers
			if ((*i)[2] < 10000)
			{
				c.push_back(Eigen::Vector3d(-(*i)[0], -(*i)[1], -(*i)[2]));
			}
		}
	}

	const rotation_mat& left2rectified() const
	{
		return left2rectified_;
	}


	
	const cv::Mat &left() const
	{
		return left_;
	}

	const cv::Mat &right() const
	{
		return right_;
	}

	const cv::Mat &left_undistrect() const
	{
		return left_undistrect_;
	}

	const cv::Mat &right_undistrect() const
	{
		return right_undistrect_;
	}

	const cv::Mat &disparity() const
	{
		return disparity_;
	}

private:
	cv::Mat Q;

	rotation_mat left2rectified_;

	cv::Mat map1_1, map2_1;
	cv::Mat map1_2, map2_2;

	mutable cv::StereoBM bm_;

	mutable cv::Mat left_, right_;
	mutable cv::Mat left_undistrect_, right_undistrect_;
	mutable cv::Mat disparity_;
	mutable cv::Mat _3dImage_;
};
	
} // namespace vslam

#endif // __VSLAM_DENSE_STEREO_HPP__
