/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_FEATURES_FEATURE_POINTS_HPP__
#define __VSLAM_FEATURES_FEATURE_POINTS_HPP__

#include <algorithm>
#include <vector>

#include <cv.h>

#include "../util/const_sub_container.hpp"

namespace vslam {

class feature
{
public:
	typedef cv::KeyPoint keypoint;
	typedef const_sub_container<std::vector<float> > signature;

	feature(const keypoint &kp,
			const signature::const_iterator &sig_begin,
			const signature::const_iterator &sig_end)
		: keypoint_(kp),
		  sig_(sig_begin, sig_end)
	{}
	
	const keypoint &kp() const
	{
		return keypoint_;
	}

	const signature &sig() const
	{
		return sig_;
	}
	
private:
	keypoint keypoint_;
	signature sig_;
};

class feature_points
{
	typedef std::vector<feature> feature_vector;
	typedef std::vector<float> descriptor_vector;
	
public:
	typedef feature value_type;
	typedef value_type& reference;
	typedef const value_type& const_reference;
	
	typedef feature_vector::const_iterator const_iterator;
	typedef descriptor_vector::size_type size_type;

	void clear()
	{
		features_.clear();
		descriptors_.clear();
	}
	
	void reserve(unsigned descriptor_size,
				 size_type size)
	{
		features_.reserve(size);
		descriptors_.reserve(descriptor_size * size);
	}

	// NOTE: call reserve() first to avoid re-allocations!
	void push_back(const feature::keypoint &kp,
				   const feature::signature::const_iterator &sig_begin,
				   const feature::signature::const_iterator &sig_end)
	{
		assert(features_.size() < features_.capacity());
		
		feature::signature::const_iterator::difference_type desc_size = sig_end - sig_begin;
		
		std::copy(sig_begin, sig_end, std::back_inserter(descriptors_));
		features_.push_back(feature(kp, descriptors_.end() - desc_size, descriptors_.end()));
	}

	void populate(const std::vector<cv::KeyPoint> &keypoints,
				  const std::vector<float> &descriptors,
				  int descriptor_size)
	{
		assert(keypoints.size() * descriptor_size == descriptors.size());

		reserve(descriptor_size,keypoints.size());
		
		std::vector<float>::const_iterator desc = descriptors.begin();
		for (std::vector<cv::KeyPoint>::const_iterator i = keypoints.begin();
			 i != keypoints.end(); ++i)
		{
			push_back(*i, desc, desc + descriptor_size);
			std::advance(desc, descriptor_size);
		}
	}

	const value_type &front() const
	{
		return features_.front();
	}

	const_iterator begin() const
	{
		return features_.begin();
	}

	const_iterator end() const
	{
		return features_.end();
	}

	const_reference operator[](size_type n) const
	{
		assert(n < size());
		return features_[n];
	}

	size_type size() const
	{
		return features_.size();
	}
	
private:
	feature_vector features_;
	descriptor_vector descriptors_;
};

} // namespace vslam
	
#endif // __VSLAM_FEATURES_FEATURE_POINTS_HPP__
