/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_FEATURES_SIFT_EXTRACTOR_HPP__
#define __VSLAM_FEATURES_SIFT_EXTRACTOR_HPP__

#include <stdlib.h>

#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>

#include <cv.h>
#include <cv.hpp>
#include <cvaux.hpp>
#include <highgui.h>

#include "feature_points.hpp"

namespace vslam {
namespace features {

class sift_extractor
{
	void append_feature(feature_points &features,
						const float &x, const float &y,
						const float &scale, const float &angle,
						const std::vector<float> &desc,
						bool do_resize, const float &factor_x, const float &factor_y) const
	{
		cv::KeyPoint kp(x, y,
						scale,
						angle // FIXME: convert into weird OpenCV angle format
			);

		if (do_resize)
		{
			kp.pt.x *= factor_x;
			kp.pt.y *= factor_y;
		}

		features.push_back(kp, desc.begin(), desc.end());
	}

	void fail(const char *msg)
	{
		std::cerr << "SIFT: error: " << msg << std::endl;
		throw msg;
	}
	
public:
	sift_extractor()
		: tmp_img_(cv::Size(512, 384), CV_8U)
	{}

	void operator()(const cv::Mat &img,
					feature_points &features,
					unsigned keep_best,
					unsigned *num_keypoints=0)
	{
		features.clear();
		
		// scale down if necessary
		bool do_resize = img.size().width > tmp_img_.size().width;
		if (do_resize)
			cv::resize(img, tmp_img_, tmp_img_.size());
		const cv::Mat &img_ = do_resize ? tmp_img_ : img;

		float factor_x = 1., factor_y = 1.;
		if (do_resize)
		{
			factor_x = img.size().width / tmp_img_.size().width;
			factor_y = img.size().height / tmp_img_.size().height;
		}

		// determine filenames
		std::stringstream infile_;
		infile_ << "sift-in-" << this << ".pgm";
		std::string infile = infile_.str();
		
		std::stringstream outfile_;
		outfile_ << "sift-out-" << this << ".key";
		std::string outfile = outfile_.str();

		// write image to temporary .pgm
		cv::imwrite(infile.c_str(), img_);

		// run SIFT
		std::stringstream cmd;
		cmd << "sift < " << infile << " > " << outfile;
		if (0 != system(cmd.str().c_str()))
			fail("error running sift executable");

		// read output from file
		std::ifstream f(outfile.c_str());
		
		if (!f.is_open())
			fail("could not open SIFT output");

		// parse header
		int n, desc_len;

		f >> n >> desc_len;
		if (f.eof())
			fail("premature EOF");
		if (desc_len != descriptor_size)
			fail("descriptor size mismatch");

		features.reserve(descriptor_size, n);
		
		for (int i = 0;
			 f.is_open() && !f.eof() && f.good()
				 && i < n; ++i)
		{
			float row, col, scale, angle;

			f >> row >> col >> scale >> angle;

			std::vector<float> desc;
			desc.reserve(descriptor_size);

			for (int k = 0; k < descriptor_size; ++k)
			{
				int d;

				f >> d;
				desc.push_back(d / 255.f);
			}

			append_feature(features, col, row,
						   scale, angle,
						   desc,
						   do_resize, factor_x, factor_y);
		}

		if (num_keypoints)
			*num_keypoints = features.size();
	}
	
	enum
	{
		descriptor_size = 128
	};
	
	static
	const char *detector_id()
	{
		return "SIFT";
	}
	
	static
	const char *descriptor_id()
	{
		return "SIFT";
	}

private:
	cv::Mat tmp_img_;
};
	
} // namespace features
} // namespace vslam

#endif // __VSLAM_FEATURES_SIFT_EXTRACTOR_HPP__
