/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_FEATURES_SORTING_HPP__
#define __VSLAM_FEATURES_SORTING_HPP__

namespace vslam {
namespace features {

namespace internal {

static
void copy_features_unsorted(const std::vector<cv::KeyPoint> &keypoints,
							const std::vector<float> &descriptors,
							int descriptor_size,
							feature_points &features)
{
	features.reserve(descriptor_size, keypoints.size());

	std::vector<float>::const_iterator desc = descriptors.begin();
	for (std::vector<cv::KeyPoint>::const_iterator i = keypoints.begin();
		 i != keypoints.end(); ++i)
	{
		features.push_back(*i,
						   desc, desc + descriptor_size);

		std::advance(desc, descriptor_size);
	}
}

struct sort_handle
{
	sort_handle(const cv::KeyPoint *kp_, std::vector<float>::const_iterator desc_)
		: kp(kp_),
		  desc(desc_)
	{}
	
	const cv::KeyPoint *kp;
	std::vector<float>::const_iterator desc;
};
	
static
bool sort_by_response(const sort_handle &a, const sort_handle &b)
{
	return a.kp->response > b.kp->response;
}

static
void copy_n_best_features(const std::vector<cv::KeyPoint> &keypoints,
						  const std::vector<float> &descriptors,
						  int descriptor_size,
						  feature_points &features,
						  unsigned keep_best)
{
	std::vector<sort_handle> handles;
	handles.reserve(keypoints.size());

	std::vector<float>::const_iterator desc = descriptors.begin();
	for (std::vector<cv::KeyPoint>::const_iterator i = keypoints.begin();
		 i != keypoints.end(); ++i)
	{
		handles.push_back(sort_handle(&*i, desc));
		std::advance(desc, descriptor_size);
	}

	unsigned keep = std::min(keep_best, keypoints.size());

	if (keep < keypoints.size())
		std::nth_element(handles.begin(), handles.begin() + keep, handles.end(), sort_by_response);
	std::sort(handles.begin(), handles.begin() + keep, sort_by_response);

	features.reserve(descriptor_size, keep);
	for (std::vector<sort_handle>::const_iterator i = handles.begin();
		 i != handles.begin() + keep; ++i)
	{
		features.push_back(*i->kp,
						   i->desc, i->desc + descriptor_size);
	}
}

static
bool sort_keypoints_by_response(const cv::KeyPoint &a, const cv::KeyPoint &b)
{
	return a.response > b.response;
}

static 
void keep_n_best_keypoints(std::vector<cv::KeyPoint> &keypoints,
						   unsigned keep_best)
{
	unsigned keep = std::min(keep_best, keypoints.size());
		
	if (keep < keypoints.size())
	{
		std::nth_element(keypoints.begin(), keypoints.begin() + keep, keypoints.end(), sort_keypoints_by_response);
		keypoints.erase(keypoints.begin() + keep, keypoints.end());
	}
	std::sort(keypoints.begin(), keypoints.end(), sort_keypoints_by_response);
}
	
} // namespace internal

} // namespace features

} // namespace vslam

#endif // __VSLAM_FEATURES_SORTING_HPP__
