/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_FEATURES_STAR_SURF_EXTRACTOR_HPP__
#define __VSLAM_FEATURES_STAR_SURF_EXTRACTOR_HPP__

#include <iostream>

#include <cv.h>
#include <cv.hpp>
#include <cvaux.hpp>

#include "feature_points.hpp"
#include "sorting.hpp"

namespace vslam {
namespace features {

class star_surf_extractor
{
public:
	star_surf_extractor()
		: tmp_img_(cv::Size(512, 384), CV_8U),
		  star_(16, // maxSize
				5,  // response threshold
				10, // line threshold projected
				8,  // lineThresholdBinarized
				5   // suppressNonmaxSize
			  )
	{}

	void operator()(const cv::Mat &img,
					feature_points &features,
					unsigned keep_best,
					unsigned *num_keypoints = 0)
	{
		features.clear();
		
		std::vector<cv::KeyPoint> keypoints_tmp;
		std::vector<float> descriptors_tmp;
		
		extract(img, keypoints_tmp, descriptors_tmp, keep_best, num_keypoints);

		features.populate(keypoints_tmp, descriptors_tmp, descriptor_size);
	}

	enum
	{
		descriptor_size = 128
	};
	
	static
	const char *detector_id()
	{
		return "Star";
	}
	
	static
	const char *descriptor_id()
	{
		return "SURF";
	}
	
private:
	void extract(const cv::Mat &img,
				 std::vector<cv::KeyPoint> &keypoints, std::vector<float> &descriptors,
				 unsigned keep_best,
				 unsigned *num_keypoints)
	{
		// scale down if necessary
		bool do_resize = img.size().width > tmp_img_.size().width;
		if (do_resize)
			cv::resize(img, tmp_img_, tmp_img_.size());
		const cv::Mat &img_ = do_resize ? tmp_img_ : img;

		// detect features using FAST
		star_(img_, keypoints);
		
		if (num_keypoints)
			*num_keypoints = keypoints.size();
		
		internal::keep_n_best_keypoints(keypoints, keep_best);
		
		// run SURF with provided keypoints
		surf_(img_, cv::Mat(), keypoints, descriptors, true);

		// adjust keypoint coordinates
		if (do_resize)
		{
			float factor_x = img.size().width / tmp_img_.size().width;
			float factor_y = img.size().height / tmp_img_.size().height;
			
			for (std::vector<cv::KeyPoint>::iterator i = keypoints.begin();
				 i != keypoints.end(); ++i)
			{
				i->pt.x *= factor_x;
				i->pt.y *= factor_y;
			}
		}
	}
	
	cv::Mat tmp_img_;
	cv::StarDetector star_;
	cv::SURF surf_;
};
	
} // namespace features
} // namespace vslam

#endif // __VSLAM_FEATURES_STAR_SURF_EXTRACTOR_HPP__
