/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_FLANN_MATCHER_HPP__
#define __VSLAM_FLANN_MATCHER_HPP__

#include <flann.h>

namespace vslam {

class flann_matcher
{
	void run_flann(float *dataset, int rows, int cols,
				   float *testset, int t_rows,
				   float **dists, int **result)
	{
		const int nn = 1;

		*dists = new float[t_rows * nn];
		*result = new int[t_rows * nn];

		FLANNParameters p;
		memset(&p, 0, sizeof(p));
		p.random_seed = 23;

		/*
		// auto-tuning
		p.target_precision = 0.9;
		p.build_weight = 0.01;
		p.memory_weight = -1;
		p.sample_fraction = 1.;
		*/

		// manual algorithm/parameter selection
		p.target_precision = -1.;

		p.algorithm = LINEAR;

		/*
		// KDTREE
		p.algorithm = KDTREE;
		p.checks = 32;
		p.trees = 16;
		*/

		/*
		// KMEANS
		p.algorithm = KMEANS;
		p.checks = 16;
		p.cb_index = 1e9;
		p.branching = 16;
		p.iterations = 10;
		p.centers_init = CENTERS_RANDOM;
		*/

		flann_find_nearest_neighbors(dataset, rows, cols,
									 testset, t_rows,
									 *result, *dists,
									 nn, &p);

		//std::cerr << "FLANN params: " << int(p.algorithm) << " " << p.checks << " " << p.cb_index << " " << p.trees << " " << p.branching << " " << p.iterations << " " << int(p.centers_init) << " " << std::endl;
	}

	template<typename container>
	void copy_signatures(float *dst, const container &kp) const
	{
		for (typename container::const_iterator i = kp.begin();
			 i != kp.end(); ++i)
		{
			dst = std::copy(i->sig().begin(), i->sig().end(), dst);
		}
	}

public:
	template<typename a_container,
			 typename b_container,
			 typename MatchFunc,
			 typename NonMatchFuncA,
			 typename NonMatchFuncB>
	void operator()(int descriptor_size,
					const a_container &kp_a,
					const b_container &kp_b,
					MatchFunc f,
					NonMatchFuncA g,
					NonMatchFuncB h,
					bool report_non_matches = true)
	{
		// allocate memory for data sets
		float *set_a = new float[kp_a.size() * descriptor_size];
		float *set_b = new float[kp_b.size() * descriptor_size];

		// copy signatures
		copy_signatures(set_a, kp_a);
		copy_signatures(set_b, kp_b);

		// run flann a -> b
		float *dist_fwd;
		int *result_fwd;
		run_flann(set_b, kp_b.size(), descriptor_size, set_a, kp_a.size(), &dist_fwd, &result_fwd);

		// run flann b -> a
		float *dist_rev;
		int *result_rev;
		run_flann(set_a, kp_a.size(), descriptor_size, set_b, kp_b.size(), &dist_rev, &result_rev);

		// iterate over results and report matches/non-matches

		// fwd matches first
		float *dist = dist_fwd;
		int *b_idx_p = result_fwd;
		for (unsigned a_idx = 0;
			 a_idx < kp_a.size(); ++a_idx, ++b_idx_p, ++dist)
		{
			const int &b_idx = *b_idx_p;

			// check whether reverse result matches
			if (unsigned(result_rev[b_idx]) == a_idx)
			{
				// we have a match
				f(kp_a[a_idx], kp_b[b_idx], *dist);
			}
			else if (report_non_matches)
			{
				// not unique
				g(kp_a[a_idx], *dist, 0);
			}
		}

		// now rev matches
		dist = dist_rev;
		int *a_idx_p = result_rev;
		for (unsigned b_idx = 0;
			 b_idx < kp_b.size(); ++b_idx, ++a_idx_p, ++dist)
		{
			const int &a_idx = *a_idx_p;
			
			// check whether forward result matches
			if (unsigned(result_fwd[a_idx]) == b_idx)
			{
				// we have a match; already handled above
			}
			else if (report_non_matches)
			{
				// not unique
				g(kp_b[b_idx], 0, *dist);
			}
		}

		// free memory
		delete[] set_a;
		delete[] set_b;
		delete[] dist_fwd;
		delete[] result_fwd;
		delete[] dist_rev;
		delete[] result_rev;
	}	
};
	
} // namespace vslam

#endif // __VSLAM_FLANN_MATCHER_HPP__
