/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_FWD_REV_GRAPH_MATCHER_HPP__
#define __VSLAM_FWD_REV_GRAPH_MATCHER_HPP__

#include <vector>

#include <cv.h>

#include <Eigen/Core>
#include <tbb/tbb.h>

#include "match_graph_node.hpp"

//#define VSLAM_NONPARALLEL_MATCHER 1

namespace vslam {
	
// O(n*m) match with bookkeeping
	
template<typename FeatureDist,
		 typename A,
		 typename B>
class fwd_rev_graph_matcher
{
	typedef typename FeatureDist::val_type val_type;
	typedef typename FeatureDist::dist_type dist_type;

	typedef std::vector<float> desc_vector;
	typedef std::vector<val_type, Eigen::aligned_allocator<__m128i> > aligned_desc_vector;

	typedef std::vector<internal::match_graph_node<A,B,dist_type> > match_node_vector_a;
	typedef std::vector<internal::match_graph_node<B,A,dist_type> > match_node_vector_b;

	class outer_loop
	{
		int descriptor_size;
		
		match_node_vector_a &a_nodes_;
		const aligned_desc_vector &a_desc_;
		
		match_node_vector_b &b_nodes_;
		const aligned_desc_vector &b_desc_;
		
	public:
		outer_loop(int descriptor_size_,
				   match_node_vector_a &a_nodes, aligned_desc_vector &a_desc,
				   match_node_vector_b &b_nodes, aligned_desc_vector &b_desc)
			: descriptor_size(descriptor_size_),
			  a_nodes_(a_nodes),
			  a_desc_(a_desc),
			  b_nodes_(b_nodes),
			  b_desc_(b_desc)
		{
		}
		
		void operator()(const tbb::blocked_range<size_t> &r) const
		{
			// only a_nodes within range
			typename match_node_vector_a::iterator a_nodes_begin = a_nodes_.begin() + r.begin();
			typename match_node_vector_a::iterator a_nodes_end = a_nodes_.begin() + r.end();
			typename aligned_desc_vector::const_iterator a_desc = a_desc_.begin() + r.begin() * descriptor_size;
			
			// all b_nodes
			typename match_node_vector_b::iterator b_nodes_begin = b_nodes_.begin();
			typename match_node_vector_b::iterator b_nodes_end = b_nodes_.end();
			
			for (typename match_node_vector_a::iterator a = a_nodes_begin;
				 a != a_nodes_end; ++a)
			{
				assert(unsigned(descriptor_size * (b_nodes_end - b_nodes_begin)) == b_desc_.size());
				typename aligned_desc_vector::const_iterator b_desc = b_desc_.begin();
				for (typename match_node_vector_b::iterator b = b_nodes_begin;
					 b != b_nodes_end; ++b)
				{
					dist_type dist = FeatureDist::compute(a_desc, a_desc+descriptor_size, b_desc);
					a->update_unsafe(&*b, dist);
					b->update(&*a, dist);
					
					std::advance(b_desc, descriptor_size);
				}
				
				std::advance(a_desc, descriptor_size);
			}
		}
	};
	
private:
	template<typename feature_container>
	static void transform_descriptor(aligned_desc_vector &desc, unsigned int descriptor_size,
									 const feature_container &features)
	{
		assert(desc.size() == features.size() * descriptor_size);
		typename aligned_desc_vector::iterator out = desc.begin(); 
		for(typename feature_container::const_iterator i = features.begin();
			i != features.end(); ++i)
		{
			assert(i->sig().size() == descriptor_size && "Feature has wrong signature size!");
			out = std::transform(i->sig().begin(), i->sig().end(), out, &FeatureDist::convert_one);
		}
	}

	template<typename feature>
	static void ignore(const feature &l, dist_type d1, dist_type d2)
	{}

public:
	template<typename a_container,
			 typename b_container,
			 typename MatchFunc>
	void operator()(int descriptor_size,
					const a_container &kp_l,
					const b_container &kp_r,
					MatchFunc f)
	{
		(*this)(descriptor_size, kp_l, kp_r, f, ignore<typename a_container::value_type>, ignore<typename b_container::value_type>, false);
	}
	
	/**
	 * match left and right keypoints
	 * 
	 * signatures of MatchFunc f and NonMatchFunc g: 
	 * f(const a_container::value_type& l, const b_container::value_type& r,
	 *   dist_type dist)
	 * 
	 * g(const a_container::value_type& kp,
	 *   dist_type dist_l, dist_type dist_r)
	 * with dist_l > dist_r, iif kp is in left camera
	 */
	template<typename a_container,
			 typename b_container,
			 typename MatchFunc,
			 typename NonMatchFuncA,
			 typename NonMatchFuncB>
	void operator()(int descriptor_size,
					const a_container &kp_l,
					const b_container &kp_r,
					MatchFunc f,
					NonMatchFuncA g,
					NonMatchFuncB h,
					bool report_non_matches = true)
	{
		using namespace internal;
		//std::cout << "descriptor size: " << descriptor_size << std::endl;
	
		aligned_desc_vector desc_l(kp_l.size() * descriptor_size);
		aligned_desc_vector desc_r(kp_r.size() * descriptor_size);
		
		transform_descriptor(desc_l, descriptor_size, kp_l);
		transform_descriptor(desc_r, descriptor_size, kp_r);

		match_node_vector_b b_nodes;
		b_nodes.reserve(kp_r.size());
		for (typename b_container::const_iterator b = kp_r.begin();
			 b != kp_r.end(); ++b)
		{
			b_nodes.push_back(match_graph_node<B,A,dist_type>(&*b));
		}
		
		match_node_vector_a a_nodes;
		a_nodes.reserve(kp_l.size());
		for (typename a_container::const_iterator a = kp_l.begin();
			 a != kp_l.end(); ++a)
		{
			a_nodes.push_back(match_graph_node<A,B,dist_type>(&*a));
		}

#ifdef VSLAM_NONPARALLEL_MATCHER
		{
			outer_loop l(descriptor_size, a_nodes, desc_l, b_nodes, desc_r);
			l(tbb::blocked_range<size_t>(0,a_nodes.size()));
		}
#else
		tbb::parallel_for(tbb::blocked_range<size_t>(0,a_nodes.size()),
						  outer_loop(descriptor_size, a_nodes, desc_l, b_nodes, desc_r));
#endif

		for (typename match_node_vector_a::const_iterator a = a_nodes.begin();
			 a != a_nodes.end(); ++a)
		{
			float a_dist = FeatureDist::convert_back(a->nn_dist);
			if (a->nn)
			{
				const match_graph_node<B, A, dist_type> *b = a->nn;
				if(b->nn == &*a)
					f(*a->self, *b->self, a_dist);
				else if (report_non_matches)
					g(*a->self, a_dist, FeatureDist::convert_back(b->nn_dist));
			}
			else if (report_non_matches)
			{
				// can only happen, if b has no entries:
				g(*a->self, a_dist, 0);
			}
		}

		if (!report_non_matches)
			return;
		
		for (typename match_node_vector_b::const_iterator b = b_nodes.begin();
			 b != b_nodes.end(); ++b)
		{
			float b_dist = FeatureDist::convert_back(b->nn_dist);
			if (b->nn)
			{
				const match_graph_node<A, B, dist_type> *a = b->nn;
				if (a->nn == &*b)
				{
					// already processed
				}
				else
					h(*b->self, FeatureDist::convert_back(a->nn_dist), b_dist); // single right keypoint
			}
			else
			{
				// can only happen, if a has no entries:
				h(*b->self, 0, b_dist);
			}
		}
	}
};

} // namespace vslam
	
#endif // __VSLAM_FWD_REV_GRAPH_MATCHER_HPP__
