/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_MATCHER_NAIVE_MATCHER_HPP__
#define __VSLAM_MATCHER_NAIVE_MATCHER_HPP__

#include <Eigen/Core>

// naive O(n*m) match

template<typename FeatureDist>
class naive_matcher
{
	typedef typename FeatureDist::val_type val_type;
	typedef typename FeatureDist::dist_type dist_type;
	typedef std::vector<val_type, Eigen::aligned_allocator<__m128i> > aligned_vector;

public:
	template<typename MatchFunc,
			 typename NonMatchFunc>
	void operator()(int descriptor_size,
					const std::vector<cv::KeyPoint> &kp_l, const std::vector<float> &desc_l_,
					const std::vector<cv::KeyPoint> &kp_r, const std::vector<float> &desc_r_,
					MatchFunc f,
					NonMatchFunc g)
	{
		std::cout << "descriptor size: " << descriptor_size << std::endl;
	
		assert(kp_l.size() * descriptor_size == desc_l_.size());
		assert(kp_r.size() * descriptor_size == desc_r_.size());

		aligned_vector desc_l(desc_l_.size());
		aligned_vector desc_r(desc_r_.size());
		std::transform(desc_l_.begin(), desc_l_.end(), desc_l.begin(), &FeatureDist::convert_one);
		std::transform(desc_r_.begin(), desc_r_.end(), desc_r.begin(), &FeatureDist::convert_one);

		std::vector<cv::KeyPoint>::const_iterator a = kp_l.begin();
		typename aligned_vector::const_iterator a_desc = desc_l.begin();
		for (;a != kp_l.end(); ++a)
		{
			std::vector<cv::KeyPoint>::const_iterator best = kp_r.begin();
			typename aligned_vector::const_iterator best_desc = desc_r.begin();
			dist_type best_dist = 1000;

			{
				std::vector<cv::KeyPoint>::const_iterator b = kp_r.begin();
				typename aligned_vector::const_iterator b_desc = desc_r.begin();
		
				for (;b != kp_r.end(); ++b)
				{
					dist_type dist = FeatureDist::compute(a_desc, a_desc+descriptor_size, b_desc);
			
					if (dist < best_dist)
					{
						//std::cout << "better dist " << a->pt.x << "," << a->pt.y << " --> " << b->pt.x << "," << b->pt.y << ": " << dist << std::endl;
						best_dist = dist;
						best_desc = b_desc;
						best = b;
					}
			
					std::advance(b_desc, descriptor_size);
				}
			}

			std::vector<cv::KeyPoint>::const_iterator best_reverse = kp_r.begin();
			dist_type best_dist_reverse = 1000;

			{
				std::vector<cv::KeyPoint>::const_iterator a2 = kp_l.begin();
				typename aligned_vector::const_iterator a_desc2 = desc_l.begin();
		
				for (;a2 != kp_l.end(); ++a2)
				{
					dist_type dist = FeatureDist::compute(best_desc, best_desc+descriptor_size, a_desc2);
			
					if (dist < best_dist_reverse)
					{
						//std::cout << "better dist " << a->pt.x << "," << a->pt.y << " --> " << b->pt.x << "," << b->pt.y << ": " << dist << std::endl;
						best_dist_reverse = dist;
						best_reverse = a2;
					}
			
					std::advance(a_desc2, descriptor_size);
				}
			}

			if (!(best_reverse != a))
				f(*a, *best, best_dist);
			else
				g(*a, *best, best_dist);
			if(0)
			{
				std::cout << "reverse best differs: "
						  << best->pt.x << "," << best->pt.y << ": " << best_dist
						  << " vs. "
						  << best_reverse->pt.x << "," << best_reverse->pt.y << ": " << best_dist_reverse
						  << std::endl;
			}

			std::advance(a_desc, descriptor_size);
		}
	}
};

#endif // __VSLAM_MATCHER_NAIVE_MATCHER_HPP__
