/*
 *  Copyright (c) 2010--2011, Universitaet Bremen and DFKI GmbH
 *  All rights reserved.
 *
 *  Author: Rene Wagner <rene.wagner@dfki.de>
 *          Christoph Hertzberg <chtz@informatik.uni-bremen.de>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Universitaet Bremen nor the DFKI GmbH 
 *     nor the names of its contributors may be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __VSLAM_MATCHER_STEREO_MATCHER_HPP__
#define __VSLAM_MATCHER_STEREO_MATCHER_HPP__

#include <boost/bind.hpp>

#include <vslam/camera/camera.hpp>
#include <vslam/camera/rectified_stereo_camera.hpp>


#include "../features/feature_points.hpp"
#include "../features/distance_euclidean_float.hpp"
#include "../features/distance_sse.hpp"
#include "fwd_rev_graph_matcher.hpp"
#include "flann_matcher.hpp"
#include "match.hpp"
#include "match_utils.hpp"

namespace vslam {
namespace matcher {

//#define VSLAM_USE_FLANN 1
//#define VSLAM_USE_BRUTE_FORCE 1

class stereo_matcher
{
#ifdef VSLAM_USE_FLANN
	typedef flann_matcher nn_matcher;
#elif VSLAM_USE_BRUTE_FORCE
	typedef fwd_rev_graph_matcher<vslam::features::distance_euclidean_float, vslam::feature, vslam::feature> nn_matcher;
#else // SSE
	typedef fwd_rev_graph_matcher<vslam::features::distance_sse, vslam::feature, vslam::feature> nn_matcher;
#endif

	typedef stereo_matcher self;
	
	void append_non_match(std::vector<mono_feature> &non_matches,
						  const feature &f,
						  const float &dist_l, const float &dist_r)
	{
		//std::cout << "mono feature found: " << f.pt.x << "," << f.pt.y << std::endl;
		non_matches.push_back(mono_feature(descriptor_size_,
										   f.kp(), f.sig().begin(),
										   dist_l, dist_r));
	}

	void append_stereo_match(std::vector<stereo_match> &stereo_matches,
			const feature &left,
			const feature &right,
			const float &dist)
	{
		// extract stereo matches from feature matches:
		
		double error;

		camera::Vector v = cam_.stereo_image2world_point(left.kp().pt, right.kp().pt, error);
		if (error < error_threshold_)
		{
			stereo_matches.push_back(stereo_match(descriptor_size_,
												  left.kp(), left.sig().begin(),
												  right.kp(), right.sig().begin(),
												  dist, v));
		}
		else
		{
			// FIXME: don't silently drop these mono features?
		}
	}

public:
	stereo_matcher(int descriptor_size,
				   const vslam::stereo_camera &cam,
				   const double &error_threshold = 7.) // FIXME Hardcoded constant
		: descriptor_size_(descriptor_size),
		  cam_(cam),
		  error_threshold_(error_threshold)
	{}

	void operator()(const feature_points &left,
					const feature_points &right,
					std::vector<stereo_match> &stereo_matches,
					std::vector<mono_feature> &not_matched)
	{
		nn_matcher m;

		// find feature matches, and unmatched points:
		m(descriptor_size_,
		  left, right,
		  boost::bind(&self::append_stereo_match, this, boost::ref(stereo_matches), _1, _2, _3),
		  boost::bind(&self::append_non_match, this, boost::ref(not_matched), _1, _2, _3),
		  boost::bind(&self::append_non_match, this, boost::ref(not_matched), _1, _2, _3));
	}

private:
	const int descriptor_size_;
	const vslam::rectified_stereo_camera cam_;
	const double error_threshold_;
};

} // namespace matcher
} // namespace vslam
	
#endif // __VSLAM_MATCHER_STEREO_MATCHER_HPP__
